/*
 * UI-independant prototypes.
 */

void show_species_info(struct species_t *s, struct species_params_t *p);
void ui_init(void);
void ui_shutdown(void);
void show_fleet_menu_help(void);
void show_fleet_menu(const char *currfleet);
int read_int(const char *prompt);
int read_char(const char *prompt);
void show_help_landed(void);
void pause_prompt(void);
void game_error(char *message);
void show_build_help(void);
void show_help_docked(void);
void show_file_paged(const char *file);
float read_float(char *prompt);
void show_equip_help(void);
void show_help_orbit(void);
void show_tech_difference(struct tech_t *real, struct tech_t *req);
void show_species_info_box(struct species_t *s);
#ifdef _INCLUDE_VECTORS_H
void ui_init_scan_result(struct coord_t *c, int range);
void ui_add_scan_result_bodies(int bodies, struct coord_t *c);
void ui_add_scan_result_system(struct sector_t *s, struct coord_t *c);
void ui_add_scan_result_fleets(struct sector_t *s, struct coord_t *c);
void ui_show_chart(struct coord_t *c, int orientation, int emphasize);
#endif
void ui_show_charts_help();
void ui_alertbox(char *message);
void show_body_scan(struct body_t *b);
void show_zone_composition(GString *z);
void show_single_orbiter(int num, struct body_t *b);
void show_location_orbit(struct body_t *b, char *sector);
void show_species_editor_help(void);

void ui_config_error(const char *message, int err);
void ui_config_error_ss(const char *s1, const char *s2, int err);
void ui_config_error_sss(const char *s1, const char *s2, const char *s3,
  int err);
void ui_config_error_sscsc(const char *s1, const char *s2,
  char c1, const char *s3, char c2, int err);

void ui_message(const char *message);
void ui_message_sd(const char *s, int d);
void ui_message_sds(const char *message1, int d, const char *message2);
void ui_message_sdsds(const char *s1, int d1, const char *s2, int d2,
  const char *s3);
void ui_message_ss(const char *s1, const char *s2);
void ui_message_sss(const char *s1, const char *s2, const char *s3);

void ui_error(const char *message);
void ui_error_sd(const char *s, int d);
void ui_error_sss(const char *s1, const char *s2, const char *s3);

void ui_debug(const char *message);

int ui_species_edit_prompt(int points);
int ui_charts_prompt(void);

void show_location_landed(struct body_t *b);
void show_gstrlist(char *header, struct gstrlist_t *list);
void show_satellite_list(struct sector_t *sect, int body, int **list, int *num);
void show_location_docked(struct sector_t *sect);

GString *read_gstring(const char *prompt);
GString *read_accesscode(const char *prompt);
GString *read_gstring_scs(const char *s1, char c, const char *s2);
GString *ui_edit_block(const char *header, GString *orig, int maxlines);

void ui_configure(struct player_t *player);

struct specie_pane_t {
	struct populii_t *specie;
	GString	*zone;
	int		*selected;
	int		pos;
	int		page;
	int		drawn;	/* Has it been drawn? */
	int		starty;	/* Starting row on screen. */
	int		pagesize;	/*Number of specie per page */
};

#define null_specie_pane (struct specie_pane_t) { \
	specie: NULL, zone: NULL, selected: NULL, pos: 0, page: 0, drawn: 0, \
	starty: 0, pagesize: 0}

void draw_specie_pane(struct specie_pane_t *pane);
void draw_species_bar_selected(struct specie_pane_t *pane);
void draw_species_bar_unselected(struct specie_pane_t *pane);
void hide_specie_pane(struct specie_pane_t *pane);
void draw_species_bar_checkbox(struct specie_pane_t *pane);

void ui_show_chatline(const char *line);

struct ui_progbar_t {
  GString	*text;			/* Text in the progress bar. */
  int		min;			/* Lowest number in the progress bar. */
  int		max;			/* Highest number in the bar. */
  int		curr;			/* Current position of counter. */
  int		screenline;		/* Line on screen of bar. */
  int		cellsize;		/* Number of increments to inc cell. */
  int		currcell;		/* Number of current cell. */
  int		lastupdated;	/* Actual position of last draw. */
};
struct ui_progbar_t ui_progbar_init(const char *label, int min, int max);
void ui_progbar_advance(struct ui_progbar_t *bar, int to);
void ui_progbar_destroy(struct ui_progbar_t *bar);

struct pane_t {
 	int	ymin;
	int	ymax;
	int	xmin;
	int	xmax;
#ifdef CURSES
	WINDOW *win;
#else
	void *win_void;
#endif
};

struct listpane_data_t {
	int	size;	/* Size of each data item. */
	int	num;	/* Number of data items. */
	char *data;	/* Data itself. */
};

struct listpane_t {
	struct pane_t pane;
	struct listpane_data_t data;
	void (*drawcode)(struct listpane_t *pane, int num, int style);
    int pagesize;
    int page;
    int lin;
};

struct listpane_t *ui_listpane_create(struct listpane_data_t *data,
  float ymin, float xmin, float ymax, float xmax,
  void pane_draw_item(struct listpane_t *pane, int num, int style));

void draw_habitatpane_line(struct listpane_t *pane, int num, int style);
void draw_listpane(struct listpane_t *pane);
int ui_habitat_edit_prompt(int points);
void listpane_down(struct listpane_t *pane);
void listpane_up(struct listpane_t *pane);
void listpane_pagedown(struct listpane_t *pane);
void listpane_pageup(struct listpane_t *pane);
void dispose_listpane(struct listpane_t *pane);

#define STYLE_UNSELECTED	0
#define STYLE_SELECTED		1

#ifdef _INCLUDE_POPULII_H
void ui_show_pop_totals(struct pop_totals_t *p);
#endif

void show_population_menu(void);
