#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ncurses.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>
#include <math.h>
#include "microcosm.h"
#include "util.h"
#include "populii.h"
#include "chronos.h"
#include "fleets.h"
#include "vectors.h"
#include "sectors.h"
//#include "util.h"
#include "ui_all.h"

int	g_fleet_errno=0;

void dispose_item(struct item_t *i)
{
  g_string_free(i->base,TRUE);
  g_string_free(i->cargo,TRUE);
  dispose_item_popstrs(i);
}

void destroy_item(struct item_t *i)
{
  dispose_item(i);
  *i=destroyed_item;
}

void handle_item_line(struct item_t *i, char *line)
{
  GString *cmd,*args;
  split(line,&cmd,&args);
  if (strcmp(cmd->str,"Add_On")==0) {
    i->add_on=1;
  } else if (strcmp(cmd->str,"BuildBio")==0) {
    i->buildtech.bio=atoi(args->str);
  } else if (strcmp(cmd->str,"BuildEMF")==0) {
    i->buildtech.emf=atoi(args->str);
  } else if (strcmp(cmd->str,"BuildGrav")==0) {
    i->buildtech.grav=atoi(args->str);
  } else if (strcmp(cmd->str,"BuildLogic")==0) {
    i->buildtech.logic=atoi(args->str);
  } else if (strcmp(cmd->str,"BuildMech")==0) {
    i->buildtech.mech=atoi(args->str);
  } else if (strcmp(cmd->str,"BuildPsi")==0) {
    i->buildtech.psi=atoi(args->str);
  } else if (strcmp(cmd->str,"BuildQuantum")==0) {
    i->buildtech.quantum=atoi(args->str);
  } else if (strcmp(cmd->str,"Cargo")==0) {
    g_string_assign(i->cargo,args->str);
  } else if (strcmp(cmd->str,"CargoMax")==0) {
    i->cargomax=atoi(args->str);
  } else if (strcmp(cmd->str,"CargoQty")==0) {
    i->cargoqty=atoi(args->str);
  } else if (strcmp(cmd->str,"CargoUse")==0) {
    i->cargouse=atoi(args->str);
  } else if (strcmp(cmd->str,"Class")==0) {
    i->class=atoi(args->str);
    if (i->class==ICLASS_DESTROYED) {
      destroy_item(i);
    }
  } else if (strcmp(cmd->str,"Completion")==0) {
    i->cyclestocompletion=atoi(args->str);
  } else if (strcmp(cmd->str,"Crew")==0) {
    i->crew=atoi(args->str);
  } else if (strcmp(line,"Custom")==0) {
    i->modified=MODDED_CUSTOM;
  } else if (strcmp(cmd->str,"DefaultEquipped")==0) {
    i->defaultequipped=atoi(args->str);
    i->equipped=i->defaultequipped;
  } else if (strcmp(cmd->str,"Equipped")==0) {
    if (args->len) {
      i->equipped=atoi(args->str);
    } else {
      ui_message("Legacy Equipped keyword used without argument.\n");
      i->equipped=1;
    }
  } else if (strcmp(cmd->str,"Flux")==0) {
    i->flux=atoi(args->str);
  } else if (strcmp(cmd->str,"HP")==0) {
    i->hp=atoi(args->str);
  } else if (strcmp(cmd->str,"Immobile")==0) {
    i->immobile=1;
  } else if (strcmp(cmd->str,"Mass")==0) {
    i->mass=atoi(args->str);
  } else if (strcmp(cmd->str,"MaxCrew")==0) {
    i->maxcrew=atoi(args->str);
  } else if (strcmp(cmd->str,"MaxHP")==0) {
    i->maxhp=atoi(args->str);
    i->hp=i->maxhp;
  } else if (strcmp(cmd->str,"MinCrew")==0) {
    i->mincrew=atoi(args->str);
  } else if (strcmp(line,"Modified")==0) {
    if (i->modified==0) {
      i->modified=1;
    }
  } else if (strcmp(cmd->str,"PopUpdated")==0) {
    ui_config_error("Deprecated PopUpdated keyword in item.",-1);
    i->updated=atoi(args->str);
  } else if (strcmp(cmd->str,"Population")==0) {
    add_popstr_to_item(i,args->str);
  } else if (strcmp(cmd->str,"RegenTime")==0) {
    i->regentime=atoi(args->str);
  } else if (strcmp(cmd->str,"Strength")==0) {
    i->strength=atoi(args->str);
  } else if (strcmp(cmd->str,"Surface")==0) {
    i->surface=atoi(args->str);
  } else if (strcmp(cmd->str,"Updated")==0) {
    i->updated=atoi(args->str);
  } else if (strcmp(cmd->str,"Volume")==0) {
    i->volume=atoi(args->str);
  } else if (cmd->str[0]!='\0') {
    printw("Unknown item line '%s'\n",line);
  }
  g_string_free(cmd,TRUE);
  g_string_free(args,TRUE);
  refresh();
}

void load_item(struct item_t *i, char *itemname, struct ship_t *s)
{
  GString	*filename=g_string_new(g_Config.datapath->str);
  FILE		*f;
  char		buf[1024];
  int		len;
  g_string_append(filename,MAKROKOSMOS_ITEM_PATH);
  g_string_append(filename,itemname);
  f=fopen(filename->str,"r");
  g_string_assign(i->base,itemname);
  if (f) {
    while (fgets(&buf[0],1024,f)) {
      len=strlen(&buf[0]);
      if ((len) && (buf[len-1]=='\n')) {
        buf[len-1]='\0';
      }
      if (buf[0]!='#') {
        handle_item_line(i,&buf[0]);
      }
    }
    fclose(f);
  }
  g_string_free(filename,TRUE);
}

void add_item_to_ship(struct ship_t *s, char *itemname)
{
  s->numitems++;
  s->items = g_realloc(s->items,sizeof(struct item_t)*s->numitems);
  s->items[s->numitems-1]=null_item;
  load_item(&s->items[s->numitems-1],itemname,s);
}

void add_ingredient(struct recipe_t *r, GString *s)
{
  r->numingredients++;
  r->ingredient=g_realloc(r->ingredient,sizeof(GString *)*r->numingredients);
  r->ingredientqty=g_realloc(r->ingredientqty,sizeof(int)*r->numingredients);
  r->ingredient[r->numingredients-1]=g_string_new(s->str);
  r->ingredientqty[r->numingredients-1]=0;
}

int ingredientidx_c(struct recipe_t *r, GString *s)
{
  int x=0;
  while (x<r->numingredients) {
    if (gstreq(s,r->ingredient[x])) {
      return x;
    }
    x++;
  }
  add_ingredient(r,s);
  return r->numingredients-1;
}

void add_ingredient_qty(struct recipe_t *stuff, GString *thing, int qty)
{
  int ingredient=ingredientidx_c(stuff,thing);
  if (ingredient>=0) {
    stuff->ingredientqty[ingredient]+=qty;
  }
}

void handle_recipe_line(char *line, struct recipe_t *r)
{
  GString *cmd,*args;
  split(line,&cmd,&args);
  if (strcmp(cmd->str,"BuildBio")==0) {
    r->tech.bio=atoi(args->str);
  } else if (strcmp(cmd->str,"BuildEMF")==0) {
    r->tech.emf=atoi(args->str);
  } else if (strcmp(cmd->str,"BuildGrav")==0) {
    r->tech.grav=atoi(args->str);
  } else if (strcmp(cmd->str,"BuildLogic")==0) {
    r->tech.logic=atoi(args->str);
  } else if (strcmp(cmd->str,"BuildMech")==0) {
    r->tech.mech=atoi(args->str);
  } else if (strcmp(cmd->str,"BuildPsi")==0) {
    r->tech.psi=atoi(args->str);
  } else if (strcmp(cmd->str,"BuildQuantum")==0) {
    r->tech.quantum=atoi(args->str);
  } else if (strcmp(cmd->str,"BuildTime")==0) {
    r->time=atoi(args->str);
  } else if (args->len) {
    add_ingredient_qty(r,cmd,atoi(args->str));
  } else {
    printw("Unknown item line %s\n",line);
    refresh();
  }
  g_string_free(cmd,TRUE);
  g_string_free(args,TRUE);
}

void load_recipe(struct recipe_t *r, char *name)
{
  GString *filename=g_string_new(g_Config.datapath->str);
  FILE *f;
  char buf[1024];
  int len;
  *r=null_recipe;
  g_string_append(filename,MAKROKOSMOS_RECIPE_PATH);
  g_string_append(filename,name);
  f=fopen(filename->str,"r");
  if (f) {
    while (fgets(&buf[0],1024,f)) {
      len=strlen(&buf[0]);
      if ((len) && (buf[len-1]=='\n')) {
        buf[len-1]='\0';
      }
      if ((buf[0]!='#') && (buf[0]!='\0')) {
        handle_recipe_line(&buf[0],r);
      }
    }
    fclose(f);
  }
  g_string_free(filename,TRUE);
}

void dispose_recipe(struct recipe_t *r)
{
  int x=0;
  while (x<r->numingredients) {
    g_string_free(r->ingredient[x],TRUE);
    x++;
  }
  g_free(r->ingredient);
  g_free(r->ingredientqty);
}

int stuffqty(struct recipe_t *r,GString *s)
{
  int x=0;
  while (x<r->numingredients) {
    if (gstreq(s,r->ingredient[x])) {
      return r->ingredientqty[x];
    }
    x++;
  }
  return 0;
}

int ingredients_obtained(struct recipe_t *r, struct ship_t *s,
  struct recipe_t *stuff, int *failure)
{
  int x=0;
  int y=0;
  int retval=1;
  int needed;
  int	*ingredientsneeded=g_malloc(r->numingredients*sizeof(int));
  while (x<r->numingredients) {
    ingredientsneeded[x]=r->ingredientqty[x];
    ingredientsneeded[x]-=lesser(stuffqty(stuff,r->ingredient[x]),
      ingredientsneeded[x]);
    x++;
  }
  y=0;
  x=0;
  while (x<r->numingredients) {
    while (y<s->numitems) {
      switch (s->items[y].class) {
      case ICLASS_HOLD:
      case ICLASS_FACTORY:
        if (gstreq(s->items[y].cargo,r->ingredient[x])) {
          needed=lesser(s->items[y].cargoqty,ingredientsneeded[x]);
          add_ingredient_qty(stuff,r->ingredient[x],needed);
          s->items[y].cargoqty-=needed;
          ingredientsneeded[x]-=needed;
          if (ingredientsneeded[x]<=0) {
            y=s->numitems;
          }
        }
      default:
      }
      y++;
    }
    if (ingredientsneeded[x]) {
      retval=0;
      if (failure) {
        *failure=x;
      }
      x=r->numingredients;
    }
    x++;
  }
  g_free(ingredientsneeded);
  return retval;
}

void extract_stuff(struct item_t *i, struct recipe_t *stuff)
{
  int x=0;
  int y;
  int lastsemi;
  *stuff=null_recipe;
  while (x<i->numpopstrs) {
    if (strncmp(i->popstrs[x]->str,"!T;",3)==0) {
      y=3;
      lastsemi=3;
      while (y<i->popstrs[x]->len) {
        switch (i->popstrs[x]->str[y]) {
        case ';':
          i->popstrs[x]->str[y]='\0';
          handle_recipe_line(&i->popstrs[x]->str[y],stuff);
          i->popstrs[x]->str[y]=';';
          lastsemi=y;
        }
        y++;
      }
    } else {
      dispose_recipe(stuff);
      *stuff=null_recipe;
      return;
    }
    x++;
  }
}

void put_stuff_in_item(struct item_t *i, struct recipe_t *r)
{
  GString *s=g_string_new("");
  int x=0;
  dispose_item_popstrs(i);
  while (x<r->numingredients) {
    g_string_sprintf(s,"!T;%s=%d;",r->ingredient[x]->str,r->ingredientqty[x]);
    add_popstr_to_item(i,s->str);
    x++;
  }
}

int ingredientidx(struct recipe_t *r, GString *s)
{
  int x=0;
  while (x<r->numingredients) {
    if (gstreq(r->ingredient[x],s)) {
      return x;
    }
    x++;
  }
  return -1;
}

void subtract_recipe(struct recipe_t *r, struct recipe_t *s)
{
  int	x=0;
  int	i;
  while (x<s->numingredients) {
    i=ingredientidx(r,s->ingredient[x]);
    if (i>=0) {
      r->ingredientqty[i]-=s->ingredientqty[x];
    }
    x++;
  }
}

void build_to_time(struct item_t *i, struct ship_t *s, int time)
{
  struct recipe_t recipe;
  struct recipe_t stuff;
  if (i->equipped) {
    /* Only build in equipped factories */
    recipe=null_recipe;
    load_recipe(&recipe,i->cargo->str);
    extract_stuff(i,&stuff);
    if ((i->updated+recipe.time<=time) && (i->cargoqty<i->cargomax)) {
      i->modified=1;
      while ((i->updated+recipe.time<=time) && (i->cargoqty<i->cargomax)) {
        if (ingredients_obtained(&recipe,s,&stuff,NULL)) {
	  // I think this leaves all ingredients in stuff.  Check and
	  // deduct if appropriate.
          subtract_recipe(&stuff,&recipe);
          i->cargoqty++;
        }
        i->updated+=recipe.time;
      }
    }
    put_stuff_in_item(i,&stuff);
    dispose_recipe(&stuff);
    dispose_recipe(&recipe);
  }
}

void airmine_to_time(struct item_t *i, struct fleet_t *fleet, int time)
{
  float	amtmined;
  int	diff;
  int	amtminedint;
  if (i->equipped) {
    if (fleet->locus==LOCUS_LANDED) {
      if ((fleet->body<fleet->sect->numbodies) &&
        (fleet->zone<fleet->sect->bodies[fleet->body].numzones)) {
        diff=time-i->updated;
        amtmined=(extract_amt(
          fleet->sect->bodies[fleet->body].zones[fleet->zone]->str,
          i->cargo->str)/128)*diff;
        amtminedint=floor(amtmined);
        amtmined-=amtminedint;
        if (amtmined>0) {
          amtminedint+=probability(amtmined);
        }
        i->cargoqty+=amtminedint;
        if (i->cargoqty>i->cargomax) {
          i->cargoqty=i->cargomax;
        }
        i->updated+=diff;
      } else {
        ui_message("Cannot airmine - zone/body not specified.");
      }
    } else {
      ui_message("Cannot airmine - fleet is not landed.\n");
    }
  }
}

int build_items_in_shipyard(struct item_t *i, struct ship_t *s,
  char *newitems, int qty, struct recipe_t *stuff)
{
  int			built=0;
  struct recipe_t	recipe;
  int			needed;
  GString		*str;
  load_recipe(&recipe,newitems);
  while (built<qty) {
    if (ingredients_obtained(&recipe,s,stuff,&needed)) {
      subtract_recipe(stuff,&recipe);
      built++;
      i->updated+=recipe.time;
    } else if (build_items_in_shipyard(i,s,recipe.ingredient[needed]->str,
      recipe.ingredientqty[needed],stuff)==0) {
      goto done;
    }
  }
  done:
  if (built) {
    str=g_string_new(newitems);
    add_ingredient_qty(stuff,str,built);
    g_string_free(str,TRUE);
  }  
  dispose_recipe(&recipe);
  return (built==qty);
}

void extract_ship_recipe(struct ship_t *s, struct recipe_t *r)
{
  int x=0;
  int i;
  *r=null_recipe;
  while (x<s->numitems) {
    i=ingredientidx_c(r,s->items[x].base);
    r->ingredientqty[i]++;
    x++;
  }
}

void build_ship_to_time(struct item_t *i, struct ship_t *ship, int time)
{
  struct ship_t		newship=null_ship;
  struct recipe_t	stuff,recipe;
  int			failure;
  if (load_ship(i->cargo->str,&newship,NULL)) {
    extract_stuff(i,&stuff);
    extract_ship_recipe(&newship,&recipe);
    while (i->updated+recipe.time<=time) {
      if (ingredients_obtained(&recipe,ship,&stuff,&failure)) {
        i->updated+=recipe.time;
        i->cargoqty++;
      } else {
        build_items_in_shipyard(i,ship,recipe.ingredient[failure]->str,
          recipe.ingredientqty[failure],&stuff);
      }
    }
    put_stuff_in_item(i,&stuff);
    dispose_recipe(&recipe);
    dispose_recipe(&stuff);
  } else {
    ui_error("Error loading ship.");
  }
  dispose_ship(&newship);
}

void breed_itempop_to_time(struct item_t *i, struct ship_t *ship,
  int time, int *damage)
{
  struct populii_t itempop;
  int sixteenth;
  extract_item_populii(i,&itempop);
  /* If habitat population is between 0xC and 0xE of max, habitat
     self-repairs by 1 HP per cycle. */
  if (i->hp<i->maxhp) {
    if (*damage) {
      sixteenth=i->maxcrew>>4;
      if ((itempop.total>=sixteenth*0xC) && (itempop.total<=sixteenth*0xE)) {
        --*damage;
        ++i->hp;
      }
    }
  }
  breed_to_time(&itempop,i->maxcrew,ship,time);
  i->updated=time;
  dispose_item_popstrs(i);
  populate_item(i,&itempop);
  dispose_populii(&itempop);
}

void age_shipyard_to_time(struct item_t *i, struct ship_t *ship, int time)
{
  int diff;
  /*  Currently, just builds the shipyard itself - does not yet build the
    *  ship it's assigned to build. */
  if (i->cyclestocompletion) {
    if (i->updated<time) {
      diff=time-i->updated;
      if (diff<=i->cyclestocompletion) {
        i->cyclestocompletion-=diff;
        i->updated+=diff;
      } else {
        i->updated+=i->cyclestocompletion;
        i->cyclestocompletion=0;
      }
    }
  } else if (i->cargo->len) {
    build_ship_to_time(i,ship,time);
    if (i->cargoqty) {
      move_cargo_to_fleet(ship->fleet,i);
    }
  }
}

void age_item_to_time(struct item_t *i, struct ship_t *ship, int time,
  int *damage)
{
  if (i->updated<time) {
    switch (i->class) {
    case ICLASS_DESTROYED:
    case 0:
    case ICLASS_WEAPON:
    case ICLASS_ARMOUR:
      break;
    case ICLASS_HABITAT:
      breed_itempop_to_time(i,ship,time,damage);
      break;
    case ICLASS_HOLD:
    case ICLASS_ORBITDRIVE:
    case ICLASS_FLEETSENSOR:
      break;
    case ICLASS_FACTORY:
      build_to_time(i,ship,time);
      break;
    case ICLASS_DRAINER:
    case ICLASS_TARGETINGCOMP:
    case ICLASS_ANTIINERTIA:
    case ICLASS_PSIWEAPON:
    case ICLASS_BODYSENSOR:
      break;
    case ICLASS_SHIPYARD:
      age_shipyard_to_time(i,ship,time);
      break;
    case ICLASS_AIRMINER:
      airmine_to_time(i,ship->fleet,time);
      break;
    default:
      ui_error_sd("Cannot age item class ",i->class);
    }
  }
}

void age_ship(struct ship_t *s)
{
  int x=0;
  int oldest=MAKROKOSMOS_NOW;
  int usingbar=0;
  int now=MAKROKOSMOS_NOW;
  int damage=s->maxhp-s->hp;
  struct ui_progbar_t bar;
  while (x<s->numitems) {
    if (s->items[x].updated) {
      if (s->items[x].updated+MAKROKOSMOS_MAX_AGING<now) {
        s->items[x].updated=now-MAKROKOSMOS_MAX_AGING;
      }
      oldest=lesser(oldest,s->items[x].updated);
    }
    damage+=(s->maxhp-s->hp);
    x++;
  }
  if ((now-oldest)>(MAKROKOSMOS_MAX_AGING/16)) {
    bar=ui_progbar_init("Aging ship...", oldest, now);
    usingbar=1;
  }
  while (oldest<now) {
    oldest++;
    x=0;
    while (x<s->numitems) {
      age_item_to_time(&s->items[x],s,oldest,&damage);
      x++;
    }
    if (usingbar) ui_progbar_advance(&bar,oldest);
  }
  if (usingbar) ui_progbar_destroy(&bar);
}

void age_fleet(struct fleet_t *f)
{
  int x=0;
  while (x<f->numships) {
    age_ship(&f->ships[x]);
    x++;
  }
}

int open_fleet_quiet(const char *name, int flags)
{
  GString *filename=g_string_new(g_Config.datapath->str);
  int fd;
  g_string_append(filename,MAKROKOSMOS_FLEET_PATH);
  g_string_append(filename,name);
  fd=open(filename->str,flags);
  if (fd>=0) {
    if (flock(fd,LOCK_EX|LOCK_NB)!=0) {
      close(fd);
      g_fleet_errno=errno;
      fd=-1;
    }
  }
  g_fleet_errno=errno;
  return fd;
}

int open_fleet(const char *name, int flags)
{
  GString *filename=g_string_new(g_Config.datapath->str);
  int fd;
  g_string_append(filename,MAKROKOSMOS_FLEET_PATH);
  g_string_append(filename,name);
  fd=open(filename->str,flags);
  if (fd>=0) {
    if (flock(fd,LOCK_EX|LOCK_NB)!=0) {
      close(fd);
      g_fleet_errno=errno;
      fd=-1;
    }
  } else {
    g_fleet_errno=errno;
    ui_error_sss("Unable to open fleet datafile '",filename->str,"'");
    refresh();
  }
  return fd;
}

void extend_ship_chain(struct fleet_t *fleet, int numships)
{
  if (numships>=fleet->numships) {
    fleet->ships =
      g_realloc(fleet->ships,sizeof(struct ship_t)*(numships+1));
    while (fleet->numships<=numships) {
      fleet->ships[fleet->numships]=null_ship;
      fleet->numships++;
    }
  }
}

void handle_ship_line(struct ship_t *s, char *line)
{
  GString	*cmd;
  GString	*args;
  int		n;
  split(line,&cmd,&args);
  if (strcmp(line,"Custom")==0) {
    s->modified=MODDED_CUSTOM;
  } else if (strcmp(cmd->str,"Immobile")==0) {
    s->immobile=1;
  } else if (strcmp(cmd->str,"Item")==0) {
    add_item_to_ship(s,args->str);
    s->curritem=s->numitems-1;
  } else if (strcmp(cmd->str,"HP")==0) {
    s->hp=atoi(args->str);
  } else if (strncmp(line,"Item.",5)==0) {
    handle_item_line(&s->items[s->curritem],&line[5]);
    /*
     *  I think item modifications are now explicitly saved.  If not, then
     *  there may be narbles.  This was commented out to stop it from
     *  clobbering custom items.
     *  s->items[s->curritem].modified=1;
     */
  } else if (strcmp(cmd->str,"ItemNum")==0) {
    n=atoi(args->str);
    while (n>=s->numitems) {
      ui_error_sd("Extending item chains to account for a population in nullspace at ",n);
      add_item_to_ship(s,"Defaults/nonexistant_hold");
      s->items[s->numitems-1].modified=1;
    }
    s->curritem=n;
  } else if (strcmp(cmd->str,"MaxHP")==0) {
    s->maxhp=atoi(args->str);
    s->hp=s->maxhp;
  } else if (strcmp(line,"Modified")==0) {
    s->modified=1;
  } else if (strcmp(cmd->str,"Surface")==0) {
    s->surface=atoi(args->str);
  } else if (strcmp(cmd->str,"Volume")==0) {
    s->volume=atoi(args->str);
  } else if (cmd->str[0]!='\0') {
    printw("Invalid ship line '%s'\n",line);
    refresh();
  }
  g_string_free(cmd,TRUE);
  g_string_free(args,TRUE);
}

int load_ship(char *shipfile, struct ship_t *s, struct fleet_t *fleet)
{
  GString	*filename=g_string_new(g_Config.datapath->str);
  FILE		*f;
  char		buf[1024];
  int		len;
  int		retval=0;
  g_string_append(filename,MAKROKOSMOS_SHIP_PATH);
  g_string_append(filename,shipfile);
  f=fopen(filename->str,"r");
  g_string_assign(s->base,shipfile);
  s->fleet=fleet;
  if (f) {
    while (fgets(&buf[0],1024,f)) {
      len=strlen(&buf[0]);
      if ((len) && (buf[len-1]=='\n')) {
        buf[len-1]='\0';
      }
      if (buf[0]!='#') {
        handle_ship_line(s,&buf[0]);
      }
    }
    fclose(f);
    retval=1;
  } else {
    printw("Error %d (%s) opening file %s\n",errno,g_strerror(errno),
      filename->str);
    refresh();
  }
  g_string_free(filename,TRUE);
  return retval;
}

void add_ship_to_fleet(struct fleet_t *fleet, char *name)
{
  extend_ship_chain(fleet,fleet->numships);
  fleet->ships[fleet->numships-1].fleet=fleet;
  load_ship(name,&fleet->ships[fleet->numships-1],fleet);
}

void read_fleet(int fd, struct fleet_t *fleet, struct sector_t *sect)
{
  int		fd2;
  FILE		*f;
  int		len;
  char		buf[1024];
  GString	*cmd;
  GString	*args;
  int		currship=-1;
  *fleet=null_fleet;
  fleet->sect=sect;
  fd2=dup(fd);
  if (fd2>=0) {
    f=fdopen(fd2,"r");
    if (f) {
      while (fgets(&buf[0],1024,f)) {
        len=strlen(&buf[0]);
        if ((len) && (buf[len-1]=='\n')) {
          buf[len-1]='\0';
        }
        if (buf[0]!='#') {
          split(&buf[0],&cmd,&args);
          if (strcmp(cmd->str,"Base")==0) {
            load_ship(args->str,&fleet->ships[currship],fleet);
          } else if (strcmp(cmd->str,"Body")==0) {
            fleet->body=atoi(args->str);
          } else if (strcmp(cmd->str,"Coords")==0) {
            g_string_assign(fleet->coords,args->str);
          } else if (strcmp(cmd->str,"Desc")==0) {
            g_string_assign_interpolated(fleet->desc,args->str);
          } else if (strcmp(cmd->str,"Immobile")==0) {
            fleet->immobile=1;
          } else if (strcmp(cmd->str,"Locus")==0) {
            fleet->locus=atoi(args->str);
          } else if (strcmp(cmd->str,"Ship")==0) {
            currship=atoi(args->str);
            if (currship>=fleet->numships) {
              extend_ship_chain(fleet,currship);
            }
          } else if (strncmp(cmd->str,"Ship.",5)==0) {
            handle_ship_line(&fleet->ships[currship],&buf[5]);
            if (fleet->ships[currship].modified==0) {
              ui_error("Warning: Unmodified ship was modified.");
              fleet->ships[currship].modified=1;
            }
          } else if (strcmp(cmd->str,"Zone")==0) {
            fleet->zone=atoi(args->str);
          } else {
            printw("Invalid fleet line %s\n",&buf[0]);
            refresh();
            shutdown_game();
          }
          g_string_free(cmd,TRUE);
          g_string_free(args,TRUE);
        }
      }
    } else {
      ui_config_error("Error in read_fleet{...fdopen()...}\n",errno);
    }
  } else {
    ui_config_error("Error in read_fleet{...dup()...}\n",errno);
  }
}

void close_fleet(int fd)
{
  if (flock(fd,LOCK_UN)!=0) {
    printw("Debug: fd=%d errno=%d\n",fd,errno);
    ui_error("Error unlocking fleet.");
  }
  close(fd);
}

void write_ship_mods(FILE *f, struct ship_t *s)
{
  int x=0;
  int y;
  if (s->modified==MODDED_CUSTOM) {
    fprintf(f,"Ship.Custom\nShip.Volume=%d\n",s->volume);
    fprintf(f,"Ship.Surface=%d\n",s->surface);
    fprintf(f,"Ship.MaxHP=%d\n",s->maxhp);
  }
  if (s->hp!=s->maxhp) {
    fprintf(f,"Ship.HP=%d\n",s->hp);
  }
  if (s->immobile) {
    fprintf(f,"Ship.Immobile\n");
  }
  while (x<s->numitems) {
    if (s->items[x].cargoqty) {
      if (s->items[x].modified==0) {
        s->items[x].modified=1;
      }
    }
    if (s->items[x].add_on) {
      fprintf(f,"Ship.Item=%s\nShip.Item.Add_On\n",s->items[x].base->str);
    } else if (s->items[x].modified) {
      fprintf(f,"Ship.ItemNum=%d\n",x);
    }
    if (s->items[x].modified) {
      if (s->items[x].modified==MODDED_CUSTOM) {
        fprintf(f,"Ship.Item.Custom\nShip.Item.MaxHP=%d\n",s->items[x].maxhp);
        fprintf(f,"Ship.Item.Volume=%d\n",s->items[x].volume);
        fprintf(f,"Ship.Item.Surface=%d\n",s->items[x].surface);
        if (s->items[x].cargo->len) {
          fprintf(f,"Ship.Item.Cargo=%s\n",s->items[x].cargo->str);
        }
        if (s->items[x].buildtech.bio) {
          fprintf(f,"Ship.Item.BuildBio=%d\n",s->items[x].buildtech.bio);
        }
        if (s->items[x].buildtech.emf) {
          fprintf(f,"Ship.Item.BuildEMF=%d\n",s->items[x].buildtech.emf);
        }
        if (s->items[x].buildtech.grav) {
          fprintf(f,"Ship.Item.BuildGrav=%d\n",
            s->items[x].buildtech.grav);
        }
        if (s->items[x].buildtech.logic) {
          fprintf(f,"Ship.Item.BuildLogic=%d\n",
            s->items[x].buildtech.logic);
        }
        if (s->items[x].buildtech.mech) {
          fprintf(f,"Ship.Item.BuildMech=%d\n",
            s->items[x].buildtech.mech);
        }
        if (s->items[x].buildtech.psi) {
          fprintf(f,"Ship.Item.BuildPsi=%d\n",s->items[x].buildtech.psi);
        }
        if (s->items[x].buildtech.quantum) {
          fprintf(f,"Ship.Item.BuildQuantum=%d\n",
            s->items[x].buildtech.quantum);
        }
      } else {
        fprintf(f,"Ship.Item.Modified\n");
      }
      if (s->items[x].class==ICLASS_DESTROYED) {
        fprintf(f,"Ship.Item.Class=%d\n",ICLASS_DESTROYED);
      }
      if (s->items[x].crew) {
        fprintf(f,"Ship.Item.Crew=%d\n",s->items[x].crew);
      }
      if (s->items[x].hp!=s->items[x].maxhp) {
        fprintf(f,"Ship.Item.HP=%d\n",s->items[x].hp);
      }
      if (s->items[x].updated) {
        fprintf(f,"Ship.Item.Updated=%d\n",s->items[x].updated);
        if (s->items[x].updated>MAKROKOSMOS_NOW) {
          ui_message("Debug: Item updated in future.");
        }
      }
      if (s->items[x].numpopstrs) {
        y=0;
        while (y<s->items[x].numpopstrs) {
          fprintf(f,"Ship.Item.Population=%s\n",s->items[x].popstrs[y]->str);
          y++;
        }
      }
      if (s->items[x].cargoqty) {
        fprintf(f,"Ship.Item.CargoQty=%d\n",s->items[x].cargoqty);
      }
      if (s->items[x].equipped!=s->items[x].defaultequipped) {
        fprintf(f,"Ship.Item.Equipped=%d\n",s->items[x].equipped);
      }
      if (s->items[x].cyclestocompletion) {
        fprintf(f,"Ship.Item.Completion=%d\n",s->items[x].cyclestocompletion);
      }
    }
    x++;
  }
}

void write_fleet(int fd, struct fleet_t *fleet)
{
  FILE *f;
  int x=0;
  int fd2;
  lseek(fd,0,SEEK_SET);
  ftruncate(fd,0);
  fd2=dup(fd);
  if (fd2>=0) {
    f=fdopen(fd2,"w");
    if (f) {
      if (fleet->body) {
        fprintf(f,"Body=%d\n",fleet->body);
      }
      if (fleet->desc->len) {
        fprintf(f,"Desc=");
        write_string_interpolated(f,fleet->desc->str);
        fprintf(f,"\n");
      }
      if (fleet->zone) {
        fprintf(f,"Zone=%d\n",fleet->zone);
      }
      if (fleet->locus) {
        fprintf(f,"Locus=%d\n",fleet->locus);
      }
      if (fleet->immobile) {
        fprintf(f,"Immobile\n");
      }
      while (x<fleet->numships) {
        fprintf(f,"Ship=%d\n",x);
        fprintf(f,"Base=%s\n",fleet->ships[x].base->str);
        if (fleet->ships[x].modified) {
          fprintf(f,"Ship.Modified\n");
          write_ship_mods(f,&fleet->ships[x]);
        }
        x++;
      }
      fclose(f);
    }
  }
}

void dispose_ship(struct ship_t *s)
{
  int x=0;
  g_string_free(s->base,TRUE);
  while (x<s->numitems) {
    dispose_item(&s->items[x++]);
  }
  g_free(s->items);
}

void dispose_fleet(struct fleet_t *f)
{
  int x=0;
  while (x<f->numships) {
    dispose_ship(&f->ships[x++]);
  }
  g_string_free(f->coords,TRUE);
  g_string_free(f->desc,TRUE);
  g_free(f->ships);
}

int update_fleet(char *fleetname, struct sector_t *sect)
{
  struct fleet_t fleet;
  int   fd;
  fd=open_fleet(fleetname,O_RDWR);
  if (fd>=0) {
    read_fleet(fd,&fleet,sect);
    age_fleet(&fleet);
    write_fleet(fd,&fleet);
    dispose_fleet(&fleet);
    close_fleet(fd);
    return 1;
  }
  return 0;
}

void equip_item(struct item_t *i)
{
  if (i->equipped==0) {
    i->equipped=1;
    if (i->equipped!=i->defaultequipped) {
      i->modified=1;
    }
    i->updated=MAKROKOSMOS_NOW;
  }
}

void equip_everything_in_fleet(char *fleetname)
{
  int fd=open_fleet(fleetname,O_RDWR);
  int x=0;
  struct fleet_t fleet;
  int y;
  if (fd>=0) {
    read_fleet(fd,&fleet,NULL);
    age_fleet(&fleet);
    while (x<fleet.numships) {
      y=0;
      while (y<fleet.ships[x].numitems) {
        equip_item(&fleet.ships[x].items[y]);
        y++;
      }
      x++;
    }
    write_fleet(fd,&fleet);
    close_fleet(fd);
  }
}

int load_fleet(char *fleetname, struct fleet_t *fleet)
{
  int retval=0;
  int fd=open_fleet(fleetname,O_RDONLY);
  if (fd>=0) {
    read_fleet(fd,fleet,NULL);
    if (fleet->numships==0) {
      ui_config_error_sss("No ships in fleet ",fleetname,".  Deleting fleet.",0);
      del_fleet_file(fleetname);
    }
    close_fleet(fd);
    retval=1;
  } else {
    *fleet=null_fleet;
  }
  return retval;
}

void move_cargo_to_fleet(struct fleet_t *f, struct item_t *i)
{
  int x=0;
  while (x<i->cargoqty) {
    add_ship_to_fleet(f,i->cargo->str);
    x++;
  }
  i->cargoqty=0;
}

int techlevel_ge(struct tech_t *t1, struct tech_t *t2)
{
  return (t1->bio>=t2->bio) && (t1->emf>=t2->emf) && (t1->grav>=t2->grav)
    && (t1->logic>=t2->logic) && (t1->mech>=t2->mech) &&
    (t1->psi>=t2->psi) && (t1->quantum>=t2->quantum);
}

void merge_itembuildtech(struct tech_t *tech, struct item_t *i)
{
  struct recipe_t	recipe;
  struct item_t		part;
  int			x=0;
  load_recipe(&recipe,i->base->str);
  while (x<recipe.numingredients) {
    part=null_item;
    load_item(&part,recipe.ingredient[x]->str,NULL);
    merge_itembuildtech(tech,&part);
    dispose_item(&part);
    x++;
  }
  tech->bio=greater(tech->bio,recipe.tech.bio);
  tech->emf=greater(tech->emf,recipe.tech.emf);
  tech->grav=greater(tech->grav,recipe.tech.grav);
  tech->logic=greater(tech->logic,recipe.tech.logic);
  tech->mech=greater(tech->mech,recipe.tech.mech);
  tech->psi=greater(tech->psi,recipe.tech.psi);
  tech->quantum=greater(tech->quantum,recipe.tech.quantum);
  dispose_recipe(&recipe);
}

void extract_ship_buildtech(struct ship_t *ship, struct tech_t *tech)
{
  int x=0;
  *tech=null_tech;
  while (x<ship->numitems) {
    merge_itembuildtech(tech,&ship->items[x]);
    x++;
  }
}

int shipyard_can_build(struct item_t *shipyard, struct ship_t *ship,
  int *err)
{
  struct tech_t	buildtech;
  int		retval;
  if (shipyard->volume<ship->volume) {
    *err=ERR_TOOBIG;
    return 0;
  }
  extract_ship_buildtech(ship,&buildtech);
  retval=techlevel_ge(&shipyard->buildtech,&buildtech);
  if (retval==0) {
    show_tech_difference(&shipyard->buildtech,&buildtech);
    *err=ERR_TOOADVANCED;
  }
  return retval;
}

int create_fleet(const char *name)
{
  GString *filename=g_string_new(g_Config.datapath->str);
  int fd;
  g_string_append(filename,MAKROKOSMOS_FLEET_PATH);
  g_string_append(filename,name);
  fd=open(filename->str,O_RDWR|O_CREAT|O_EXCL,g_Config.datamode);
  if (fd>=0) {
    if (flock(fd,LOCK_EX|LOCK_NB)!=0) {
      close(fd);
      fd=-1;
    }
    fchmod(fd, g_Config.datamode);
  } else {
    ui_error("Unable to open fleet.");
  }
  return fd;
}

void move_ship_to_fleet(struct fleet_t *src, int ship,
  struct fleet_t *dest)
{
  dest->numships++;
  dest->ships=g_realloc(dest->ships,dest->numships*sizeof(struct ship_t));
  dest->ships[dest->numships-1]=src->ships[ship];
  /*
   * Moving an immobile ship into a fleet makes the fleet immobile.
   */
  dest->immobile|=src->ships[src->numships].immobile;
  src->numships--;
  src->ships[ship]=src->ships[src->numships];
  src->ships=g_realloc(src->ships,src->numships*sizeof(struct ship_t));
}

void apply_history_to_fleetpair(struct fleet_t *dest,
  struct fleet_t *src, struct history_chain *h)
{
  int x=0;
  while (x<h->numchanges) {
    switch (h->c[x].action) {
    case CHANGE_MOVESHIP:
      if (h->c[x].srcfleet==0) {
        move_ship_to_fleet(dest,h->c[x].srcship,src);
      } else if (h->c[x].srcfleet==1) {
        move_ship_to_fleet(src,h->c[x].srcship,dest);
      } else {
        ui_error("Invalid source fleet in apply_history_to_fleetpair()");
        return;
      }
      break;
    case CHANGE_MOVEITEM:
      if (h->c[x].srcfleet==0) {
        if (move_item_to_ship(&src->ships[h->c[x].srcship], h->c[x].srcitem,
          &dest->ships[h->c[x].destship])==0) {
          ui_error("Move of item to dest failed.");
          return;
        } else {
          ui_message("move of item to dest succeeded.");
        }
      } else {
        if (move_item_to_ship(&dest->ships[h->c[x].srcship], h->c[x].srcitem,
          &src->ships[h->c[x].destship])==0) {
          ui_error("Move of item to src failed.");
          return;
        } else {
          ui_message("move of item to src succeeded.");
        }
      }
      break;
    default:
      ui_error("Unknown action in apply_history_to_fleetpair()");
      return;
    }
    x++;
  }
}

int techeq(struct tech_t *t1, struct tech_t *t2)
{
  return (t1->bio==t2->bio) && (t1->emf==t2->emf) && (t1->grav==t2->grav) &&
    (t1->logic==t2->logic) && (t1->mech==t2->mech) &&
    (t1->psi==t2->psi) && (t1->quantum==t2->quantum);
}

int itemeq(struct item_t *i1, struct item_t *i2)
{
  int x;
  if (techeq(&i1->buildtech,&i2->buildtech) && (i1->buildtime==i2->buildtime) &&
    (i1->class==i2->class) &&
    (i1->surface==i2->surface) && (i1->mass==i2->mass) &&
    (i1->volume==i2->volume) && (i1->maxcrew==i2->maxcrew) &&
    (i1->mincrew==i2->mincrew) &&
    (i1->crew==i2->crew) && (i1->maxhp==i2->maxhp) &&
    (i1->hp==i2->hp) && (i1->strength==i2->strength) && (i1->flux==i2->flux) &&
    (i1->regentime==i2->regentime) && (i1->cargomax==i2->cargomax) &&
    (i1->cargoqty==i2->cargoqty) && (i1->cargouse==i2->cargouse) &&
    (i1->updated==i2->updated) && (i1->numpopstrs==i2->numpopstrs) &&
    gstreq(i1->base,i2->base) && gstreq(i1->cargo,i2->cargo)) {
    x=0;
    while (x<i1->numpopstrs) {
      if (gstreq(i1->popstrs[x],i2->popstrs[x])==0) {
        return 0;
      }
      x++;
    }
    return 1;
  }
  return 0;
}

int shipeq(struct ship_t *s1, struct ship_t *s2)
{
  int x;
  /* Does not compare parent fleet to prevent infinite recursion */
  /* Curritem not compared as it's a scratch var */
  if ((s1->volume==s2->volume) && (s1->surface==s2->surface) &&
    (s1->maxhp==s2->maxhp) && (s1->hp==s2->hp) &&
    (s1->modified==s2->modified) && (s1->numitems==s2->numitems) &&
    (gstreq(s1->base,s2->base))) {
    x=0;
    while (x<s1->numitems) {
      if (itemeq(&s1->items[x],&s2->items[x])==0) {
        return 0;
      }
      x++;
    }
    return 1;
  }
  return 0;
}

int fleeteq(struct fleet_t *f1, struct fleet_t *f2)
{
  int x;
  if ((f1->numships==f2->numships) && (f1->sect==f2->sect) &&
    (f1->locus==f2->locus) && (f1->body==f2->body) &&
    (f1->zone==f2->zone) && gstreq(f1->coords,f2->coords) &&
    gstreq(f1->desc,f2->desc)) {
    x=0;
    while (x<f1->numships) {
      if (shipeq(&f1->ships[x],&f2->ships[x])==0) {
        return 0;
      }
      x++;
    }
    return 1;
  }
  return 0;
}

void del_fleet_file(const char *fleetname)
{
  GString *file=g_string_new(g_Config.datapath->str);
  g_string_append(file,MAKROKOSMOS_FLEET_PATH);
  g_string_append(file,fleetname);
  unlink(file->str);
  g_string_free(file,TRUE);
}

int create_fleet_from_template(const char *fleetname,
  const char *template, struct fleet_t *f)
{
  int fd;
  fd=open_fleet(template,O_RDONLY);
  if (fd>=0) {
    *f=null_fleet;
    read_fleet(fd,f,NULL);
    close_fleet(fd);
  } else {
    return -1;
  }
  fd=create_fleet(fleetname);
  if (fd>=0) {
    write_fleet(fd,f);
  } else {
    dispose_fleet(f);
    return -1;
  }
  return fd;
}

void populate_fleet_file(char *fleetname, struct populii_t *p)
{
  struct fleet_t        fleet;
  int                   fd;
  fd=open_fleet(fleetname,O_RDWR);
  if (fd>=0) {
    read_fleet(fd,&fleet,NULL);
    move_population_to_fleet(&fleet,&p->pops[0]);
    write_fleet(fd,&fleet);
    close_fleet(fd);
    dispose_fleet(&fleet);
  }
}

void itemlist_push(struct itemlist_t *l, int ship, int item)
{
  l->listlen++;
  if (l->listlen>l->alloced) {
    l->alloced+=128;
    l->list=g_realloc(l->list,l->alloced*sizeof(struct item_p));
  }
  l->list[l->listlen-1]=(struct item_p){ship: ship, item: item};
}

void dispose_itemlist(struct itemlist_t *l)
{
  g_free(l->list);
  *l=null_itemlist;
}

int check_fleet_immobility(struct fleet_t *f)
{
  int x=0;
  while (x<f->numships) {
    if (f->ships[x].immobile) {
      return 1;
    }
    x++;
  }
  return 0;
}

const char *item_class_string(int iclass)
{
  switch (iclass) {
  case ICLASS_DESTROYED:	return("Destroyed");
  case ICLASS_WEAPON:		return("Weapon");
  case ICLASS_HABITAT:		return("Habitat");
  case ICLASS_ORBITDRIVE:	return("Orbital drive");
  }
  return "Unknown item class";
}

int move_item_to_ship(struct ship_t *srcship, int srcitem,
  struct ship_t *destship)
{
  int destsurf=destship->surface-srcship->items[srcitem].surface;
  int destvol=destship->volume-srcship->items[srcitem].volume;
  int x=0;
  if (srcitem>=srcship->numitems) {
    return 0;
  }
  while (x<destship->numitems) {
    destsurf-=destship->items[x].surface;
    destvol-=destship->items[x].volume;
    x++;
  }
  if ((destsurf<0) || (destvol<0)) {
    ui_error("destsurf<0 || destvol<0");
    return 0;
  }
  destship->numitems++;
  destship->items = 
    g_realloc(destship->items,sizeof(struct item_t)*destship->numitems);
  destship->items[destship->numitems-1] = srcship->items[srcitem];
  destship->items[destship->numitems-1].add_on = 1;
  if (srcship->items[srcitem].add_on) {
    srcship->numitems--;
    srcship->items[srcitem]=srcship->items[srcship->numitems];
    srcship->items=g_realloc(srcship->items,
      sizeof(struct item_t)*srcship->numitems);
  } else {
    if (srcship->modified==0) {
      srcship->modified=1;
    }
    srcship->items[srcitem] = destroyed_item;
    srcship->items[srcitem].modified=1;
    srcship->items[srcitem].class = ICLASS_DESTROYED;
  }
  return 1;
}
