/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *   Makrokosmos (C) 2000 Deekoo L.
 *		Licensed under the GPL (v2.0)
 *
 *	Patents on the following technologies are pending:
 *
 *	Not commenting one's code for reasons of laziness and/or
 *	copy-protection.
 *
 *	Symmetrical use of angle brackets and parentheses.
 *
 *	Use of flock() to prevent data corruption.
 *
 *	Sarcasm, satire, and related humours.
 *
 *	To license the above concepts, please contact the Imperial sales
 *	offices on Mercury.  Have suitable offerings on hand.
 *
 */

#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <fcntl.h>
#include <curses.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <math.h>
#include <regex.h>
#include <crypt.h>
#include "microcosm.h"
#include "util.h"
#include "chronos.h"
#include "fleets.h"
#include "populii.h"
#include "sectors.h"
#include "vectors.h"
#include "ui_all.h"
#include "ui_curses.h"
#include "comm.h"
#include "charts.h"

GString *g_PlayerName=NULL;
struct player_t g_Player=(struct player_t) {theme: null_theme};
struct config_t g_Config;
int g_Config_Horizon=0;
int g_Config_MaxSystems=0;
int g_Config_Systems=0;
int g_Chat_fd=-1;

void get_player_id(void)
{
  g_PlayerName=read_gstring("Player ID: ");
}

void write_theme(FILE *f, struct theme_t *theme)
{
  if (theme->custom) {
    fprintf(f,"Theme.ChartBlank=%d\n",theme->chart_blank);
  }
}

void save_player(char *name, struct player_t *p)
{
  int fd;
  FILE *f;
  GString *file=g_string_new(g_Config.datapath->str);
  GString *targfile;
  g_string_append(file,MAKROKOSMOS_PLAYER_PATH);
  g_string_append(file,name);
  targfile=g_string_new(file->str);
  g_string_append(file,".new");
  fd=open(file->str,O_RDWR|O_CREAT,0640);
  if (fd>-1) {
    if (flock(fd,LOCK_EX)==0) {
      f=fdopen(fd,"w");
      if (f) {
        ftruncate(fd,0);
        fprintf(f,"Coords=%s\n",p->coords->str);
        switch (g_Player.locus) {
        case LOCUS_SECTOR:
          break;
        case LOCUS_ORBIT:
          fprintf(f,"Orbit=%d\n",p->orbit);
          break;
        case LOCUS_LANDED:
          fprintf(f,"Orbit=%d\nLanded\n",p->orbit);
          break;
        case LOCUS_DOCKED:
          fprintf(f,"Orbit=%d\nDock=%s\n",p->orbit,p->dock->str);
          break;
        case LOCUS_HYPERSPACE:
          fprintf(f,"Destination=%s\n",p->dest->str);
          break;
        default:
          ui_message_sd("Invalid locus ",p->locus);
          refresh();
        }
        if (p->accesscode->len) {
          fprintf(f,"Accesscode=%s\n",p->accesscode->str);
        }
        fprintf(f,"Fleet=%s\nTurns=%d\nTurnsUpdated=%d\n",
          p->fleet->str, p->turns, p->turnsupdated);
        fprintf(f,"ResearchUpdated=%d\n",p->researchupdated);
        if (rename(file->str,targfile->str)) {
          ui_error("An error occurred while renaming the player file.\n");
        }
        fprintf(f,"Bio=%d\nEMF=%d\nGrav=%d\nLogic=%d\nMech=%d\n",
          p->tech.bio, p->tech.emf, p->tech.grav, p->tech.logic,
          p->tech.mech);
        fprintf(f,"Psi=%d\nQuantum=%d\n", p->tech.psi, p->tech.quantum);
        fprintf(f,"Logread=%d\n",(int)p->logread);
        if (p->biopoints) {
          fprintf(f,"BioPoints=%d\n",p->biopoints);
        }
        write_theme(f,&p->theme);
/*          fprintf(f,"Theme=%x,%x,%x,%x,%x %x,%x,%x,%x,%x,%x\n",
            p->colours.text, p->colours.hotkey, p->colours.input,
            p->colours.prompt, p->colours.error,
            p->colours.chart_blank, p->colours.chart_visited,
            p->colours.chart_fullscanned, p->colours.chart_bodyscanned,
            p->colours.chart_fleetscanned, p->colours.menuitem
          );*/
        fclose(f);
      } else {
        ui_error("fdopen err");
        refresh();
      }
      flock(fd,LOCK_UN);
    } else {
      ui_error("flock() err.");
      refresh();
    }
    close(fd);
  } else {
    ui_error("open err");
    refresh();
  }
  g_string_free(targfile,TRUE);  
  g_string_free(file,TRUE);  
}

void update_turns(char *playername, struct player_t *player)
{
  int now=MAKROKOSMOS_NOW;
  int t=now-player->turnsupdated;
  if (t>0) {
    ui_message_sdsds("You haven't played in ",t," cycles.  Adding ",
      t*MAKROKOSMOS_TURNS_PER_CYCLE," turns.\n");
    refresh();
    player->turnsupdated=now;
    player->turns+=t*MAKROKOSMOS_TURNS_PER_CYCLE;
    save_player(playername,player);
  }
}

void research_emf(struct player_t *player)
{
  if (player->researchupdated+pow(player->tech.emf,2)<=MAKROKOSMOS_NOW) {
    player->researchupdated+=pow(player->tech.emf,2);
    attrset(COLOR_TEXT);
    player->tech.emf++;
    ui_message_sd("Your researchers have advanced your electromagnetic technology\nto level ",
      player->tech.emf);
  }
}

void research_bio(struct player_t *player)
{
  if (player->researchupdated+pow(player->tech.bio,2)<=MAKROKOSMOS_NOW) {
    player->researchupdated+=pow(player->tech.bio,2);
    attrset(COLOR_TEXT);
    ui_message("One of your researchers staggers up to you, blasted out of its mind\non some synthetic neurotransmitter or other, and informs you that your\nbiochemists have made some impressive discovery or other.  It hands\nyou an attractively-coloured pill and staggers off.");
    player->tech.bio++;
  }
}

void research_grav(struct player_t *player)
{
  if (player->researchupdated+pow(player->tech.grav,2)<=MAKROKOSMOS_NOW) {
    player->researchupdated+=pow(player->tech.grav,2);
    attrset(COLOR_TEXT);
    ui_message("Your laboratories report that your knowledge of gravitotemporal\nprinciples has improved.\n");
    player->tech.grav++;
  }
}

void research_logic(struct player_t *player)
{
  if (player->researchupdated+pow(player->tech.logic,2)<=MAKROKOSMOS_NOW) {
    player->researchupdated+=pow(player->tech.logic,2);
    attrset(COLOR_TEXT);
    ui_message("Several cryptic papers on transmetaboolean philosophy are published\nby your researchers.  This causes a new style of footnotes to come into\nvogue.\n");
    player->tech.logic++;
  }
}

void research_mech(struct player_t *player)
{
  if (player->researchupdated+pow(player->tech.mech,2)<=MAKROKOSMOS_NOW) {
    player->researchupdated+=pow(player->tech.mech,2);
    attrset(COLOR_TEXT);
    ui_message("Your engineers report progress.\n");
    player->tech.mech++;
  }
}

void research_psi(struct player_t *player)
{
  if (player->researchupdated+pow(player->tech.psi,2)<=MAKROKOSMOS_NOW) {
    player->researchupdated+=pow(player->tech.psi,2);
    attrset(COLOR_TEXT);
    ui_message("A vision informs you that your empire has made new developments in\npsionomagical research.\n");
    player->tech.psi++;
  }
}

void research_quantum(struct player_t *player)
{
  if (player->researchupdated+pow(player->tech.quantum,2)<=MAKROKOSMOS_NOW) {
    player->researchupdated+=pow(player->tech.quantum,2);
    attrset(COLOR_TEXT);
    ui_message("Your Empire reports an advance in quantum physics.\nAll existing textbooks are obselete; your researchers demand new ones.\n");
    player->tech.quantum++;
  }
}


void update_research(struct player_t *player)
{
  switch (randval_i(8)) {
  case 0:
    research_bio(player);
    break;
  case 1:
    research_emf(player);
    break;
  case 2:
    research_grav(player);
    break;
  case 3:
    research_logic(player);
    break;
  case 4:
    research_mech(player);
    break;
  case 5:
    research_psi(player);
    break;
  case 6:
    research_quantum(player);
    break;
  default:
    research_bio(player);
    research_emf(player);
    research_grav(player);
    research_logic(player);
    research_mech(player);
    research_psi(player);
    research_quantum(player);
  }
}

void spawn_newplayer_fleet(char *fleetname)
{
  int			fd;
  struct fleet_t	defaultfleet;
  load_fleet("Defaults/NewImperialFlagship",&defaultfleet);
  fd=create_fleet(fleetname);
  if (fd>=0) {
    write_fleet(fd,&defaultfleet);
    close_fleet(fd);
  } else {
    ui_error("Unable to open new flagship.");
  }
  dispose_fleet(&defaultfleet);
}

void mkfleetsdir(char *name)
{
  GString *s=g_string_new(g_Config.datapath->str);
  g_string_append(s,MAKROKOSMOS_FLEET_PATH);
  g_string_append(s,name);
  mkdir(s->str,g_Config.dirmode);
  chmod(s->str,g_Config.dirmode);
  g_string_free(s,TRUE);
}

void add_fleet_to_locus(char *fleet, char *coords, int locus, int body)
{
  int			fd;
  struct sector_t	sect;
  fd=open_sector(coords);
  if (fd>=0) {
    read_sector(fd,&sect);
    switch (locus) {
    case LOCUS_SECTOR:
      add_fleet(&sect,fleet);
      break;
    case LOCUS_ORBIT:
      if (body<sect.numbodies) {
        add_orbital_fleet(&sect.bodies[body],fleet);
      }
      break;
    case LOCUS_LANDED:
      if (body<sect.numbodies) {
        add_landed_fleet(&sect.bodies[body],fleet);
      }
      break;
    default:
      ui_error("Unknown locus in add_fleet_to_locus.");
    }
    write_sector(fd,&sect);
    close_sector(fd);
    dispose_sector(&sect);
  } else {
    ui_error("Error opening sector.");
  }
}

void find_suitable_fleet(struct player_t *player)
{
  int			fd;
  GString		*s=g_string_new(player->name->str);
  mkfleetsdir(player->name->str);
  g_string_append(s,G_DIR_SEPARATOR_S);
  g_string_append(s,"ImperialFlagship");
  fd=open_fleet_quiet(s->str,O_RDWR);
  if (fd>=0) {
    close_fleet(fd);
  } else {
    spawn_newplayer_fleet(s->str);
    add_fleet_to_locus(s->str,player->coords->str,player->locus,player->orbit);
  }
  g_string_assign(player->fleet,s->str);
  g_string_free(s,TRUE);
}

void split_gstring(GString *s, char c, GString ***list, int *listlen)
{
  int x=0;
  int y=0;
  *list=NULL;
  *listlen=0;
  while (x<s->len) {
    if (s->str[x]==c) {
      s->str[x]='\0';
      *listlen+=1;
      *list=g_realloc(*list,(*listlen+1)*sizeof(GString *));
      (*list)[*listlen-1]=g_string_new(&s->str[y]);
      s->str[x]=c;      
      if (x<s->len) {
        y=x+1;
      } else {
        y=x;
      }
    }
    x++;
  }
  if ((x==s->len) && (s->len) && (s->str[x-1]!=c)) {
    *listlen+=1;
    *list=g_realloc(*list,*listlen*sizeof(GString *));
    (*list)[*listlen-1]=g_string_new(&s->str[y]);
  }
}

void set_gender_name(GString *genders, int num, char *name)
{
  GString **list;//=NULL;
  int listlen;//=0;
  int x=0;
  split_gstring(genders,',',&list,&listlen);
  if (num>=listlen) {
    list=g_realloc(list,(num+1)*sizeof(GString *));
    while (num>=listlen) {
      list[listlen]=g_string_new("");
      listlen++;
    }
  }
  g_string_assign(list[num],name);
  g_string_assign(genders,"");
  while (x<listlen) {
    g_string_append(genders,list[x]->str);
    g_string_free(list[x],TRUE);
    x++;
    if (x<listlen) {
      g_string_append(genders,",");
    }
  }
  g_free(list);
}

int safe_file_name(GString *g)
{
  int x=0;
  if (g->len) {
    while (x<g->len) {
      switch (g->str[x]) {
      case ' ':
      case '0' ... '9':
      case '-':
      case 'A' ... 'Z':
      case 'a' ... 'z':
        break;
      default:
        return 0;
      }
      x++;
    }
  } else {
    return 0;
  }
  return 1;
}

int save_new_species(struct species_t *s)
{
  FILE		*f;
  int		fd;
  GString	*filename=g_string_new(g_Config.datapath->str);
  int		retval=0;
  g_string_append(filename,MAKROKOSMOS_SPECIES_PATH);
  g_string_append(filename,s->name->str);
  fd=open(filename->str,O_CREAT|O_EXCL|O_WRONLY,g_Config.datamode);
  if (fd>=0) {
    if (flock(fd,LOCK_EX)==0) {
      f=fdopen(fd,"w");
      if (f) {
        fprintf(f,"# %s - created by %s\n",s->name->str,g_PlayerName->str);
        fprintf(f,"MinTemp=%d\nMaxTemp=%d\n",s->mintemp,s->maxtemp);
        fprintf(f,"Genders=%s\n",s->genders->str);
        fprintf(f,"Sexes=%d\nReproduceNum=%d\n",s->sexes,s->reproducenum);
        fprintf(f,"ReproduceOdds=%f\nMortality=%f\n",s->reproduceodds,
          s->mortality);
        fprintf(f,"Puberty=%d\nMenopause=%d\n",s->puberty,s->menopause);
        fprintf(f,"Senescence=%d\n",s->senescence);
        fprintf(f,"HP=%d\nMinPress=%f\nMaxPress=%f\n", s->hp, s->minpress,
          s->maxpress);
        fprintf(f,"Habitat=%s\n",s->habitat->str);
        fclose(f);
        retval=1;
      }
      flock(fd,LOCK_UN);
    }
    close(fd);
  }
  return retval;
}

GString *edit_habitat(GString *habitatstr, int *points)
{
  int c;
  int notdone=1;
  GString *ret;
  struct listpane_data_t *data=habitatpane_data(habitatstr);
  struct listpane_t *pane=ui_listpane_create(data,1,0,-2,.75,draw_habitatpane_line);
  while (notdone) {
    draw_listpane(pane);
    c=ui_habitat_edit_prompt(*points);
    switch (toupper(c)) {
	case KEY_DOWN:
      listpane_down(pane);
      break;
    case KEY_NPAGE:
      listpane_pagedown(pane);
      break;
    case KEY_PPAGE:
      listpane_pageup(pane);
      break;
    case KEY_UP:
      listpane_up(pane);
      break;
    case 'Q':
      notdone=0;
      break;
    default:
      beep();
    }
  }
  ret=habitatpane_gstring(data);
  dispose_listpane(pane);
  return ret;
}

void create_new_species(struct species_t *s)
{
  int		notdone=1;
  int		c;
  GString	*str;
  int		x;
  float		f;
  int		basetemp;
  struct species_params_t params=null_species_params;
  *s=null_species;
  str=g_string_new("");
  g_string_sprintf(str,"Defaults/Template%d",
    randval_i(MAKROKOSMOS_SPECIES_TEMPLATES));
  load_species(str->str,s);
  extract_species_params(s,&params);
  params.points=randval_i(3)+4;
  params.basepressure=s->minpress+((s->maxpress-s->minpress)/2);
  basetemp=s->mintemp+((s->maxtemp-s->mintemp)/2);
  g_string_free(str,TRUE);
  s->hp=randval_i(5)+4;
  while (notdone) {
    show_species_info(s,&params);
    c=ui_species_edit_prompt(params.points);
    switch (c) {
    case '#':
      x=read_int("Number to reproduce: ");
      if ((x>s->sexes) && (s->sexes>1)) {
        ui_message("Invalid number.  Number to reproduce must be less than or equal to the number of sexes in multisexual species.");
        break;
      }
      /* Points no longer set in gender counter. */
      s->reproducenum=x;
      break;
    case '1' ... '9':
      str=read_gstring_scs("Name of gender ",c,": ");
      set_gender_name(s->genders,c-'1',str->str);
      g_string_free(str,TRUE);
      break;
    case '/':
    case '?':
      show_species_editor_help();
      break;
    case 'A':
    case 'a':
      edit_habitat(s->habitat,&params.points);
      break;
    case 'E':
    case 'e':
      x=read_int("Age of death: ");
      if (x>s->menopause) {
        s->senescence=x;
      } else {
        ui_message("Age at death must be after menopause.");
        pause_prompt();
      }
      break;
    case 'F':
    case 'f':
      x=read_int("Age at fertility onset: ");
      if (x>s->menopause) {
        ui_message("Fertility onset must occur before menopause.");
        pause_prompt();
      } else if (x>s->menopause-x) {
        ui_message("Fertility onset must be sooner than the duration of the fertile period.");
        pause_prompt();
      } else if (x<=0) {
        ui_message("Aphids are born pregnant.  However, they don't have interstellar\nempires.");
        pause_prompt();
      } else {
        s->puberty=x;
      }
      break;
    case 'h':
      if (s->hp>0) {
        s->hp--;
        params.points++;
      }
      break;
    case 'H':
      if (params.points>0) {
        params.points--;
        s->hp++;
      }
      break;
    case 'M':
    case 'm':
      x=read_int("Age at menopause: ");
      if (x>s->senescence) {
        ui_alertbox("Age at menopause must be less than age of death.");
      } else if ((x-s->puberty<s->puberty) || (x<=s->puberty)) {
        ui_alertbox("Duration of fertile period must be greater than time to fertility.");
      } else {
        s->menopause=x;
      }
      break;
    case 'N':
    case 'n':
      str=read_gstring("Species name: ");
      g_string_assign(s->name,str->str);
      g_string_free(str,TRUE);
      break;
    case 'O':
    case 'o':
      f=read_float("Odds of reproduction: ");
      if ((f<0.1) || (f>0.9)) {
        ui_alertbox("Odds of reproduction must be between 0.1 and 0.9");
      } else {
        s->reproduceodds=f;
      }
      break;
    case 'P':
      f=read_float("Baseline pressure: ");
      if (f!=0) {
        params.basepressure=f;
        s->minpress=f-(params.pressuretolerance*cbrt(f));
        if (s->minpress<0) {
          s->minpress=0;
        }
        s->maxpress=f+(params.pressuretolerance*cbrt(f));
      }
      break;
    case 'p':
      x=read_int("Pressure variation tolerance: ");
      if (x<0) {
        ui_message_sds("Leaving pressure variation at ",
          params.pressuretolerance," points.");
      } else if (x-params.pressuretolerance>params.points) {
        ui_alertbox("You don't have enough points to increase your pressure tolerance that much.");
      } else {
        params.points-=(x-params.pressuretolerance);
        params.pressuretolerance=x;
        s->minpress=params.basepressure-(params.pressuretolerance*cbrt(params.basepressure));
        if (s->minpress<0) {
          s->minpress=0;
        }
        s->maxpress=params.basepressure+(params.pressuretolerance*cbrt(params.basepressure));
      }
      break;
    case 'Q':
    case 'q':
      if (s->name->len) {
        if (safe_file_name(s->name)) {
          if (toupper(read_char("Is your species complete?"))=='Y') {
            if (save_new_species(s)) {
              g_Player.biopoints+=params.points;
              notdone=0;
              break;
            } else {
              ui_alertbox("Error saving species.");
            }
          }
        } else {
          ui_alertbox("Invalid characters in species name.  Only letters, numbers, and spaces are allowed.");
        }
      } else {
        ui_alertbox("You need to name your species.");
      }
      break;
    case 'R':
    case 'r':
      f=read_float("Mortality: ");
      if ((f>1) || (f<0.1)) {
        ui_alertbox("Mortality must be a number between 0.1 and 1.0");
      } else {
        s->mortality=f;
      }
      break;
    case 'S':
    case 's':
      s->sexes=read_int("Number of sexes: ");
      if (s->sexes>10) {
        ui_alertbox("Limit 9 sexes.");
        s->sexes=9;
      }
      break;
    case 'T':
      x=read_int("Base temperature: ");
      if (x>=0) {
        basetemp=x;
        s->maxtemp=basetemp+params.temptolerance+
          (params.temptolerance*sqrt(basetemp))+cbrt(basetemp);
        s->mintemp=basetemp-(params.temptolerance+
          (params.temptolerance*sqrt(basetemp))+cbrt(basetemp));
      }
      break;
    case 't':
      x=read_int("Temperature variation tolerance: ");
      if (x-params.temptolerance>params.points) {
        ui_alertbox("Insufficient points to extend habitable temperature range that far.");
      } else if (x>=0) {
        s->maxtemp=basetemp+x+(x*sqrt(basetemp))+cbrt(x);
        s->mintemp=basetemp-(x+(x*sqrt(basetemp))+cbrt(x));
        params.points-=x-params.temptolerance;
        params.temptolerance=x;
      }
      break;
    default:
      ui_alertbox("Invalid command.");
    }
  }
}

void g_string_append_from_file(GString *s, const char *file, int num)
{
  char *chunk=g_malloc(num+1);
  int fd=open(file,O_RDONLY);
  int numread;
  int x;
  if (fd>=0) {
    chunk[num]='\0';
    while (num>0) {
      x=s->len;
      numread=read(fd,chunk,num);
      if (numread>0) {
        chunk[numread]='\0';
        g_string_append(s,chunk);
        num-=s->len-x;
      } else {
        num=0;
      }
    }
    close(fd);
  }
  g_free(chunk);
}

void wipe_string(char *s)
{
  /* Blank strings in memory.  This isn't really a Useful Algorythm,
     as someone with a decent electron microscope can probably rip your
     DIMMs out and read 'em.  And there's no protection against the OS
     swapping to your HD, where an OK electron microscope (or a *LIGHT*
     microscope on old MFMs...) can get at it.

     OTOH, anyone using an electron microscope to cheat at Makrokosmos
     deserves to read your data.
     */
  while (s[0]!='\0') {
    s[0]='\0';
    s[0]='\252';
    s[0]='\377';
    s[0]='\125';
    s[0]='\222';
    s[0]='\111';
    s[0]='\x24';
    s[0]='\x6D';
    s[0]='\xB6';
    s[0]='\xDB';
    s++;
  }
}

void set_newplayer_accesscode(struct player_t *player)
{
  GString *ac1;
  GString *ac2;
  GString *salt=g_string_new("$1$");
  char *crypted;
  int notdone=1;
  while (notdone) {
    ac1=read_accesscode("Select a password to access your empire: ");
    ac2=read_accesscode("Confirm password: ");
    if (gstreq(ac1,ac2)) {
      notdone=0;
    } else {
      ui_message("Passwords don't match.");
    }
  }
  wipe_string(ac2->str);
  g_string_append_from_file(salt,"/dev/urandom",8);
  crypted=crypt(ac1->str,salt->str);
  wipe_string(ac1->str);
  wipe_string(salt->str);
  g_string_assign(player->accesscode,crypted);
  wipe_string(crypted);
  g_string_free(ac1,TRUE);
  g_string_free(ac2,TRUE);
  g_string_free(salt,TRUE);
  /* Should crypt()'s return value be freed? */
}

void create_new_player(char *playername, struct player_t *player)
{
  GString *fleetdir=g_string_new(g_Config.datapath->str);
  struct species_t species;
  struct populii_t populii;
  struct coord_t   c;
  int              ch;
  ui_message_sss("No player called '",playername,"' known.");
  ch=read_char("Create a new player? ");
  if (toupper(ch) != 'Y') {
    ui_message("Player creation cancelled.");
    ui_shutdown();
    exit(0);
  }
  *player=null_player;
  set_newplayer_accesscode(player);
  ui_configure(player);
  ui_init();
  create_new_species(&species);
  g_string_assign(player->name,playername);
  g_string_append(fleetdir,MAKROKOSMOS_FLEET_PATH);
  g_string_append(fleetdir,playername);
  if (mkdir(fleetdir->str,g_Config.dirmode)==0) {
    find_suitable_fleet(player);
    player->turns=MAKROKOSMOS_NEWPLAYER_TURNS;
    randomize_coords(&c);
    g_string_free(g_Player.coords,TRUE);
    g_Player.coords=coords_to_string(&c);
    save_player(playername,player);
    create_newplayer_population(&populii,&species);
    populate_fleet_file(player->fleet->str,&populii);
    dispose_populii(&populii);
  } else {
    ui_error("Unable to create new player directory.");
    exit(errno);
  }
  dispose_species(&species);
  show_file_paged("newplayer.txt");
}

void handle_theme_line(struct theme_t *theme, const char *cmd, const char *arg)
{
  if (strcmp(cmd,"ChartBlank")==0) {
    theme->chart_blank=atoi(arg);
  } else {
    ui_config_error_sscsc("Invalid theme line '",cmd,'=',arg,'\'',0);
    return;
  }
  theme->custom=1;
}

int load_player_file(char *playername, struct player_t *player, int quiet)
{
  GString	*playerfile=g_string_new(g_Config.datapath->str);
  int		retval=0;
  FILE		*f;
  char		buf[1024];
  int		len;
  GString	*cmd;
  GString	*args;
  g_string_append(playerfile,MAKROKOSMOS_PLAYER_PATH);
  g_string_append(playerfile,playername);
  f=fopen(playerfile->str,"r");
  if (f) {
    retval=1;
    *player=null_player;
    g_string_assign(player->name,playername);
    while (fgets(&buf[0],1024,f)) {
      len=strlen(&buf[0]);
      if ((len) && (buf[len-1]=='\n')) {
        buf[len-1]='\0';
      }
      if (buf[0]!='#') {
        split(&buf[0],&cmd,&args);
        if (strcmp(cmd->str,"Accesscode")==0) {
          g_string_assign(player->accesscode,args->str);
        } else if (strcmp(cmd->str,"Bio")==0) {
          player->tech.bio=atoi(args->str);
        } else if (strcmp(cmd->str,"BioPoints")==0) {
          player->biopoints=atoi(args->str);
        } else if (strcmp(cmd->str,"Coords")==0) {
          g_string_assign(player->coords,args->str);
        } else if (strcmp(cmd->str,"Destination")==0) {
          g_string_assign(player->dest,args->str);
          player->locus=LOCUS_HYPERSPACE;
        } else if (strcmp(cmd->str,"Dock")==0) {
          g_string_assign(player->dock,args->str);
          player->locus=LOCUS_DOCKED;
        } else if (strcmp(cmd->str,"EMF")==0) {
          player->tech.emf=atoi(args->str);
        } else if (strcmp(cmd->str,"Fleet")==0) {
          g_string_assign(player->fleet,args->str);
        } else if (strcmp(cmd->str,"Grav")==0) {
          player->tech.grav=atoi(args->str);
        } else if (strcmp(cmd->str,"Landed")==0) {
          player->locus=LOCUS_LANDED;
        } else if (strcmp(cmd->str,"Logic")==0) {
          player->tech.logic=atoi(args->str);
        } else if (strcmp(cmd->str,"Logread")==0) {
          player->logread=atoi(args->str);
        } else if (strcmp(cmd->str,"Mech")==0) {
          player->tech.mech=atoi(args->str);
        } else if (strcmp(cmd->str,"Orbit")==0) {
          player->locus=LOCUS_ORBIT;
          player->orbit=atoi(args->str);
        } else if (strcmp(cmd->str,"Overlords")==0) {
          g_string_assign(player->overlords,args->str);
        } else if (strcmp(cmd->str,"Psi")==0) {
          player->tech.psi=atoi(args->str);
        } else if (strcmp(cmd->str,"Quantum")==0) {
          player->tech.quantum=atoi(args->str);
        } else if (strcmp(cmd->str,"ResearchUpdated")==0) {
          player->researchupdated=atoi(args->str);
        } else if (strncmp(cmd->str, "Theme.",6)==0) {
          handle_theme_line(&player->theme,&cmd->str[6],args->str);
        } else if (strcmp(cmd->str,"Turns")==0) {
          player->turns=atoi(args->str);
        } else if (strcmp(cmd->str,"TurnsUpdated")==0) {
          player->turnsupdated=atoi(args->str);
        } else {
          ui_config_error_ss("Invalid player line ",&buf[0],0);
          retval=0;
        }
        g_string_free(cmd,TRUE);
        g_string_free(args,TRUE);
      }
    }
    fclose(f);
  } else {
    if (errno!=ENOENT || quiet==0) {
      ui_config_error_ss("Error opening player file ",playerfile->str,errno);
    }
  }
  g_string_free(playerfile,TRUE);
  return retval;
}

int acccode_match(GString *crypted, GString *plain)
{
  char *cryptedplain;
  int ret;
  if (crypted->len==0 && plain->len==0) {
    return 1;
  }
  cryptedplain=crypt(plain->str,crypted->str);
  ret=!strcmp(crypted->str,cryptedplain);
  wipe_string(cryptedplain);
//  wipe_string(crypted->str);
  wipe_string(plain->str);
//  g_free(cryptedplain);
  return ret;
}

void load_player(char *playername, struct player_t *player)
{
  int	   fd;
  GString *acccode;
  if (load_player_file(playername,player,TRUE)) {
    acccode=read_accesscode("Access code: ");
    if (!acccode_match(player->accesscode,acccode)) {
      ui_message("Access code mismatch.");
      shutdown_game();
    }
    update_turns(playername,player);
    update_research(player);
    if (player->fleet->len==0) {
      ui_message("You're not in a fleet.  Trying to find a suitable one.\n");
      find_suitable_fleet(player);
    }
    if (update_fleet(g_Player.fleet->str,NULL)==0) {
      fd=open_fleet(player->fleet->str,O_RDWR);
      if (fd>=0) {
        close_fleet(fd);
      } else {
        ui_message("Unable to open your fleet.  Finding an appropriate one");
        find_suitable_fleet(player);
      }
    }
  } else {
    if (errno==ENOENT) {
      create_new_player(playername,player);
    } else {
      exit(errno);
    }
  }
}

void read_global_config(int fd)
{
  FILE *f;
  GString *cmd,*arg;
  char	buf[1024];
  int newfd=dup(fd);
  if ((f=fdopen(newfd,"r"))) {
    while (fgets(&buf[0],1023,f)) {
      buf[1023]='\0';
      split(&buf[0],&cmd,&arg);
      if (strcmp(cmd->str,"Horizon")==0) {
        g_Config_Horizon=atoi(arg->str);
      } else if (strcmp(cmd->str,"MaxSystems")==0) {
        g_Config_MaxSystems=atoi(arg->str);
      } else if (strcmp(cmd->str,"Systems")==0) {
        g_Config_Systems=atoi(arg->str);
      } else if ((buf[0]!='#') && (buf[0]!='\0') && (buf[0]!='\n')) {
        ui_error_sss("Invalid config line \n'",&buf[0],"'");
      }
    }
    fclose(f);
  }
}

void write_global_config(int fd)
{
  int newfd;
  FILE *f;
  lseek(fd,0,SEEK_SET);
  ftruncate(fd,0);
  newfd=dup(fd);
  f=fdopen(newfd,"w");
  if (f) {
    fprintf(f,"Horizon=%d\nMaxSystems=%d\nSystems=%d\n",
      g_Config_Horizon, g_Config_MaxSystems, g_Config_Systems);
    fclose(f);
    close(newfd);
  } else {
    ui_error("Error writing global config file.");
  }
}

void process_global_config(int fd)
{
  read_global_config(fd);
  if (g_Config_Systems<g_Config_MaxSystems) {
    if (create_random_system()) {
      g_Config_Systems++;
      write_global_config(fd);
    }
  }
}

void load_global_config_file(void)
{
  GString *s=g_string_new(MAKROKOSMOS_DATA_PATH);
  int fd;
  g_Config=null_config_t;
  g_string_append(s,MAKROKOSMOS_CONFIG_FILE);
  fd=open(s->str,O_RDWR);
  if (fd>=0) {
    if (flock(fd,LOCK_EX)==0) {
      process_global_config(fd);
/*
      read_global_config(fd);
      if (g_Config_Systems<g_Config_MaxSystems) {
        if (create_random_system()) {
          g_Config_Systems++;
          write_global_config(fd);
        }
      }
*/
      flock(fd,LOCK_UN);
    } else {
      ui_config_error("Cannot lock global config file.\n",errno);
    }
    close(fd);
  } else {
    struct passwd *pwd=getpwuid(getuid());
    int		loaded=0;
    if (pwd) {
      g_string_assign(g_Config.datapath, pwd->pw_dir);
      g_Config.datamode=0600;
      g_Config.dirmode=0700;
      g_string_append_c(g_Config.datapath, '/');
      g_string_append(g_Config.datapath, "makrokosmos/");
      g_string_assign(s, g_Config.datapath->str);
      g_string_append(s, MAKROKOSMOS_CONFIG_FILE);
      fd=open(s->str,O_RDWR);
      if (fd>=0) {
        if (flock(fd,LOCK_EX)==0) {
          process_global_config(fd);
          flock(fd,LOCK_UN);
          loaded=1;
        } else {
          ui_config_error("Error locking config file.\n", errno);
        }
        close(fd);
      } else {
        ui_config_error_ss("Error opening ", s->str, errno);
      }
    } else {
      ui_config_error("cannot getpwuid().\n", errno);
    }
    if (!loaded) {
      ui_shutdown();
      exit(1);
    }
  }
  g_string_free(s,TRUE);
}

void initialize(void)
{
  srandom(time(NULL));
  ui_init();
  load_global_config_file();
  get_player_id();
  if (g_PlayerName->len==0) {
    ui_message("Unable to determine player name.");
    ui_shutdown();
    exit(0);
  }
  load_player(g_PlayerName->str,&g_Player);
  open_chart_file();
  open_chat_file();
  read_mail(g_PlayerName->str);
}

void list_bodies_orbiting(struct sector_t *sect, int primary)
{
  int x=0;
  while (x<sect->numbodies) {
    if (sect->bodies[x].primary==primary) {
      show_single_orbiter(x+1, &sect->bodies[x]);
/*      printw("%d: %s",x+1,body_name(&sect->bodies[x]));
      if (sect->bodies[x].owner->len) {
        printw("   (Claimed by %s)\n",sect->bodies[x].owner->str);
      } else {
        printw("\n");
      }*/
    }
    x++;
  }
}

void list_fleets(struct sector_t *sect)
{
  int x=0;
  while (x<sect->numfleets) {
    printw("%s\n",sect->fleets[x]->str);
    x++;
  }
}

void list_fleets_numbered(GString **list, int num)
{
  int x=1;
  while (x<=num) {
    printw("%d: %s\n",x,list[x-1]->str);
    x++;
  }
}

void show_curr_location(void)
{
  struct sector_t sect;
  attrset(COLOR_TEXT);
  load_sector(g_Player.coords->str,&sect);
  switch (g_Player.locus) {
  case LOCUS_SECTOR:
    printw("Sector %s",sect.name->str);
    if (sect.owner->len) {
      printw("(Claimed by %s) ",sect.owner->str);
    }
    printw("\nCoords: %s\n",g_Player.coords->str);
    if (sect.numbodies) {
      printw("Bodies:\n");
      list_bodies_orbiting(&sect,-1);
    }
    if (sect.numfleets) {
      printw("Fleets:\n");
      list_fleets(&sect);
    }
    break;
  case LOCUS_ORBIT:
    if (g_Player.orbit<sect.numbodies) {
      show_location_orbit(&sect.bodies[g_Player.orbit],g_Player.coords->str);
    } else {
      ui_message("The body you were orbiting isn't there anymore.");
      g_Player.locus=LOCUS_SECTOR;
    }
    break;
  case LOCUS_LANDED:
    if (g_Player.orbit<sect.numbodies) {
      show_location_landed(&sect.bodies[g_Player.orbit]);
    } else {
      g_Player.locus=LOCUS_SECTOR;
      ui_message("The body you were landed on has disappeared.");
    }
    break;
  case LOCUS_DOCKED:
    show_location_docked(&sect);
    break;
  case LOCUS_HYPERSPACE:
    printw("In route from %s to %s.\n",g_Player.coords->str,
       g_Player.dest->str);
    printw("ETA %d turns.\n",0-g_Player.turns);
    refresh();
    break;
  default:
    wprintw(stdscr,"Invalid locus type %d\n",g_Player.locus);
    refresh();
    exit(255);
  }
  dispose_sector(&sect);
  refresh();
}

void shutdown_game(void)
{
  save_player(g_PlayerName->str,&g_Player);
  close_chart_file();
  printw("Shutting down.\n");
  refresh();
  ui_shutdown();
  exit(0);
}

/*
void list_satellites(struct sector_t *sect, int body, int **list, int *num)
{
  int x=0;
  int lines=0;
  *list=g_malloc(sizeof(int));
  *num=1;
  if (body>=0) {
    if (sect->bodies[body].primary>-1) {
      *list[0]=sect->bodies[body].primary;
      printw("0: %s\n",body_name(&sect->bodies[*list[0]]));
    } else {
      printw("0: Leave orbit\n");
      *list[0]=-1;
    }
  }
  while (x<sect->numbodies) {
    if (sect->bodies[x].primary==body) {
      if (lines>20) {
        printw("[Mo[i]re]\n");
        lines=0;
        refresh();
        getch();
      }
      printw("%d: %s\n",*num,body_name(&sect->bodies[x]));
      lines++;
      ++*num;
      *list=g_realloc(*list,*num*sizeof(int));
      (*list)[*num-1]=x;
    }
    x++;
  }
  refresh();
}*/

/*int gstreq(GString *s1, GString *s2)
{
  if (s1==s2) {
     return TRUE;
  }
  if (s1->len==s2->len) {
    return (strcmp(s1->str,s2->str)==0);
  }
  return FALSE;
}*/

void del_orbital_fleet(struct body_t *body, GString *fleet)
{
  int x=0;
  int delled=0;
  while (x<body->numofleets) {
    if (gstreq(body->ofleets[x],fleet)) {
      if (delled) {
        printw("Warning: Multiple copies of ship %s.  %d already moved.\n",
          fleet->str,delled);
      }
      g_string_free(body->ofleets[x],TRUE);
      body->numofleets--;
      body->ofleets[x]=body->ofleets[body->numofleets];
      x--;
      body->ofleets=
        g_realloc(body->ofleets,sizeof(GString *)*body->numofleets);
      delled++;
    }
    x++;
  }
}

void del_landed_fleet(struct body_t *body, GString *fleet)
{
  int x=0;
  int delled=0;
  while (x<body->numlfleets) {
    if (gstreq(body->lfleets[x],fleet)) {
      if (delled) {
        printw("Warning: Multiple copies of ship %s.  %d already moved.\n",
          fleet->str,delled);
      }
      g_string_free(body->lfleets[x],TRUE);
      body->numlfleets--;
      body->lfleets[x]=body->lfleets[body->numlfleets];
      x--;
      body->lfleets=
        g_realloc(body->lfleets,sizeof(GString *)*body->numlfleets);
      delled++;
    }
    x++;
  }
}

void del_sector_fleet(struct sector_t *sect, GString *fleet)
{
  int x=0;
  int delled=0;
  while (x<sect->numfleets) {
    if (gstreq(sect->fleets[x],fleet)) {
      if (delled) {
        printw("Warning: Multiple copies of ship %s.  %d already moved.\n",
          fleet->str,delled);
      }
      g_string_free(sect->fleets[x],TRUE);
      sect->numfleets--;
      sect->fleets[x]=sect->fleets[sect->numfleets];
      x--;
      sect->fleets=
        g_realloc(sect->fleets,sizeof(GString *)*sect->numfleets);
      delled++;
    }
    x++;
  }
}

void move_ofleet_to_orbit(char *coords, int currsat,
  GString *fleet, int targsat)
{
  struct sector_t sect;
  int fd=open_sector(coords);
  if (fd>-1) {
    read_sector(fd,&sect);
    del_orbital_fleet(&sect.bodies[currsat],fleet);
    add_orbital_fleet(&sect.bodies[targsat],fleet->str);
    write_sector(fd,&sect);
    dispose_sector(&sect);
    close_sector(fd);
  }
}

void move_sfleet_to_orbit(char *coords, GString *fleet, int targsat)
{
  struct sector_t sect;
  int fd=open_sector(coords);
  if (fd>-1) {
    read_sector(fd,&sect);
    del_sector_fleet(&sect,fleet);
    add_orbital_fleet(&sect.bodies[targsat],fleet->str);
    write_sector(fd,&sect);
    dispose_sector(&sect);
    close_sector(fd);
  }
}

void move_ofleet_to_sector(char *coords, int currsat,
  GString *fleet)
{
  struct sector_t sect;
  int fd=open_sector(coords);
  if (fd>-1) {
    read_sector(fd,&sect);
    del_orbital_fleet(&sect.bodies[currsat],fleet);
    add_fleet(&sect,fleet->str);
    write_sector(fd,&sect);
    dispose_sector(&sect);
    close_sector(fd);
  }
}

void move_ofleet_to_land(char *coords, GString *fleet, int body)
{
  struct sector_t sect;
  int fd=open_sector(coords);
  if (fd>-1) {
    read_sector(fd,&sect);
    del_orbital_fleet(&sect.bodies[body],fleet);
    add_landed_fleet(&sect.bodies[body],fleet->str);
    write_sector(fd,&sect);
    dispose_sector(&sect);
    close_sector(fd);
  }
}

void move_lfleet_to_orbit(char *coords, GString *fleet, int body)
{
  struct sector_t sect;
  int fd=open_sector(coords);
  if (fd>-1) {
    read_sector(fd,&sect);
    del_landed_fleet(&sect.bodies[body],fleet);
    add_orbital_fleet(&sect.bodies[body],fleet->str);
    write_sector(fd,&sect);
    dispose_sector(&sect);
    close_sector(fd);
  }
}

int acquire_target(struct fleet_t *f)
{
  if (f->numships) {
    return 0;
  }
  return -1;
}

int surface_item(struct ship_t *s, int point)
{
  int x=0;
  int area=0;
  while (x<s->numitems) {
    area+=s->items[x].surface;
    if (area>=point) {
      return x;
    }
    x++;
  }
  return -1;
}

void del_item(struct ship_t *s, int itemnum) {
  dispose_item(&s->items[itemnum]);
  s->numitems--;
  s->items[itemnum]=s->items[s->numitems];
  s->items=g_realloc(s->items,sizeof(struct item_t)*s->numitems);
}

void del_ship(struct fleet_t *f, int ship)
{
  dispose_ship(&f->ships[ship]);
  f->numships--;
  f->ships[ship]=f->ships[f->numships];
  f->ships=g_realloc(f->ships,sizeof(struct ship_t)*f->numships);
}

void fire(struct item_t *i, struct ship_t *s, char *report)
{
  int damage= (i->strength-i->flux) +
    (1+(int) ((i->flux*2)*(random()/(RAND_MAX+1.0))));
  int impactpoint=1+(s->surface*(random()/(RAND_MAX+1.0)));
  int itemhit=surface_item(s,impactpoint);
  if (itemhit>=0) {
    switch (s->items[itemhit].class) {
    case ICLASS_ARMOUR:
      if (s->items[itemhit].strength>=damage) {
        return;
      }
      damage-=s->items[itemhit].strength;
    }
    s->items[itemhit].modified=1;
    s->items[itemhit].hp-=damage;
    if (s->items[itemhit].hp<0) {
      s->items[itemhit].crew=0;
      s->items[itemhit].equipped=0;
      if (0-s->items[itemhit].hp>s->items[itemhit].hp/10) {
        s->items[itemhit].class=ICLASS_DESTROYED;
        if (report) {
          printw("%s %s's %s has been destroyed.\n",report,
            s->base->str,s->items[itemhit].base->str);
          refresh();
        }
        if (s->items[itemhit].add_on) {
          del_item(s,itemhit);
        } else {
          if (s->items[itemhit].popstrs) {
            dispose_item_popstrs(&s->items[itemhit]);
          }
        }
      } else if (report) {
        printw("%s %s's %s has been disabled.\n",report,
          s->base->str,s->items[itemhit].base->str);
        refresh();
      }
    }
    s->modified=1;
  } else {
    s->modified=1;
    s->hp-=damage;
  }
  refresh();
}

void do_melee_halfround(struct fleet_t* agg, struct fleet_t *def,
  struct fleet_t *junk, char *report)
{
  int x=0;
  int y;
  int target;
  while (x<agg->numships) {
    target=acquire_target(def);
    if (target>=0) {
      y=0;
      while (y<agg->ships[x].numitems) {
        switch (agg->ships[x].items[y].class) {
        case ICLASS_DESTROYED:
	case 0:
	  break;
        case ICLASS_WEAPON:
          if (target>=0) {
            fire(&agg->ships[x].items[y],&def->ships[target],report);
            if (def->ships[target].hp<0) {
              if (0-def->ships[target].hp>def->ships[target].maxhp/10) {
                if (report) {
                  printw("%s %s has been destroyed.\n",report,
                    def->ships[target].base->str);
                  refresh();
                }
                del_ship(def,target);
                target=-1;
                refresh();
              } else {
                if (report) {
                  printw("%s %s has been disabled.\n",report,
                    def->ships[target].base->str);
                  refresh();
                } /* report */
                if (def->ships[target].modified==0) {
                  def->ships[target].modified=1;
                }
                def->ships[target].immobile=1;
                move_ship_to_fleet(def,target,junk);
                target=-1;
              } /* else disabled */
            } /* hp<0 */
          } /* target>=0 */
          break;
        case ICLASS_ARMOUR:
        case ICLASS_HABITAT: /* Should provide crew to damaged components. */
        case ICLASS_HOLD:
          break;
        case ICLASS_ORBITDRIVE:
          /* Add code for approach when targeting bad, at such point
             as battlespace simulation in relevant.  */
          break;
        case ICLASS_FLEETSENSOR:	/* Does nothing in battle */
        case ICLASS_FACTORY:
	  break;
        case ICLASS_DRAINER:
          ui_error("Life drainer not yet supported.");
          break;
        case ICLASS_TARGETINGCOMP:
	  target=acquire_target(def);
          if (target<0) {
            y=agg->ships[x].numitems;
          }
          break;
	case ICLASS_ANTIINERTIA:
	  /* To be used in nav */
	  break;
        case ICLASS_PSIWEAPON:
          ui_error("Psionomagical weapons not yet supported.");
          break;
        default:
          printw("Unsupported item class %d.\n",agg->ships[x].items[y].class);
        }
        y++;
      }
    }/* else {
      ui_message("No targets in range.\n");
    }*/
    x++;
  }
}

char *junk_string(int seed)
{
  switch (seed && 0x3) {
  case 0: return "Debris/";
  case 1: return "Detritus/";
  case 2: return "Wreckage/";
  case 3: return "Junk/";
  }
  return "Cantor dust/";
}

char locuschar(int locus)
{
  switch (locus) {
  case LOCUS_SECTOR: return('s');
  case LOCUS_ORBIT: return('o');
  case LOCUS_LANDED: return('l');
  default: return('X');
  }
}

void absorb_fleet(struct fleet_t *dest, struct fleet_t *src)
{
  int x=0;
  dest->ships=g_realloc(
    dest->ships,(dest->numships+src->numships)*sizeof(struct ship_t));
  while (x<src->numships) {
    dest->ships[dest->numships++]=src->ships[x++];
  }
  g_free(src->ships);
  src->ships=NULL;
  src->numships=0;
}

void write_junk(struct fleet_t *junk, char *coords, int body, int locus)
{
  int fd;
  GString *coords2=g_string_new(coords);
  GString *fleetname=g_string_new("");
  struct sector_t s;
  struct fleet_t oldjunk;
  int x=0;
  while (x<coords2->len) {
    if (coords2->str[x]=='/') {
      coords2->str[x]='-';
    }
    x++;
  }
  if (body>=0) {
    g_string_sprintf(fleetname,"%s%s.%d%c",junk_string(body+locus),
      coords2->str,body,locuschar(locus));
  } else {
    g_string_sprintf(fleetname,"%s%s.%c",junk_string(body+locus),
      coords2->str,locuschar(locus));
  }
  fd=open_fleet_quiet(fleetname->str,O_RDWR);
  if (fd>=0) {
    read_fleet(fd,&oldjunk,NULL);
    absorb_fleet(&oldjunk,junk);
    write_fleet(fd,&oldjunk);
    close_fleet(fd);
    dispose_fleet(&oldjunk);
  } else {
    /* No preexisting junk */
    fd=create_fleet(fleetname->str);
    if (fd>=0) {
      write_fleet(fd,junk);
      close_fleet(fd);
      fd=open_sector(coords);
      if (fd>=0) {
        read_sector(fd,&s);
        switch (locus) {
        case LOCUS_SECTOR:
          add_fleet(&s,fleetname->str);
          break;
        case LOCUS_ORBIT:
          add_orbital_fleet(&s.bodies[body],fleetname->str);
          break;
        case LOCUS_LANDED:
          add_landed_fleet(&s.bodies[body],fleetname->str);
          break;
        default:
          add_fleet(&s,fleetname->str);
          ui_error("Unknown locus adding junk to sector.");
        }
        write_sector(fd,&s);
        dispose_sector(&s);
        close_sector(fd);
      } else {
        ui_error("Error adding junk fleet to host sector.");
      }
    } else {
      ui_error("Error creating junk fleet.");
    }
  }
  g_string_free(fleetname,TRUE);
  g_string_free(coords2,TRUE);
}

void del_fleet_from_sector(GString *fleet, struct sector_t *sect)
{
  int x=0;
  int y;
  while (x<sect->numfleets) {
    if (gstreq(fleet,sect->fleets[x])) {
      sect->numfleets--;
      g_string_free(sect->fleets[x],TRUE);
      sect->fleets[x]=sect->fleets[sect->numfleets];
      sect->fleets=g_realloc(sect->fleets,sect->numfleets*sizeof(GString *));
      return;
    }
    x++;
  }
  x=0;
  while (x<sect->numbodies) {
    y=0;
    while (y<sect->bodies[x].numofleets) {
      if (gstreq(fleet,sect->bodies[x].ofleets[y])) {
        sect->bodies[x].numofleets--;
        g_string_free(sect->bodies[x].ofleets[y],TRUE);
        sect->bodies[x].ofleets[y]=
          sect->bodies[x].ofleets[sect->bodies[x].numofleets];
        sect->bodies[x].ofleets=g_realloc(sect->bodies[x].ofleets,
          sect->bodies[x].numofleets*sizeof(GString *));
        return;
      }
      y++;
    }
    y=0;
    while (y<sect->bodies[x].numlfleets) {
      if (gstreq(fleet,sect->bodies[x].lfleets[y])) {
        sect->bodies[x].numlfleets--;
        g_string_free(sect->bodies[x].lfleets[y],TRUE);
        sect->bodies[x].lfleets[y]=
          sect->bodies[x].lfleets[sect->bodies[x].numlfleets];
        sect->bodies[x].lfleets=g_realloc(sect->bodies[x].lfleets,
          sect->bodies[x].numlfleets*sizeof(GString *));
        return;
      }
      y++;
    }
    x++;
  }
  printw("Unable to find and delete obliterated fleet %s\n",fleet->str);
  refresh();
}

void del_fleet(char *coords, GString *fleetname, int *ffd)
{
  int sfd=open_sector(coords);
  struct sector_t sect;
  del_fleet_file(fleetname->str);
  if (*ffd>=0) {
    ftruncate(*ffd,0);
    close_fleet(*ffd);
    *ffd=-1;
  }
  if (sfd>=0) {
    read_sector(sfd,&sect);
    del_fleet_from_sector(fleetname,&sect);
    write_sector(sfd,&sect);
    close_sector(sfd);
    dispose_sector(&sect);
  }
}

int probability(double d) {
  if (((double)random())/RAND_MAX>d) {
    return FALSE;
  }
  return TRUE;
}

int lesser(int x, int y)
{
  if (x<y) {
    return x;
  }
  return y;
}

/*int randval(int max)
{
  return ((double)random()/RAND_MAX)*max;
}*/

int greater(int x, int y)
{
  if (x>y) {
    return x;
  }
  return y;
}

double greater_d(double x, double y)
{
  if (x>y) {
    return x;
  } else {
    return y;
  }
}

struct fleet_t *do_attack(char *coords, GString *aggressor,
  GString *defender, struct sector_t *sect, int *aggressor_defeated)
	/* Returns null if no junk, otherwise returns the junk fleet */
{
  struct fleet_t agg;
  struct fleet_t def;
  struct fleet_t *junk=g_malloc(sizeof(struct fleet_t));
  int	aggfd;
  int	deffd;
  *aggressor_defeated=0;
  *junk=null_fleet;
  printw("Your fleet %s moves to attack %s.\n",
    aggressor->str, defender->str);
  refresh();
  aggfd=open_fleet(aggressor->str,O_RDWR);
  if (aggfd>=0) {
    deffd=open_fleet(defender->str,O_RDWR);
    if (deffd>=0) {
      read_fleet(aggfd,&agg,sect);
      age_fleet(&agg);
      read_fleet(deffd,&def,sect);
      age_fleet(&def);
      do_melee_halfround(&agg,&def,junk,NULL);
      do_melee_halfround(&def,&agg,junk,"Your");
      if (agg.numships) {
        write_fleet(aggfd,&agg);
      } else {
        ui_message("The enemy has defeated you.");
        del_fleet(coords,aggressor,&aggfd);
        *aggressor_defeated=1;
      }
      if (def.numships) {
        write_fleet(deffd,&def);
        close_fleet(deffd);
      } else {
        printw("You have destroyed the enemy fleet %s\n",
           defender->str);
        refresh();
        del_fleet(coords,defender,&deffd);
      }
      dispose_fleet(&def);
      dispose_fleet(&agg);
    } else {
      if (g_fleet_errno==ENOENT) {
        ui_message("The enemy fleet does not appear to be here.");
        del_fleet(coords,defender,&deffd);
      } else {
        ui_message("The enemy fleet is manouvering.  You cannot lock on.");
      }
    }
    if (aggfd>=0) { /* Won't need to close fleet if del_fleet got it. */
      close_fleet(aggfd);
    }
  } else {
    ui_message("Your fleet is unable to lock onto the enemy.");
  }
  if (junk->numships) {
    return junk;
  } else {
    dispose_fleet(junk);
    g_free(junk);
    return NULL;
  }
}

void attack_fleet(void)
{
  struct sector_t sect;
  struct fleet_t *junk=NULL;
  int targfleet;
  int destroyed=0;
  load_sector(g_Player.coords->str,&sect);
  switch (g_Player.locus) {
  case LOCUS_SECTOR:
    list_fleets_numbered(sect.fleets,sect.numfleets);
    targfleet=read_int("Fleet to attack: ");
    if ((targfleet>0) && (targfleet<=sect.numfleets)) {
      targfleet--;
      junk=do_attack(g_Player.coords->str,g_Player.fleet,
        sect.fleets[targfleet],&sect,&destroyed);
    } else {
      ui_message("Invalid input.");
    }
    break;
  case LOCUS_ORBIT:
    list_fleets_numbered(sect.bodies[g_Player.orbit].ofleets,
      sect.bodies[g_Player.orbit].numofleets);
    targfleet=read_int("Fleet to attack: ");
    if ((targfleet>0) &&
      (targfleet<=sect.bodies[g_Player.orbit].numofleets)) {
      targfleet--;
      junk=do_attack(g_Player.coords->str,g_Player.fleet,
        sect.bodies[g_Player.orbit].ofleets[targfleet],&sect,&destroyed);
    } else {
      ui_message("Invalid input.");
    }
    break;
  case LOCUS_LANDED:
    list_fleets_numbered(sect.bodies[g_Player.orbit].lfleets,
      sect.bodies[g_Player.orbit].numlfleets);
    targfleet=read_int("Fleet to attack: ");
    if ((targfleet>0) &&
      (targfleet<=sect.bodies[g_Player.orbit].numlfleets)) {
      targfleet--;
      junk=do_attack(g_Player.coords->str,g_Player.fleet,
        sect.bodies[g_Player.orbit].lfleets[targfleet],&sect,&destroyed);
    } else {
      ui_message("Invalid input.");
    }
    break;
  case LOCUS_DOCKED:
    ui_message("You decide against wrecking your spaceport and destroying your allies.");
    break;
  default:
    ui_error("Unknown locus; cannot attack fleets.");
  }
  if (junk) {
    write_junk(junk,g_Player.coords->str,g_Player.orbit,g_Player.locus);
    dispose_fleet(junk);
    g_free(junk);
  }
  dispose_sector(&sect);
  if (destroyed) {
    g_Player.turns-=MAKROKOSMOS_TURNS_PER_CYCLE*3;
    shutdown_game();
  }
}

void show_help()
{
  switch (g_Player.locus) {
  case LOCUS_SECTOR:
    show_help_sector();
    break;
  case LOCUS_ORBIT:
    show_help_orbit();
    break;
  case LOCUS_LANDED:
    show_help_landed();
    break;
  case LOCUS_DOCKED:
    show_help_docked();
    break;
  case LOCUS_HYPERSPACE:
    printw("Commands available in hyperspace\n");
    printw("?: Help             L: Leave hyperspace\n");
    printw("                    r: repair (not yot working)\n");
    printw("                    T: Time\n");
    printw("E: Equip            :: Hyperwave Commlink\n");
    printw("F: Fleet menu       Q:     Quit / quit to main\n");
    break;
  default: 
    printw("Unknown locus - no specific help.  Generic help:\n");
    printw("?: Help             L:     Land / Leave\n");
    printw("A: Attack           O:     Orbit\n");
    printw("B: Build / Board    ^O, 0: Claim ownership\n");
    printw("C: Colonize         R:     Repair\n");
    printw("D: Dock             T:     Time\n");
    printw("E: Equip            ::     Hyperwave Commlink\n");
    printw("F: Fleet menu       Q:     Quit / quit to main\n");
    break;
  }
}

void set_sector_ownership(char *sectorname, char *claimant)
{
  struct sector_t sect;
  int fd=open_sector(sectorname);
  if (fd>=0) {
    read_sector(fd,&sect);
    g_string_assign(sect.owner,claimant);
    write_sector(fd,&sect);
    dispose_sector(&sect);
    close_sector(fd);
  }
}

void set_body_ownership(char *coords, int body, char *claimant)
{
  struct sector_t sect;
  int fd=open_sector(coords);
  if (fd>=0) {
    read_sector(fd,&sect);
    if (body<sect.numbodies) {
      g_string_assign(sect.bodies[body].owner,claimant);
    } else {
      ui_error("Tried to claim invalid body.\n");
    }
    write_sector(fd,&sect);
    dispose_sector(&sect);
    close_sector(fd);
  }
}

void claim_ownership(void)
{
  struct sector_t sect;
  int fd;
  int notdone=1;
  fd=open_sector(g_Player.coords->str);
  if (fd>=0) {
    switch (g_Player.locus) {
    case LOCUS_SECTOR:
      read_sector(fd,&sect);
      if (sect.owner->len) {
        close_sector(fd);
        if (strcmp(sect.owner->str,g_PlayerName->str)) {
          printw("Sector %s is already claimed by %s.\nAre you sure you want to claim it?",
            g_Player.coords->str, sect.owner->str);
          while (notdone) {
            switch (toupper(getch())) {
            case 'Y':
              set_sector_ownership(g_Player.coords->str,g_PlayerName->str);
            case 'N':
              printw("\n");
              notdone=0;
            }
          }
        } else {
          ui_message("You already claim this sector.");
        }
      } else {
        g_string_assign(sect.owner,g_PlayerName->str);
        write_sector(fd,&sect);
        close_sector(fd);
      }
      dispose_sector(&sect);
      break;
    case LOCUS_ORBIT:
      ui_message("You must land on a body to claim ownership of it.");
      close_sector(fd);
      break;
    case LOCUS_LANDED:
      read_sector(fd,&sect);
      if (sect.bodies[g_Player.orbit].owner->len) {
        close_sector(fd);
        printw("Body %s is already claimed by %s.\nAre you sure you want to claim it?",
          sect.bodies[g_Player.orbit].name->str,
          sect.bodies[g_Player.orbit].owner->str);
        while (notdone) {
          switch (toupper(getch())) {
          case 'Y':
            set_body_ownership(g_Player.coords->str,g_Player.orbit,
              g_PlayerName->str);
          case 'N':
            printw("\n");
            notdone=0;
          }
        }
      } else {
        g_string_assign(sect.bodies[g_Player.orbit].owner,g_PlayerName->str);
        write_sector(fd,&sect);
        close_sector(fd);
      }
      dispose_sector(&sect);
      break;
    default:
      close_sector(fd);
      ui_message("Unknown locus in claim_ownership.");
    }
  }
}

void colonize_entire(char *srcfleetname, char *colonyname, int body,
  int zone, const char *colonytemplate,
  const struct gstrlist_t *specie)
{
  int colfd,srcfd,secfd;
  struct fleet_t col;
  struct fleet_t src;
  struct sector_t sect;
  struct populii_t colpop;
  int x=0,y;
  srcfd=open_fleet(srcfleetname,O_RDWR);
  if (srcfd>=0) {
    read_fleet(srcfd,&src,NULL);
    age_fleet(&src);
  } else {
    printw("Cannot open source fleet %s.\n",src);
    refresh();
    return;
  }
  secfd=open_sector(g_Player.coords->str);
  if (secfd>=0) {
    read_sector(secfd,&sect);
  } else {
    ui_message("Error opening sector file.\n");
    close_fleet(srcfd);
    dispose_fleet(&src);
    return;
  }
  colfd=create_fleet_from_template(colonyname,colonytemplate,&col);
  if ((col.numships<=0) || (col.ships[0].numitems<=0) ||
    (col.ships[0].items[0].class!=ICLASS_HABITAT)) {
    ui_message("Improperly formatted colony template.");
  }
  age_fleet(&src);
  if (colfd>=0) {
    col.zone=zone;
    col.body=body;
    col.locus=LOCUS_LANDED;
    colpop=null_populii;
    while (x<src.numships) {
      y=0;
      while (y<src.ships[x].numitems) {
        if (src.ships[x].items[y].popstrs) {
          absorb_item_populii_specie(&colpop, &src.ships[x].items[y], specie,
            col.ships[0].items[0].maxcrew);
        }
        y++;
      }
      x++;
    }
    populate_item(&col.ships[0].items[0],&colpop);
    dispose_populii(&colpop);
    if (col.ships[0].items[0].popstrs) {
      add_colony(&sect.bodies[g_Player.orbit],colonyname);
      col.ships[0].modified=1;
      col.ships[0].items[0].modified=1;
      write_fleet(colfd,&col);
      write_fleet(srcfd,&src);
      write_sector(secfd,&sect);
    } else {
      ui_message("No non-crew population in fleet.  Cannot plant colony.");
      del_fleet_file(colonyname);
    }
    close_fleet(colfd);
    dispose_fleet(&col);
  } else {
    ui_message("colonize_entire() could not create colony.");
  }
  close_fleet(srcfd);
  dispose_fleet(&src);
  close_sector(secfd);
  dispose_sector(&sect);
}

void colonize_maximally(char *srcfleetname, char *colonyname, int body,
	int zone, const char *colonytemplate, struct gstrlist_t *specie)
{
  int srcfd,colfd,secfd;
  int x,y;
  struct fleet_t src,col;
  struct sector_t sect;
  struct populii_t colpop, itempop, seed;
  srcfd=open_fleet(srcfleetname,O_RDWR);
  if (srcfd>=0) {
    read_fleet(srcfd,&src,NULL);
    age_fleet(&src);
    secfd=open_sector(g_Player.coords->str);
    if (secfd>=0) {
      colfd=create_fleet_from_template(colonyname,
        colonytemplate,&col);
      if (colfd>=0) {
        col.zone=zone;
        col.body=body;
        col.locus=LOCUS_LANDED;
        read_sector(secfd,&sect);
        colpop=null_populii;
        x=0;
        while (x<src.numships) {
          y=0;
          while (y<src.ships[x].numitems) {
            if (src.ships[x].items[y].popstrs) {
              extract_item_populii(&src.ships[x].items[y],&itempop);
                 /* If you plant a colony exactly on a cycle boundary, you
			may lose a generation.  Not worth fixing. */
              extract_seed_populii(&itempop,&seed,specie);
              absorb_populii_limited(&colpop,&itempop,
                col.ships[0].items[0].maxcrew);
              dispose_item_popstrs(&src.ships[x].items[y]);
              if (itempop.total) {
                absorb_populii(&seed,&itempop);
              }
              populate_item(&src.ships[x].items[y],&seed);
              dispose_populii(&seed);
              dispose_populii(&itempop);
            }
            y++;
          }
          x++;
        }
        if ((colpop.pops) && (colpop.pops[0].total)) {
          if (colpop.pops[0].total<=0) {
            ui_error("Somehow a population has gotten a negative total.");
          }
          x=0;
          while (x<colpop.numpops) {
            if (breedable(&colpop.pops[x])==0) {
              colpop.total-=colpop.pops[x].total;
              move_population_to_fleet(&src,&colpop.pops[x]);
            }
            x++;
          }
        }
        if (colpop.total>0) {
          colpop.lastupdated=MAKROKOSMOS_NOW;
          populate_item(&col.ships[0].items[0],&colpop);
          write_fleet(srcfd,&src);
          col.ships[0].modified=1;
          write_fleet(colfd,&col);
          add_colony(&sect.bodies[g_Player.orbit],colonyname);
          write_sector(secfd,&sect);
        } else {
          ui_message("No populace left over for colony.  Not planting.");
          del_fleet_file(colonyname);
        }
        dispose_populii(&colpop);
        close_fleet(colfd);
        dispose_fleet(&col);
        dispose_sector(&sect);
      } else {
        ui_message("Cannot open colony fleet.\n");
      }
      close_sector(secfd);
    } else {
      ui_message("Cannot open sector.");
    }
    close_fleet(srcfd);
  } else {
    ui_message("colonize_maximally(){...open_fleet(src)...} failed.\n");
  }
}

void colonize_half(char *srcfleetname, char *colonyname, int body,
	int zone, const char *colonytemplate,
	struct gstrlist_t *specieslist)
{
  int srcfd,colfd,secfd;
  int x,y;
  struct fleet_t src,col;
  struct sector_t sect;
  struct populii_t colpop, itempop, seed;
  srcfd=open_fleet(srcfleetname,O_RDWR);
  if (srcfd>=0) {
    read_fleet(srcfd,&src,NULL);
    age_fleet(&src);
    secfd=open_sector(g_Player.coords->str);
    if (secfd>=0) {
      colfd=create_fleet_from_template(colonyname,
        colonytemplate,&col);
      if (colfd>=0) {
        col.zone=zone;
        col.body=body;
        col.locus=LOCUS_LANDED;
        read_sector(secfd,&sect);
        colpop=null_populii;
        x=0;
        while (x<src.numships) {
          y=0;
          while (y<src.ships[x].numitems) {
            if (src.ships[x].items[y].popstrs) {
              extract_item_populii(&src.ships[x].items[y],&itempop);
                 /* If you plant a colony exactly on a cycle boundary, you
			may lose a generation.  Not worth fixing. */
              extract_half_populii(&itempop,&seed,specieslist);
              absorb_populii_limited(&colpop,&seed,
                col.ships[0].items[0].maxcrew);
              if (seed.total) {
                absorb_populii(&itempop,&seed);
              }
              dispose_item_popstrs(&src.ships[x].items[y]);
              populate_item(&src.ships[x].items[y],&itempop);
              dispose_populii(&seed);
              dispose_populii(&itempop);
            }
            y++;
          }
          x++;
        }
        if ((colpop.pops) && (colpop.pops[0].total)) {
          if (colpop.pops[0].total<=0) {
            ui_error("Somehow a population has gotten a negative total.");
          }
          x=0;
          while (x<colpop.numpops) {
            if (breedable(&colpop.pops[x])==0) {
              colpop.total-=colpop.pops[x].total;
              move_population_to_fleet(&src,&colpop.pops[x]);
            }
            x++;
          }
        }
        if (colpop.total>0) {
          colpop.lastupdated=MAKROKOSMOS_NOW;
          populate_item(&col.ships[0].items[0],&colpop);
          write_fleet(srcfd,&src);
          col.ships[0].modified=1;
          write_fleet(colfd,&col);
          add_colony(&sect.bodies[g_Player.orbit],colonyname);
          write_sector(secfd,&sect);
        } else {
          ui_message("No populace left over for colony.  Not planting.");
          del_fleet_file(colonyname);
        }
        dispose_populii(&colpop);
        close_fleet(colfd);
        dispose_fleet(&col);
        dispose_sector(&sect);
      } else {
        ui_message("Cannot open colony fleet.\n");
      }
      close_sector(secfd);
    } else {
      ui_message("Cannot open sector.");
    }
    close_fleet(srcfd);
  } else {
    ui_message("colonize_half(){...open_fleet(src)...} failed.\n");
  }
}


int colonize_open_source(const char *srcfname, int *srcfd,
	struct fleet_t *src, int *secfd)
{
  *srcfd=open_fleet(srcfname,O_RDWR);
  if (*srcfd>=0) {
    read_fleet(*srcfd,src,NULL);
    *secfd=open_sector(g_Player.coords->str);
    if (secfd>=0) {
      return 1;
    } else {
      ui_message("Cannot open sector.");
    }
    close_fleet(*srcfd);
    dispose_fleet(src);
  } else {
    ui_message("Cannot open source fleet.");
  }
  return 0;
}

int colony_init(struct fleet_t *col, const char *colname, int body,
  int zone, const char *colonytemplate)
{
  int colfd=create_fleet_from_template(colname,colonytemplate,col);
  if (colfd>=0) {
    col->zone=zone;
    col->body=body;
    col->locus=LOCUS_LANDED;
  }
  return colfd;
}

void colonize_minimally(const char *srcfleetname, const char *colonyname, 
	int body, int zone, const char *colonytemplate,
	struct gstrlist_t *specie)
{
  int srcfd,colfd,secfd;
  int x,y;
  struct fleet_t src,col;
  struct sector_t sect;
  struct populii_t colpop, itempop, seed;
  if (colonize_open_source(srcfleetname,&srcfd,&src,&secfd)) {
    if ((colfd=colony_init(&col,colonyname,body,zone,colonytemplate))>=0) {
      age_fleet(&src);
      read_sector(secfd,&sect);
      colpop=null_populii;
      x=0;
      while (x<src.numships) {
        y=0;
        while (y<src.ships[x].numitems) {
          if (src.ships[x].items[y].popstrs) {
            extract_item_populii(&src.ships[x].items[y],&itempop);
//            breed_to_now(&itempop,src.ships[x].items[y].maxcrew,
//              &src.ships[x]);
               /* If you plant a colony exactly on a cycle boundary, you
		may lose a generation.  Not worth fixing. */
            extract_seed_populii(&itempop,&seed,specie);
            if (breedable(&itempop.pops[0])) {
              absorb_populii(&colpop,&seed);
              dispose_item_popstrs(&src.ships[x].items[y]);
              populate_item(&src.ships[x].items[y],&itempop);
            }
            dispose_populii(&seed);
            dispose_populii(&itempop);
          }
          y++;
        }
        x++;
      }
      if ((colpop.pops) && (colpop.pops[0].total)) {
        if (colpop.pops[0].total<=0) {
          ui_error("Somehow a population has gotten a negative total.");
        }
        x=0;
        while (x<colpop.numpops) {
          if (breedable(&colpop.pops[x])==0) {
            colpop.total-=colpop.pops[x].total;
            move_population_to_fleet(&src,&colpop.pops[x]);
          }
          x++;
        }
      }
      if (colpop.total>0) {
        colpop.lastupdated=MAKROKOSMOS_NOW;
        populate_item(&col.ships[0].items[0],&colpop);
        write_fleet(srcfd,&src);
        col.ships[0].modified=1;
        write_fleet(colfd,&col);
        add_colony(&sect.bodies[g_Player.orbit],colonyname);
        write_sector(secfd,&sect);
      } else {
        ui_message("No populace left over for colony.  Not planting.");
        del_fleet_file(colonyname);
      }
      dispose_populii(&colpop);
      close_fleet(colfd);
      dispose_fleet(&col);
      dispose_sector(&sect);
    } else {
      ui_message("Cannot open colony fleet.\n");
    }
    close_sector(secfd);
    close_fleet(srcfd);
  } else {
    ui_message("colonize_minimally(){...open_fleet(src)...} failed.\n");
  }
}

void show_composition_short(GString *z)
{
  int linelen=5;
  int min,max;
  min=extract_amt(z->str,"p-");
  max=extract_amt(z->str,"p+");
  printw("Pressure: %d-%d ",min,max);
  linelen+=14;
  linelen+=(min/10)+(max/10);
  printw("\n");
}

void list_zones(struct body_t *b)
{
  int x=0;
  while (x<b->numzones) {
    printw("%d: ",x+1);
    show_composition_short(b->zones[x]);
    x++;
  }
}

void load_fleet_populii(char *fleetname, struct populii_t *p)
{
  struct fleet_t fleet;
  struct populii_t itempop;
  int x=0;
  int y;
  load_fleet(fleetname,&fleet);
  *p=null_populii;
  while (x<fleet.numships) {
    y=0;
    while (y<fleet.ships[x].numitems) {
      extract_item_populii(&fleet.ships[x].items[y],&itempop);
      absorb_populii(p,&itempop);
      dispose_populii(&itempop);
      y++;
    }
    x++;
  }
  dispose_fleet(&fleet);
}

void init_specie_pane(struct specie_pane_t *pane, struct populii_t *pop,
  GString *zone)
{
  int x=0;
  *pane=null_specie_pane;
  pane->specie=pop;
  pane->zone=zone;
  pane->selected=g_malloc(sizeof(int)*pop->numpops);
  while (x<pop->numpops) {
    pane->selected[x++]=1;
  }
}

void specie_pane_down(struct specie_pane_t *pane)
{
  if (pane->pos<pane->specie->numpops) {
    draw_species_bar_unselected(pane);
    pane->pos++;
    if (pane->pos-pane->page<pane->pagesize) {
      draw_species_bar_selected(pane);
    } else {
      pane->page+=pane->pagesize;
      draw_specie_pane(pane);
    }
  }
}

void specie_pane_up(struct specie_pane_t *pane)
{
  if (pane->pos>0) {
    draw_species_bar_unselected(pane);
    pane->pos--;
    if (pane->pos-pane->page>=0) {
      draw_species_bar_selected(pane);
    } else {
      pane->page-=pane->pagesize;
      draw_specie_pane(pane);
    }
  }
}

void dispose_specie_pane(struct specie_pane_t *p)
{
  g_free(p->selected);
}

void show_species_selector(struct populii_t *p, struct gstrlist_t *list, 
  GString *zone)
{
  int x;
  int key;
  int notdone;
  struct specie_pane_t pane;
  *list=null_gstrlist;
  init_specie_pane(&pane,p,zone);
  draw_specie_pane(&pane);
  notdone=1;
  x=0;
  while (notdone) {
    key=getch();
    switch (key) {
    case 0x0A:
    case 0x0D:
    case ' ':
      if (pane.pos>=p->numpops) {
        hide_specie_pane(&pane);
        notdone=0;
        break;
      } else {
        pane.selected[pane.pos]^=1;
        draw_species_bar_checkbox(&pane);
      }
      break;
    case '?':
      if (pane.pos<p->numpops) {
        show_species_info_box(&p->pops[pane.pos].species);
        draw_specie_pane(&pane);
      }
      break;
    case KEY_DOWN:
      specie_pane_down(&pane);
      break;
    case KEY_UP:
      specie_pane_up(&pane);
      break;
    default:
    }
  }
  for (x=0; x<p->numpops; x++) {
    if (pane.selected[x]) {
      gstrlist_add(list,p->pops[x].species.name->str);
    }
  }
  dispose_specie_pane(&pane);
}

int colonization_turncost(int habitability)
{
  return 2+(2-habitability)+sqrt(2-habitability);
}

int colonization_buildtime(int habitability)
{
  return 3+(2-habitability)+(sqrt(2-habitability)*4);
}

void colonize(void)
{
  GString		*colonyname;
  int			zone;
  struct sector_t 	sect;
  struct populii_t	fleetpopulii;
  struct gstrlist_t	specieslist=null_gstrlist;
  int			habitability;
  int			habmod;
  const char		*colonytemplate=NULL;
  if (g_Player.locus==LOCUS_LANDED) {
    load_sector(g_Player.coords->str,&sect);
    if (sect.bodies[g_Player.orbit].zones==NULL) {
      dispose_sector(&sect);
      generate_zones(g_Player.coords->str,g_Player.orbit);
      load_sector(g_Player.coords->str,&sect);
    }
    list_zones(&sect.bodies[g_Player.orbit]);
    zone=read_int("Zone to colonize: ");
    if (sect.bodies[g_Player.orbit].zones==NULL) {
      generate_zones(g_Player.coords->str,g_Player.orbit);
      dispose_sector(&sect);
      load_sector(g_Player.coords->str,&sect);
    }
    if ((zone>0) && (zone<=sect.bodies[g_Player.orbit].numzones)) {
      load_fleet_populii(g_Player.fleet->str,&fleetpopulii);
      if (fleetpopulii.total<=0) {
        ui_message("No population in fleet.  Cannot plant colony.");
        goto colonization_done;
      }
      zone--;
      show_species_selector(&fleetpopulii,&specieslist,
        sect.bodies[g_Player.orbit].zones[zone]);
      if (specieslist.num==0) {
        goto colonization_done;
      }
      habitability=habitability_to_populii(sect.bodies[g_Player.orbit].zones[zone]->str,&fleetpopulii,&specieslist);
      if (habitability<-15) {
        printw("Conditions on this world are too severe to plant a colony.\n");
        refresh();
        goto colonization_done;
      }
      printw("Zone habitability rating is %d\n",habitability);
      if (sect.bodies[g_Player.orbit].radius<110) {
        habmod=-7+(sect.bodies[g_Player.orbit].radius>>4);
      } else {
        habmod=0;
      }
      printw("This colony will take %d turns and %d cycles to complete.\n",
        colonization_turncost(habitability+habmod),
        colonization_buildtime(habitability+habmod));
      if (habitability<0 || sect.bodies[g_Player.orbit].radius<110) {
        if (habitability<0) {
          ui_message("This world's environment can only support an outpost.");
        } else {
          ui_message("This body is too small for a full colony.  Building an outpost...");
        }
        colonytemplate=MAKROKOSMOS_OUTPOST_TEMPLATE;
      } else {
        colonytemplate=MAKROKOSMOS_SETTLEMENT_TEMPLATE;
      }
      colonyname=read_gstring("Colony name: ");
      if (safe_file_name(colonyname)==0) {
        if (colonyname->len) {
          ui_message("Invalid characters in colony name.");
        } else {
          ui_message("Colonization cancelled.");
        }
        g_string_free(colonyname,TRUE);
        return;
      }
      g_string_prepend(colonyname,G_DIR_SEPARATOR_S);
      g_string_prepend(colonyname,g_PlayerName->str);
      printw("Populate colony with:\n(E)ntire non-crew population of your fleet\n(A)lmost all - leave a minimal breeding population in each habitat module\n(H)alf the population of your fleet\n(M)inimal breeding population\n");
      refresh();
      switch (toupper(getch())) {
      case 'E':
        colonize_entire(g_Player.fleet->str, colonyname->str, 
          g_Player.orbit, zone, colonytemplate, &specieslist);
        break;
      case 'A':
        colonize_maximally(g_Player.fleet->str, colonyname->str,
          g_Player.orbit, zone, colonytemplate, &specieslist);
        break;
      case 'H':
        colonize_half(g_Player.fleet->str, colonyname->str,
          g_Player.orbit, zone, colonytemplate, &specieslist);
        break;
      case 'M':
        colonize_minimally(g_Player.fleet->str, colonyname->str,
          g_Player.orbit, zone, colonytemplate, &specieslist);
        break;
      default:
        ui_message("Invalid input - cancelling colonization");
        return;
      }
      g_string_free(colonyname,TRUE);
      colonization_done:
      dispose_populii(&fleetpopulii);
      dispose_gstrlist(&specieslist);
      dispose_sector(&sect);
    }
  } else {
    ui_message("You must land before you can plant a colony.");
  }
}

int regex_match(const char *regexp, const char *s)
{
  regex_t	preg;
  int		err;
  int		retval=0;
  char		c[1024];
  err=regcomp(&preg,regexp,0);
  if (err==0) {
    err=regexec(&preg,s,0,NULL,REG_EXTENDED);
    if (err==0) {
      retval=1;
    } else {
      printw("Err(%s,%s): %d/%s\n",regexp,s,
        regerror(err,&preg,&c[0],1000),&c[0]);
    }
    regfree(&preg);
  } else {
    ui_error("Regex compilation error.");
  }
  return retval;
}

void dispose_player(struct player_t *p)
{
  g_string_free(p->name,TRUE);
  g_string_free(p->coords,TRUE);
  g_string_free(p->dest,TRUE);
  g_string_free(p->fleet,TRUE);
  g_string_free(p->dock,TRUE);
  g_string_free(p->overlords,TRUE);
}

int boardable(char *fleet)
{
  GString *playerpath=g_string_new(g_PlayerName->str);
  struct player_t targetplayer;
  int retval=0;
  int x;
  g_string_append(playerpath,"/");
  if (strncmp(playerpath->str,fleet,g_PlayerName->len+1)==0) {
    retval=1;
  } else {
    g_string_assign(playerpath,fleet);
    x=0;
    while ((x<playerpath->len) && (playerpath->str[x]!=G_DIR_SEPARATOR)) {
      x++;
    }
    g_string_truncate(playerpath,x);
    load_player_file(playerpath->str,&targetplayer,FALSE);
    if (targetplayer.overlords->len) {
      if (regex_match(targetplayer.overlords->str, g_PlayerName->str)) {
        retval=1;
      }
    }
    dispose_player(&targetplayer);
  }
  g_string_free(playerpath,TRUE);
  return retval;
}

void do_board(GString *target)
{
  /* Is the fleet boardable? */
  if (boardable(target->str)) {
    printw("Boarding %s\n",target->str);
    g_string_assign(g_Player.fleet,target->str);
    save_player(g_PlayerName->str,&g_Player);
  } else {
    printw("The fleet %s refuses to allow you to board.\n",target->str);
    /* Add code to signal the target fleet of your attempt */
    /* Add code to signal the target fleet's owner of your attempt. */
  }
}

void board_fleet(void)
{
  struct sector_t sect;
  int targfleet;
  load_sector(g_Player.coords->str,&sect);
  switch (g_Player.locus) {
  case LOCUS_SECTOR:
    list_fleets_numbered(sect.fleets,sect.numfleets);
    targfleet=read_int("Fleet to board: ");
    if ((targfleet>0) && (targfleet<=sect.numfleets)) {
      targfleet--;
      do_board(sect.fleets[targfleet]);
    } else {
      ui_message("Invalid input.");
    }
    break;
  case LOCUS_ORBIT:
    list_fleets_numbered(sect.bodies[g_Player.orbit].ofleets,
      sect.bodies[g_Player.orbit].numofleets);
    targfleet=read_int("Fleet to board: ");
    if ((targfleet>0) && 
      (targfleet<=sect.bodies[g_Player.orbit].numofleets)) {
      targfleet--;
      do_board(sect.bodies[g_Player.orbit].ofleets[targfleet]);
    } else {
      ui_message("Invalid input.");
    }
    break;
  case LOCUS_LANDED:
    list_fleets_numbered(sect.bodies[g_Player.orbit].lfleets,
      sect.bodies[g_Player.orbit].numlfleets);
    targfleet=read_int("Fleet to board: ");
    if ((targfleet>0) && 
      (targfleet<=sect.bodies[g_Player.orbit].numlfleets)) {
      targfleet--;
      do_board(sect.bodies[g_Player.orbit].lfleets[targfleet]);
    } else {
      ui_message("Invalid input.");
    }
    break;
  case LOCUS_DOCKED:
    ui_message("Cannot board fleets while docked.");
    break;
  default:
    ui_error("Unknown locus; cannot board fleets.");
  }
  dispose_sector(&sect);
}

int heal_item(struct item_t *i, int *amt)
{
  int diff;
  int repaired=0;
  if (i->hp<i->maxhp) {
    diff=i->maxhp-i->hp;
    if (diff<=*amt) {
      i->hp+=diff;
      *amt-=diff;
      repaired+=diff;
    } else {
      i->hp+=*amt;
      repaired+=*amt;
      *amt=0;
    }
  }
  return repaired;
}

int heal_ship(struct ship_t *s, int *amt)
{
  int diff;
  int repaired=0;
  int x;
  int broken_immobile=0;
  struct ship_t ship;
  if (s->hp<s->maxhp) {
    if ((s->immobile) && (s->hp<s->maxhp>>3)) {
      broken_immobile=1;
    }
    diff=s->maxhp-s->hp;
    if (diff<=*amt) {
      s->hp+=diff;
      *amt-=diff;
      repaired+=diff;
    } else {
      s->hp+=*amt;
      repaired+=*amt;
      *amt=0;
    }
    if ((broken_immobile) && (s->hp>s->maxhp>>3)) {
      ship=null_ship;
      load_ship(s->base->str,&ship,NULL);
      if (ship.immobile==0) {
        ui_message_sss("Your ",s->base->str," is mobile again.");
        s->immobile=0;
        ((struct fleet_t *)s->fleet)->immobile =
          check_fleet_immobility(s->fleet);
      }
      dispose_ship(&ship);
    }
    if (*amt<=0) {
      return repaired;
    }
  }
  x=0;
  while (x<s->numitems) {
    repaired+=heal_item(&s->items[x],amt);
    x++;
  }
  return repaired;
}

void repair(void)
{
  int r;
  int repaired=0;
  int x=0;
  struct fleet_t fleet;
  int fd;
  printw("How many hitpoints do you want to repair (Cost: 1 turn/HP)? ");
  r=read_int("");
  if (r>0) {
    if (r<=g_Player.turns) {
      fd=open_fleet(g_Player.fleet->str,O_RDWR);
      if (fd>=0) {
        read_fleet(fd,&fleet,NULL);
        while (x<fleet.numships) {
          repaired+=heal_ship(&fleet.ships[x],&r);
          if (r<=0) {
            goto done;
          }
          x++;
        }
        done:
        write_fleet(fd,&fleet);
        close_fleet(fd);
        printw("%d hitpoints healed.\n",repaired);
        g_Player.turns-=repaired;
        save_player(g_PlayerName->str,&g_Player);
      } else {
        ui_config_error("Error opening fleet (%d: %s)\n",errno);
      }
    } else {
      printw("You only have %d turns.\n",g_Player.turns);
      refresh();
    }
  } else {
    ui_message("Cancelled.\n");
  }
}

void fleetpane_delve(struct fleetpane_t *pane)
{
  if (pane->state<PANE_ITEM) {
    pane->state++;
    if (pane->state==PANE_SHIP) {
      pane->spos=0;
      pane->spage=0;
    }
    draw_fleet_pane(pane);
    if (pane->state==PANE_SHIP) {
      draw_item_bar_selected(pane);
    }
  }
}

void fleetpane_out(struct fleetpane_t *pane)
{
  if (pane->state>0) {
    pane->state--;
    draw_fleet_pane(pane);
  }
}

void equip_single_item(void)
{
  struct history_chain	changes=null_history_chain;
  struct fleetpane_t		pane;
  struct fleet_t	fleet,fleet2;
  int			notdone=1;
  int			fd;
  pane=new_fleetpane(subwin(stdscr,20,79,0,0),20,&fleet);
  load_fleet(g_Player.fleet->str,&fleet);
  clear();
  draw_fleet_pane(&pane);
  draw_ship_bar_selected(&pane);
  draw_equip_footer(&pane);
  while (notdone) {
    switch (toupper(getch())) {
    case 0xC:
      clear();
      draw_fleet_pane(&pane);
      draw_bar_selected(&pane);
      draw_equip_footer(&pane);
      break;
    case 0xA:
    case 0xD:
    case KEY_RIGHT:
      fleetpane_delve(&pane);
      draw_equip_footer(&pane);
      break;
    case ' ':
      if (pane.state==PANE_SHIP) {
        pane.fleet->ships[pane.pos].items[pane.spos].equipped^=1;
        if ((pane.fleet->ships[pane.pos].items[pane.spos].equipped !=
          pane.fleet->ships[pane.pos].items[pane.spos].defaultequipped)
          && (pane.fleet->ships[pane.pos].items[pane.spos].modified==0)) {
          pane.fleet->ships[pane.pos].items[pane.spos].modified=0;
          if (pane.fleet->ships[pane.pos].modified==0) {
            pane.fleet->ships[pane.pos].modified=1;
          }
        }
        draw_item_bar_selected(&pane);
        add_history_entry(&changes,CHANGE_EQUIP,0,pane.pos,pane.spos,0,0,0);
      }
      break;
    case 'Q':
      notdone=0;
      break;
    case KEY_DOWN:
      fleetpane_down(&pane);
      break;
    case KEY_LEFT:
      fleetpane_out(&pane);
      draw_pane_bar_selected(&pane);
      draw_equip_footer(&pane);
      break;
    case KEY_UP:
      fleetpane_up(&pane);
      break;
    default:
      move(pane.height+3,0);
      ui_error("Invalid keystroke.");
    }
  }
  dispose_fleetpane(&pane);
  fd=open_fleet(g_Player.fleet->str,O_RDWR);
  if (fd>=0) {
    read_fleet(fd,&fleet2,NULL);
    apply_history_to_fleet(&fleet2,&changes);
    if (fleeteq(&fleet,&fleet2)) {
      write_fleet(fd,&fleet2);
    } else {
      ui_error("Consistency lost.");
    }
    close_fleet(fd);
    dispose_fleet(&fleet2);
  }
  dispose_history_chain(&changes);
  dispose_fleet(&fleet);
}

void equip_menu(void)
{
  int notdone=1;
  int c;
  while (notdone) {
    c=read_char("Equip menu (?=help): ");
    switch (toupper(c)) {
    case 0x0C:
      clear();
      break;
    case '?':
      show_equip_help();
      break;
    case 'E':
      equip_everything_in_fleet(g_Player.fleet->str);
      break;
    case 'Q':
      notdone=0;
      break;
    case 'T':
      equip_single_item();
      break;
    default:
      printw("Invalid key %c pressed.\n",c);
    }
  }
}

int dockable(char *colonyname)
{
//  return boardable(colonyname);
  return TRUE;
}

void del_colony(struct body_t *body, GString *colony)
{
  int x=0;
  while (x<body->numcolonies) {
    if (gstreq(body->colonies[x],colony)) {
      g_string_free(body->colonies[x],TRUE);
      body->numcolonies--;
      body->colonies[x]=body->colonies[body->numcolonies];
      body->colonies=g_realloc(body->colonies,body->numcolonies*sizeof(GString *));
      return;
    }
    x++;
  }
}

void remove_colony_from_body(char *coords, int body, GString *colony)
{
  struct sector_t sect;
  int fd=open_sector(coords);
  if (fd>=0) {
    read_sector(fd,&sect);
    if ((body>=0) && (body<=sect.numbodies)) {
      del_colony(&sect.bodies[body],colony);
      write_sector(fd,&sect);
    }
    close_sector(fd);
    dispose_sector(&sect);
  } else {
    ui_error("Could not open sector file.");
  }
}

GString *owner(GString *fleet)
{
  GString *ret=g_string_new(fleet->str);
  int x=0;
  while (x<ret->len) {
    if (ret->str[x]=='/') {
      g_string_truncate(ret,x);
    }
    x++;
  }
  return ret;
}

void send_docking_notification(struct player_t *player)
{
  GString *s;
  GString *to;
  if ((strncmp(g_Player.name->str,g_Player.dock->str,g_Player.name->len)==0)
    && (g_Player.dock->str[g_Player.name->len]=='/')) {
    /* No need to notify oneself of dockings. */
    return;
  } else {
    s=g_string_new("");
    g_string_sprintf(s,"%s docked at your colony %s.",
      player->name->str,player->dock->str);
    to=owner(g_Player.dock);
    send_mail(player->dock->str,to->str,s->str);
    g_string_free(to,TRUE);
    g_string_free(s,TRUE);
  }
}

void dock_fleet(void)
{
  struct sector_t sect;
  int target;
  int fd;
  struct fleet_t fleet;
  if ((g_Player.locus!=LOCUS_LANDED) && (g_Player.locus!=LOCUS_DOCKED)) {
    ui_message("Can only dock at a colony if already landed.");
    return;
  }
  load_sector(g_Player.coords->str,&sect);
  printw("Colonies:\n");
  list_fleets_numbered(sect.bodies[g_Player.orbit].colonies,
    sect.bodies[g_Player.orbit].numcolonies);
  target=read_int("Dock fleet at: ");
  if ((target>0) && (target<=sect.bodies[g_Player.orbit].numcolonies)) {
    target--;
    if (dockable(sect.bodies[g_Player.orbit].colonies[target]->str)) {
      fd=open_fleet(sect.bodies[g_Player.orbit].colonies[target]->str,O_RDWR);
      if (fd>=0) {
        read_fleet(fd,&fleet,&sect);
	if (fleet.locus!=LOCUS_LANDED) {
	  /* Note: If docking at non-colonies is allowed, this will have
           * to be changed. */
          ui_error("Colony is not landed!  Fixing...");
          fleet.locus=LOCUS_LANDED;
        }
        age_fleet(&fleet);
        write_fleet(fd,&fleet);
        close_fleet(fd);
        dispose_fleet(&fleet);
        g_Player.locus=LOCUS_DOCKED;
        g_string_assign(g_Player.dock,
          sect.bodies[g_Player.orbit].colonies[target]->str);
        send_docking_notification(&g_Player);
        ui_message("Docked.");
      } else {
        remove_colony_from_body(g_Player.coords->str,g_Player.orbit,
          sect.bodies[g_Player.orbit].colonies[target]);
        printw("The colony you attempted to dock at has been destroyed.\n");
      }
    }
  }
  dispose_sector(&sect);
}

void set_shipyard_specs(struct item_t *i, int tech, int volume)
{
  i->modified=MODDED_CUSTOM;
//  i->volume=volume+greater((volume/128),5);
  i->surface=greater(cbrt(volume)*6,7);
  i->maxhp=(volume/500)+(tech*30);
  i->hp=(volume/700)+(tech*22);
  i->volume=volume;
  i->updated=MAKROKOSMOS_NOW;
  i->add_on=1;
  i->cyclestocompletion=(tech*10)+(volume/1000);
  i->buildtech.bio=lesser(tech,g_Player.tech.bio);
  i->buildtech.emf=lesser(tech,g_Player.tech.emf);
  i->buildtech.grav=lesser(tech,g_Player.tech.grav);
  i->buildtech.logic=lesser(tech,g_Player.tech.logic);
  i->buildtech.mech=lesser(tech,g_Player.tech.mech);
  i->buildtech.psi=lesser(tech,g_Player.tech.psi);
  i->buildtech.quantum=lesser(tech,g_Player.tech.quantum);
}

void build_shipyard(void)
{
  struct sector_t	sect;
  int			volume;
  int			tech;
  struct fleet_t	fleet;
  int			fd;
  int			c;
  int			hightech;
  if (g_Player.locus!=LOCUS_DOCKED) {
    ui_message("You must be docked to build a shipyard.");
    return;
  }
  volume=read_int("Shipyard volume: ");
  hightech=greater(greater(greater(g_Player.tech.bio,g_Player.tech.emf),
    greater(g_Player.tech.grav,g_Player.tech.logic)),
    greater(greater(g_Player.tech.mech,g_Player.tech.psi),
    g_Player.tech.quantum));
  tech=read_int("Techlevel: ");
  if (hightech<tech) {
/*    attrset(COLOR_TEXT);*/
    ui_message_sds("The most advanced shipyard you can build has a maximum technological\nlevel of ",hightech,".");
    tech=hightech;
  }
  if (g_Player.turns<(tech/8)+2) {
    printw("Insufficient turns to build that shipyard (%d needed, %d available.",
      (tech/8)+2,g_Player.turns);
  }
  printw("This shipyard will take %d cycles and %d turns to complete.\nBuild it? ",
    (tech*10)+(volume/1000),(tech/8)+2);
  c=getch();
  printw("\n");
  if (toupper(c)=='Y') {
    g_Player.turns-=(tech/8)+2;
    fd=open_fleet(g_Player.dock->str,O_RDWR);
    if (fd>=0) {
      load_sector(g_Player.coords->str,&sect);
      read_fleet(fd,&fleet,&sect);
      age_fleet(&fleet);
      if (fleet.numships>0) {
        add_item_to_ship(&fleet.ships[0],"Defaults/shipyard");
        set_shipyard_specs(&fleet.ships[0].items[fleet.ships[0].numitems-1],
	  tech,volume);
      } else {
        ui_error("No ships in fleet.");
      }
      write_fleet(fd,&fleet);
      close_fleet(fd);
      dispose_fleet(&fleet);
      dispose_sector(&sect);
    } else {
      switch (g_fleet_errno) {
      case ENOENT:
        remove_colony_from_body(g_Player.coords->str,g_Player.orbit,
          g_Player.dock);
        ui_message("You cannot build here - this colony no longer exists.");
        g_Player.locus=LOCUS_LANDED;
        break;
      default:
        ui_config_error("Error %d (%s) opening colony.\n",g_fleet_errno);
      }
    }
  } else {
    ui_message("Construction cancelled.");
  }
}

void get_item_list(struct fleet_t *f, struct item_p **list, int *num,
	int iclass)
{
	/* Caller is expected to initialize list and num.  This allows
	   get_item_list() to append items to a preexisting list. */
  int x=0;
  int y;
  while (x<f->numships) {
    y=0;
    while (y<f->ships[x].numitems) {
      if (f->ships[x].items[y].class==iclass) {
        ++*num;
        *list=g_realloc(*list,sizeof(struct item_p)*(*num));
        (*list)[*num-1]=(struct item_p){ ship: x, item: y };
      }
      y++;
    }
    x++;
  }
}

void show_shipyard_status(int x, struct item_t *i)
{
  if (i->cyclestocompletion) {
    printw("%d: %9d | Incomplete, %d cycles remaining\n",x,
      i->volume,i->cyclestocompletion);
  } else if (i->cargo->len) {
    printw("%d: %9d | Building a %s.\n",x,i->volume,i->cargo->str);
  } else {
    printw("%d: %9d | Free\n",x,i->volume);
  }
}

void shipyard_build(struct fleet_t *f, struct item_p i, struct ship_t *ship)
{
  int err;
  if ((i.ship>=f->numships) || (i.item>=f->ships[i.ship].numitems) ||
    (f->ships[i.ship].items[i.item].class!=ICLASS_SHIPYARD)) {
    ui_message("Fleet has changed under us.  Aborting.");
    return;
  }
  if (f->ships[i.ship].items[i.item].cyclestocompletion) {
    ui_message("Shipyard is busy.");
    return;
  }
  if (f->ships[i.ship].items[i.item].cargoqty) {
    move_cargo_to_fleet(f,&f->ships[i.ship].items[i.item]);
  }
  if (shipyard_can_build(&f->ships[i.ship].items[i.item],ship,&err)) {
    g_string_assign(f->ships[i.ship].items[i.item].cargo,ship->base->str);
    f->ships[i.ship].items[i.item].updated=MAKROKOSMOS_NOW;
  } else {
    switch (err) {
    case ERR_TOOBIG:
      ui_message("This ship is too big to build in this shipyard.");
      break;
    case ERR_TOOADVANCED:
      ui_message("This ship requires technology exceeding the capacity of this shipyard.");
      break;
    default:
      ui_error("This ship cannot be built in this shipyard for unknown reasons.");
    }
  }
}

void build_ship(void)
{
  struct fleet_t	fleet;
  struct sector_t	sect;
  int			fd;
  struct item_p		*list=NULL;
  int			numyards=0;
  int			x;
  int			shipyard;
  GString		*shiptype;
  struct ship_t		ship;
  if (g_Player.locus!=LOCUS_DOCKED) {
    ui_message("You must be docked to build a ship.\n");
    return;
  }
  load_sector(g_Player.coords->str,&sect);
  fd=open_fleet(g_Player.dock->str,O_RDWR);
  if (fd>=0) {
    read_fleet(fd,&fleet,&sect);
    age_fleet(&fleet);
    get_item_list(&fleet,&list,&numyards,ICLASS_SHIPYARD);
    write_fleet(fd,&fleet);
    close_fleet(fd);
    printw("#     Volume | Status\n");
    x=0;
    while (x<numyards) {
      show_shipyard_status(x+1,&fleet.ships[list[x].ship].items[list[x].item]);
      x++;
    }
    refresh();
    if (boardable(g_Player.dock->str)) {
      shipyard=read_int("Shipyard to build in: ");
      if ((shipyard>0) && (shipyard<=numyards)) {
        shipyard--;
        if (fleet.ships[list[shipyard].ship].items[list[shipyard].item].cyclestocompletion) {
          ui_message("This shipyard is incomplete.");
        } else {
          shiptype=read_gstring("Ship to build: ");
          ship=null_ship;
          if (load_ship(shiptype->str,&ship,NULL)) {
            dispose_fleet(&fleet);
            fd=open_fleet(g_Player.dock->str,O_RDWR);
            read_fleet(fd,&fleet,NULL);
            shipyard_build(&fleet,list[shipyard],&ship);
            write_fleet(fd,&fleet);
          }
          g_string_free(shiptype,TRUE);
          close_fleet(fd);
          dispose_ship(&ship);
        }
      } else {
        ui_message("shipyard out of range.");
      }
    } else {
      ui_message("You are not allowed to control this colony's construction.");
    }
    dispose_fleet(&fleet);
    g_free(list);
  }
  dispose_sector(&sect);
}

void set_new_airminer_defaults(struct item_t *i, char *chem)
{
  g_string_assign(i->cargo,chem);
  i->add_on=1;
  i->updated=MAKROKOSMOS_NOW;
}

void build_airminer(void)
{
  int fd;
  struct fleet_t fleet;
  struct sector_t sect;
  GString *chem;
  if (g_Player.locus!=LOCUS_DOCKED) {
    ui_message("You must be docked to build an air miner.");
    return;
  }
  fd=open_fleet(g_Player.dock->str,O_RDONLY);
  if (fd>=0) {
    load_sector(g_Player.coords->str,&sect);
    read_fleet(fd,&fleet,&sect);
    close_fleet(fd);
    if (fleet_in_sector(&fleet,&sect)) {
      show_zone_composition(sect.bodies[fleet.body].zones[fleet.zone]);
      chem=read_gstring("Chemical to mine: ");
      if (extract_amt(fleet.sect->bodies[fleet.body].zones[fleet.zone]->str,chem->str)) {
        fd=open_fleet(g_Player.dock->str,O_RDWR);
        if (fd>=0) {
          dispose_fleet(&fleet);
          read_fleet(fd,&fleet,&sect);
          age_fleet(&fleet);
          if (fleet.numships>=0) {
            add_item_to_ship(&fleet.ships[0],"Defaults/airminer");
            set_new_airminer_defaults(
              &fleet.ships[0].items[fleet.ships[0].numitems-1]
              ,chem->str);
          } else {
            ui_error("No ships in fleet");
          }
          write_fleet(fd,&fleet);
          close_fleet(fd);
        }
      } else {
        printw("Insufficient %s to fuel airminer in this zone.\n",chem->str);
        refresh();
      }
    } else {
      ui_message("Error: Colony isn't in a zone.");
    }
    dispose_fleet(&fleet);
  } else {
    ui_error("Cannot read fleet.");
  }
  g_string_free(chem,TRUE);
}

void build_menu(void)
{
  int c;
  int notdone=1;
  while (notdone) {
    attrset(COLOR_PROMPT);
    printw("Build command (? for help): ");
    attrset(COLOR_INPUT);
    refresh();
    c=getch();
    printw("\n");
    switch (toupper(c)) {
    case '?':
      show_build_help();
      break;
    case 'A':
      build_airminer();
      break;
    case 'Q':
      notdone=0;
      break;
    case 'S':
      build_ship();
      break;
    case 'Y':
      build_shipyard();
      break;
    default:
      printw("Invalid key %c.  Press ? for help.\n",c);
    }
  }
}

void show_current_time(void)
{
  printw("Current cycle: %d\n",(int)MAKROKOSMOS_NOW);
  printw("(Seconds per cycle: %d)\n",MAKROKOSMOS_CYCLE_LENGTH);
  refresh();
}

void scan_from_sector(void)
{
  struct coord_t	cbase,cmin,cmax,f,g,slicesize;
  struct sector_t	sect;
  int			fd;
  struct fleet_t	fleet;
  GString		*s;
  int			range=2; /* Default range is 2: Eyes. */
  int			fleetrange=0;
  int			bodyrange=2;
  int			ship=0;
  int			item;
  int			dist;
  int			slicelen;
  char			*slice;
  int			slicepos;
  if (load_fleet(g_Player.fleet->str,&fleet)==0) {
    ui_error("Could not load fleet.");
    dispose_fleet(&fleet);
    return;
  }
  /* Identify the strongest scanners. */
  while (ship<fleet.numships) {
    item=0;
    while (item<fleet.ships[ship].numitems) {
      item++;
      switch (fleet.ships[ship].items[item].class) {
      case ICLASS_FLEETSENSOR:
        fleetrange=greater(fleetrange,fleet.ships[ship].items[item].strength);
        break;
      case ICLASS_BODYSENSOR:
        bodyrange=greater(bodyrange,fleet.ships[ship].items[item].strength);
      }
    }
    ship++;
  }
  dispose_fleet(&fleet);
  range=greater(range,greater(fleetrange,bodyrange));
  if (range>=10) {
    ui_error("Warning: Scanner range >=10.");
  }
  string_to_coords(g_Player.coords,&cbase);
  cmin=(struct coord_t){x: greater(cbase.x-range,0),
    y: greater(cbase.y-range,0), z: greater(cbase.z-range,0)};
  cmax=(struct coord_t){x: lesser(cbase.x+range,g_Config_Horizon),
    y: lesser(cbase.y+range,g_Config_Horizon),
    z: lesser(cbase.z+range,g_Config_Horizon)};
  slicesize=(struct coord_t){x: (cmax.x-cmin.x)+1,
    y: (cmax.y-cmin.y)+1, z: (cmax.z-cmin.z)+1};
  slicelen=slicesize.x*slicesize.y*slicesize.z;
  slice=g_malloc(slicelen+1024);
  load_chart_slice(&cmin,&cmax,slice,slicelen);
  ui_init_scan_result(&cbase,range);
  f.x=0;
  while (f.x<slicesize.x) {
    f.y=0;
    while (f.y<slicesize.y) {
      f.z=0;
      while (f.z<slicesize.z) {
        slicepos=f.x+(f.y*slicesize.x)+(f.z*slicesize.x*slicesize.y);
        if (slicepos>=slicelen) {
          ui_error("Slicepos>=slicelen");
        }
        slice[slicepos] |= CHART_BODYSCANNED;
        g=(struct coord_t){x:cmin.x+f.x, y:cmin.y+f.y, z:cmin.z+f.z};
        s=coords_to_string(&g);
        fd=open_sector(s->str);
        if (fd>=0) {
          sect=null_sector;
          read_sector(fd,&sect);
          dist=vector_distance_3d(cbase,g);
          if (sect.numbodies>0) {
            if (dist<=bodyrange) {
              ui_add_scan_result_bodies(sect.numbodies,&g);
            } else {
              ui_add_scan_result_system(&sect,&g);
            }
            slice[slicepos] |= CHART_BODIES;
          }
          if (dist<=fleetrange) {
            slice[slicepos] |= CHART_FLEETSCANNED;
            ui_add_scan_result_fleets(&sect,&g);
          }
          dispose_sector(&sect);
          close_sector(fd);
        }
	g_string_free(s,TRUE);
        f.z++;
      }
      f.y++;
    }
    f.x++;
  }
  write_chart_slice(slice,&cmin,&slicesize);
  g_free(slice);
}

void scan_from_body(void)
{
  struct sector_t sect;
  load_sector(g_Player.coords->str,&sect);
  if (g_Player.orbit<sect.numbodies) {
    show_body_scan(&sect.bodies[g_Player.orbit]);
  } else {
    ui_error("Invalid body number.");
  }
  dispose_sector(&sect);
}

void scan_from_orbit(void)
{
  if (g_Player.turns<=0) {
    ui_message("Insufficient turns to scan from orbit.\n");
    return;
  }
  g_Player.turns--;
  scan_from_body();
}

void scan_from_dock(void)
{
  struct sector_t	sect;
  struct fleet_t	colony;
  printw("Add code to show colony info:\n");
  printw("Environment surrounding colony:\n");
  load_sector(g_Player.coords->str,&sect);
  load_fleet(g_Player.dock->str,&colony);
  if (fleet_in_sector(&colony,&sect)) {
    show_zone_composition(sect.bodies[colony.body].zones[colony.zone]);
  } else {
    ui_error("Fleet is not properly located within sector.");
    printw("Add code to perform sector integrity check.\n");
  }
  refresh();
}

void scan(void)
{
  switch (g_Player.locus) {
  case LOCUS_SECTOR:
    scan_from_sector();
    break;
  case LOCUS_ORBIT:
	scan_from_orbit();
    break;
  case LOCUS_LANDED:
    scan_from_body();
    break;
  case LOCUS_DOCKED:
    scan_from_dock();
    break;
  default:
    printw("Unknown locus.  Cannot scan.\n");
  }
}

int move_to_sector(struct coord_t *dest)
{
  GString		*s;
  int			length;
  struct sector_t	sect;
  int			fd;
  struct coord_t	c;
  struct fleet_t	fleet;
  int			retval=0;
  if (g_Player.turns<=0) {
    ui_message("Insufficient turns to enter hyperspace.");
    return 0;
  }
  if (g_Player.locus==LOCUS_SECTOR) {
    if (dest) {
      s=coords_to_string(dest);
      c=*dest;
    } else {
      s=read_gstring("Destination: ");
      if (s->len==0) {
        goto dont_move_ship;
      }
      string_to_coords(s,&c);
      g_string_sprintf(s,"%d/%d/%d",c.x,c.y,c.z);
    }
    fd=open_fleet(g_Player.fleet->str,O_RDONLY);
    if (fd>=0) {
      read_fleet(fd,&fleet,NULL);
      close_fleet(fd);
      fd=fleet.immobile;
      dispose_fleet(&fleet);
      if (fd) {
        ui_message("Fleet cannot be moved.\nEither split your immobile ships off into another fleet or repair them.");
        goto dont_move_ship;
      }
    } else {
      ui_config_error("Error opening fleet.",g_fleet_errno);
    }
    if (coords_oob(&c)) {
      ui_message_sds("The edge of the universe is at ",g_Config_Horizon,
        ".  You cannot go beyond it.");
      goto dont_move_ship;
    }
    length=vector_distance_s(g_Player.coords,s);
    if (length>g_Player.turns) {
      printw("This voyage will take %d turns.  You only have %d left.  Go anyway?",
        length,g_Player.turns);
      if (toupper(getch())!='Y') {
        goto dont_move_ship;
      }
    }
    fd=open_sector(g_Player.coords->str);
    if (fd>=0) {
      read_sector(fd,&sect);
      del_sector_fleet(&sect,g_Player.fleet);
      write_sector(fd,&sect);
      close_sector(fd);
      dispose_sector(&sect);
    } else {
      ui_error("Error opening source sector.");
    }
    g_Player.turns -= length;
    if (g_Player.turns<0) {
      g_Player.locus=LOCUS_HYPERSPACE;
      g_string_assign(g_Player.dest,s->str);
      retval=1;
    } else {
      g_string_assign(g_Player.coords,s->str);
      fd=open_sector(g_Player.coords->str);
      if (fd<0) {
        fd=create_sector(g_Player.coords->str);
      }
      if (fd>=0) {
        read_sector(fd,&sect);
        add_fleet(&sect,g_Player.fleet->str);
        write_sector(fd,&sect);
        close_sector(fd);
        dispose_sector(&sect);
        retval=1;
      }
    }
    append_chart_dotor(&c,CHART_VISITED);
    dont_move_ship:
    g_string_free(s,TRUE);
  } else {
    ui_message("You must leave orbit before embarking on interstellar voyages.");
  }
  return retval;
}

void leave_hyperspace(void)
{
  struct coord_t	src,dest,new;
  int			length,travelled;
  int			fd;
  struct sector_t	sect;
  string_to_coords(g_Player.coords,&src);
  string_to_coords(g_Player.dest,&dest);
  length=vector_distance_3d(src,dest);
  travelled=length+g_Player.turns;
  determine_vector_pos(&src,&dest,&new,length,travelled);
  g_string_sprintf(g_Player.coords,"%d/%d/%d",new.x,new.y,new.z);
  fd=open_sector(g_Player.coords->str);
  if (fd<0) {
    fd=create_sector(g_Player.coords->str);
  }
  if (fd>=0) {
    read_sector(fd,&sect);
    add_fleet(&sect,g_Player.fleet->str);
    write_sector(fd,&sect);
    close_sector(fd);
    dispose_sector(&sect);
    g_Player.turns--;
    g_Player.locus=LOCUS_SECTOR;
  } else {
    ui_message("Unable to leave hyperspace - destination has low reality quotient.");
  }
}

void fleet_panes(struct fleetpane_t *pane1, struct fleetpane_t *pane2,
  struct history_chain *changes)
{
  int notdone=1;
  int currpanenum=0;
  struct fleetpane_t *currpane=pane1;
  int paneheight;
  int c;
  struct fleetpane_t *panes[2]={pane1,pane2};
  draw_fleet_pane(pane1);
  draw_fleet_pane(pane2);
  getmaxyx(pane1->win,paneheight,c);
  refresh();
  if (pane2->fleet->numships) {
    draw_ship_bar_target(pane2->win,&pane2->fleet->ships[0],0);
  }
  while (notdone) {
    draw_pane_bar_selected(currpane);
    c=getch();
    wmove(currpane->win,currpane->pos-currpane->page,0);
    draw_pane_bar_unselected(currpane);
    switch (toupper(c)) {
    case '\t':
      if (currpane->state==PANE_FLEET) {
        draw_ship_bar_target(currpane->win,
          &currpane->fleet->ships[currpane->pos],
          currpane->pos-currpane->page);
      } else if (currpane->state) {
        draw_pane_bar_unselected(currpane);
      }
      currpanenum=currpanenum^1;
      currpane=panes[currpanenum];
      break;
    case 0xA:
    case 0xD:
    case KEY_RIGHT:
      fleetpane_delve(currpane);
      draw_equip_footer(currpane);
      break;
    case ' ':
      switch (currpane->state) {
      case PANE_FLEET:
        if (currpane->pos<currpane->fleet->numships) {
          move_ship_to_fleet(currpane->fleet,currpane->pos,
            panes[currpanenum^1]->fleet);
          add_history_entry(changes,CHANGE_MOVESHIP,currpanenum,
            currpane->pos,0,currpanenum^1,0,0);
          if (currpane->pos>=currpane->fleet->numships) fleetpane_up(currpane);
          draw_fleet_pane(currpane);
          draw_fleet_pane(panes[currpanenum^1]);
        }
        break;
      case PANE_SHIP:
        if (currpane->fleet->ships[currpane->pos].items[currpane->spos].immobile==0) {
          if (move_item_to_ship(&currpane->fleet->ships[currpane->pos],
            currpane->spos,
            &panes[currpanenum^1]->fleet->ships[panes[currpanenum^1]->pos])) {
            add_history_entry(changes,CHANGE_MOVEITEM,currpanenum^1,
              currpane->pos,currpane->spos,
              currpanenum,panes[currpanenum^1]->pos,0);
            if (currpane->spos>=currpane->fleet->ships[currpane->pos].numitems) {
              fleetpane_up(currpane);
            }
            draw_fleet_pane(currpane);
            draw_fleet_pane(panes[currpanenum^1]);
            if (panes[currpanenum^1]->state==PANE_FLEET) {
              draw_ship_bar_target(panes[currpanenum^1]->win,
                &panes[currpanenum^1]->fleet->ships[panes[currpanenum^1]->pos],
                panes[currpanenum^1]->pos-panes[currpanenum^1]->page);
            }
          } else {
            ui_pane_warning(currpane,"Unable to move item.");
          }
        } else {
          ui_pane_warning(currpane,"Item is immobile.");
        }
        break;
      default:
        ui_error("Pane mode not supported in move.");
      }
      break;
    case KEY_DOWN:
      fleetpane_down(currpane);
      break;
    case KEY_LEFT:
      fleetpane_out(currpane);
      draw_pane_bar_selected(currpane);
      break;
    case KEY_UP:
      fleetpane_up(currpane);
      break;
    case 'Q':
      notdone=0;
    }
  }
}

void fleet_organization_panes(char *destfleetname, char *srcfleetname)
{
  struct history_chain changes=null_history_chain;
  struct fleet_t src, dest;
  int srcfd,destfd;
  struct fleet_t workdest, worksrc;
  struct fleetpane_t destpane, srcpane;
  int paneheight=20;
  destpane=new_fleetpane(subwin(stdscr,paneheight,39,0,0),paneheight,&workdest);
  /*(struct fleetpane_t){win: subwin(stdscr,paneheight,39,0,0), 
    pos: 0, page: 0, height: paneheight, state: 0, spage: 0,
    fleet: &workdest};*/
  srcpane=new_fleetpane(subwin(stdscr,paneheight,39,0,40),paneheight,&worksrc);
  /*(struct pane_t){win: subwin(stdscr,paneheight,39,0,40),
    pos: 0, page: 0, height: paneheight, fleet: &worksrc};*/
  load_fleet(destfleetname,&workdest);
  load_fleet(srcfleetname,&worksrc);
  clear();
  move(paneheight+1,0);
  printw("Up/Down arrows move, [tab] switches panes, [space] moves a ship.\n");
  printw("[Enter] will eventually let you delve into ships.\n");
  refresh();
  fleet_panes(&destpane, &srcpane, &changes);
  move(paneheight+3,0);
  printw("Save changes to fleet? ");
  refresh();
  if (toupper(getch())=='Y') {
    printw("Y\n");
    srcfd=open_fleet(srcfleetname,O_RDWR);
    if (srcfd>=0) {
      destfd=open_fleet(destfleetname,O_RDWR);
      if (destfd>=0) {
        read_fleet(srcfd,&src,NULL);
        read_fleet(destfd,&dest,NULL);
        apply_history_to_fleetpair(&dest,&src,&changes);
        if (fleeteq(&workdest,&dest) && fleeteq(&worksrc,&src)) {
          if (dest.numships) {
            dest.immobile=check_fleet_immobility(&dest);
            write_fleet(destfd,&dest);
          } else {
            del_fleet_file(destfleetname);
          }
          if (src.numships) {
            src.immobile=check_fleet_immobility(&src);
            write_fleet(srcfd,&src);
          } else {
            del_fleet_file(srcfleetname);
          }
        } else {
          ui_message("Consistency lost.");
        }
        close_fleet(destfd);
        dispose_fleet(&dest);
        dispose_fleet(&src);
      }
      close_fleet(srcfd);
    }
  } else {
    printw("\n");
  }
  delwin(destpane.win);
  delwin(srcpane.win);
  dispose_history_chain(&changes);
  dispose_fleet(&workdest);
  dispose_fleet(&worksrc);
}

GString *select_fleet(const char *prompt, struct sector_t *sect, int locus,
  int body)
{
  GString *retval=NULL;
  int	targfleet;
  switch (locus) {
  case LOCUS_SECTOR:
    list_fleets_numbered(sect->fleets,sect->numfleets);
    targfleet=read_int(prompt);
    if ((targfleet>0) && (targfleet<=sect->numfleets)) {
      targfleet--;
      retval=g_string_new(sect->fleets[targfleet]->str);
    } else {
      ui_message("Invalid input.");
    }
    break;
  case LOCUS_ORBIT:
    list_fleets_numbered(sect->bodies[body].ofleets,
      sect->bodies[body].numofleets);
    targfleet=read_int(prompt);
    if ((targfleet>0) && (targfleet<=sect->bodies[body].numofleets)) {
      targfleet--;
      retval=g_string_new(sect->bodies[body].ofleets[targfleet]->str);
    } else {
      ui_message("Invalid input.");
    }
    break;
  case LOCUS_LANDED:
    list_fleets_numbered(sect->bodies[body].lfleets,
      sect->bodies[body].numlfleets);
    targfleet=read_int(prompt);
    if ((targfleet>0) &&
      (targfleet<=sect->bodies[body].numlfleets)) {
      targfleet--;
      retval=g_string_new(sect->bodies[body].lfleets[targfleet]->str);
    } else {
      ui_message("Invalid input.");
    }
    break;
  default:
    ui_error("select_fleet() called in invalid locus.");
    retval=NULL;
  }
  return retval;
}

void exchange_ships(void)
{
  GString		*targfleet;
  struct sector_t	sect;
  int			fd;
  load_sector(g_Player.coords->str,&sect);
  targfleet=select_fleet("Fleet to exchange ships with: ",&sect,g_Player.locus,g_Player.orbit);
  if (targfleet) {
    if (boardable(targfleet->str)) {
      if (gstreq(targfleet,g_Player.fleet)) {
        ui_message("Exchanging ships with the fleet you're currently in is a little silly.");
      } else {
        fleet_organization_panes(g_Player.fleet->str,targfleet->str);
        fd=open_fleet_quiet(g_Player.fleet->str,O_RDWR);
        if (fd<0) {
          if (g_fleet_errno==ENOENT) {
            /* fleet deleted */
            del_fleet(g_Player.coords->str,g_Player.fleet,&fd);
            /* If you moved all your ships into the target fleet, you
		move into the fleet too. */
            g_string_assign(g_Player.fleet,targfleet->str);
          } else {
            ui_error("Unhandled error opening fleet.");
          }
        } else {
          /*  organize_fleet_panes deletes the fleet if it's empty, so no
            *  need to try to do so here.
             */
          close_fleet(fd);
        }
        fd=open_fleet_quiet(targfleet->str,O_RDWR);
        if (fd<0) {
          if (g_fleet_errno==ENOENT) {
            del_fleet(g_Player.coords->str,targfleet,&fd);
          } else {
            ui_error("Unhandled error opening target fleet.");
          }
        } else {
          close_fleet(fd);
        }
      }
    } else {
      printw("%s, %s, %s\n",targfleet->str,g_PlayerName->str,
        g_Player.name->str);
      ui_message("The other fleet's commander does not wish to let you take her ships.");
    }
    g_string_free(targfleet,TRUE);
  }
  dispose_sector(&sect);
}

void organize_fleets(void)
{
  switch (g_Player.locus) {
  case LOCUS_SECTOR:
  case LOCUS_ORBIT:
  case LOCUS_LANDED:
    exchange_ships();
    break;
  case LOCUS_DOCKED:
    if (boardable(g_Player.dock->str)) {
      fleet_organization_panes(g_Player.fleet->str,g_Player.dock->str);
    } else {
      ui_message("The Port Authority ridicules you.");
    }
    break;
  default:
    ui_error("Fleet organization not supported in this locus.\n");
  }
}

void split_fleet(void)
{
  GString		*s;
  int			fd;
  int			secfd;
  struct sector_t	sect;
  struct stat		st;
  switch (g_Player.locus) {
  case LOCUS_SECTOR:
  case LOCUS_LANDED:
  case LOCUS_ORBIT:
    s=read_gstring("Name for new fleet: ");
    if (safe_file_name(s)) {
      g_string_prepend(s,G_DIR_SEPARATOR_S);
      g_string_prepend(s,g_Player.name->str);
      fd=create_fleet(s->str);
      if (fd>=0) {
        close_fleet(fd);
        fleet_organization_panes(s->str,g_Player.fleet->str);
        secfd=open_sector(g_Player.coords->str);
        if (secfd>=0) {
          fd=open_fleet(s->str,O_RDONLY);
          if (fd>=0) {
            if (fstat(fd,&st)==0) {
              if (st.st_size) {
                read_sector(secfd,&sect);
                switch (g_Player.locus) {
                case LOCUS_SECTOR:
                  add_fleet(&sect,s->str);
                  break;
                case LOCUS_LANDED:
                  if (g_Player.orbit<sect.numbodies) {
                    add_landed_fleet(&sect.bodies[g_Player.orbit],s->str);
                  } else {
                    ui_message("Body has disappeared under us!\nThe new fleet has been annihilated.");
                    del_fleet_file(s->str);
                  }
                  break;
                case LOCUS_ORBIT:
                  if (g_Player.orbit<sect.numbodies) {
                    add_orbital_fleet(&sect.bodies[g_Player.orbit],s->str);
                  } else {
                    ui_message("Body has disappeared under us!\nThe new fleet has been annihilated.");
                    del_fleet_file(s->str);
                  }
                  break;
                }
                write_sector(secfd,&sect);
                dispose_sector(&sect);
              } else {
                del_fleet_file(s->str);
              }
            }
            close_fleet(fd);
          }
          close_sector(secfd);
        } /* secfd>=0 */
      } /* fd>=0 */
    } /* valid fleet name(s) */
    g_string_free(s,TRUE);
    break;
  case LOCUS_DOCKED:
    ui_message("Cannot split fleet while docked.");
    break;
  case LOCUS_HYPERSPACE:
    ui_message("Cannot split fleet in hyperspace.");
    break;
  default:
    ui_error("Cannot split fleet in this locus.");
  }
}

void recruit_colonists(char *destfleet, char *srcfleet, int num)
{
  int destfd,srcfd;
  struct fleet_t src,dest;
  struct populii_t itempop,seed;
  int s=0,i,done=0;
  destfd=open_fleet(destfleet,O_RDWR);
  if (destfd>=0) {
    srcfd=open_fleet(srcfleet,O_RDWR);
    if (srcfd>=0) {
      read_fleet(srcfd,&src,NULL);
      read_fleet(destfd,&dest,NULL);
      while (s<src.numships) {
        i=0;
        while (i<src.ships[s].numitems) {
          if (src.ships[s].items[i].class==ICLASS_HABITAT) {
            extract_item_populii(&src.ships[s].items[i],&itempop);
            extract_seed_populii(&itempop,&seed,NULL);
            if (itempop.total) {
              if (itempop.total>num) {
                trim_populii(&itempop,&seed,num);
              }
              move_populii_to_fleet(&dest,&itempop);
              if (itempop.total) {
                absorb_populii(&seed,&itempop);
                done=1;
              }
              dispose_item_popstrs(&src.ships[s].items[i]);
              populate_item(&src.ships[s].items[i],&seed);
              if (done) {
                s=src.numships-1;
                i=src.ships[s].numitems;
              }
            }
            dispose_populii(&itempop);
            dispose_populii(&seed);
          }
          i++;
        }
        s++;
      }
      write_fleet(srcfd,&src);
      write_fleet(destfd,&dest);
      close_fleet(srcfd);
    }
    close_fleet(destfd);
  }
}

void recruit_populace(void)
{
  int numrecruits;
  if (g_Player.locus!=LOCUS_DOCKED) {
    ui_message("You must be docked to recruit colonists.");
    return;
  }
  if (boardable(g_Player.dock->str)) {
    numrecruits=read_int("Number of colonists to recruit: ");
    if (numrecruits>0) {
      recruit_colonists(g_Player.fleet->str,g_Player.dock->str,numrecruits);
    } else {
      ui_message("Nevermind...");
    }
  } else {
    ui_error("Loyalty-check code not yet written.");
  }
}

void train_crew(void)
{
  int fleetfd,ship,item;
  int understaffing=0, vacancies=0;
  int recruited=0;
  int batch;
  struct fleet_t fleet;
  int numtotrain=read_int("Number of crewmembers to train: ");
  if (numtotrain>0) {
    fleetfd=open_fleet(g_Player.fleet->str,O_RDWR);
    if (fleetfd>=0) {
      read_fleet(fleetfd,&fleet,NULL);
      age_fleet(&fleet);
      ship=0;
      while ((ship<fleet.numships) && (numtotrain>0)) {
        item=0;
        while (item<fleet.ships[ship].numitems) {
          if (fleet.ships[ship].items[item].crew<
            fleet.ships[ship].items[item].mincrew) {
            understaffing+=fleet.ships[ship].items[item].mincrew-
              fleet.ships[ship].items[item].crew;
          } else if (fleet.ships[ship].items[item].class!=ICLASS_HABITAT) {
            /* Habitats don't have vacancies the same way other items do. */
            if (fleet.ships[ship].items[item].crew<
              fleet.ships[ship].items[item].maxcrew) {
              vacancies+=fleet.ships[ship].items[item].maxcrew-
                fleet.ships[ship].items[item].mincrew;
            }
          }
          item++;
        }
        ship++;
      }
      if (numtotrain>understaffing+vacancies) {
        numtotrain=understaffing+vacancies;
        ui_message_sds("You only have space for ",numtotrain," crew members.");
      }
      ship=0;
      while ((ship<fleet.numships) && (numtotrain>0)) {
        item=0;
        while (item<fleet.ships[ship].numitems) {
          if (fleet.ships[ship].items[item].class==ICLASS_HABITAT) {
            batch = shanghai_itempop_senescent(
              &fleet.ships[ship].items[item], numtotrain);
	// Make sure shanghai funcs can set ship modded bit!
            numtotrain-=batch;
            recruited+=batch;
          }
          item++;
        }
        ship++;
      }
      ship=0;
      while ((numtotrain>0) && (ship<fleet.numships)) {
        item=0;
        while ((numtotrain>0) && (item<fleet.ships[ship].numitems)) {
          if (fleet.ships[ship].items[item].class==ICLASS_HABITAT) {
            batch = shanghai_itempop_maximal(
              &fleet.ships[ship].items[item], numtotrain);
	// Make sure shanghai funcs can set ship modded bit!
            numtotrain-=batch;
            recruited+=batch;
          }
          item++;
        }
        ship++;
      }
      if (recruited>understaffing) {
        vacancies=recruited-understaffing;
      } else {
        understaffing=recruited;
        vacancies=0;
      }
      ship=0;
      while ((ship<fleet.numships) && (understaffing || vacancies)) {
        item=0;
        while ((item<fleet.ships[ship].numitems) &&
          (understaffing || vacancies)) {
          if (fleet.ships[ship].items[item].class != ICLASS_HABITAT) {
            batch=fleet.ships[ship].items[item].mincrew-
              fleet.ships[ship].items[item].crew;
            if (batch>0) {
              batch=lesser(batch,understaffing);
              understaffing-=batch;
              fleet.ships[ship].items[item].crew+=batch;
              set_item_modified(&fleet,ship,item);
            }
            batch=fleet.ships[ship].items[item].maxcrew-
              fleet.ships[ship].items[item].crew;
            if (batch>0) {
              batch=lesser(batch,vacancies);
              vacancies-=batch;
              fleet.ships[ship].items[item].crew+=batch;
              set_item_modified(&fleet,ship,item);
            }
          }
          item++;
        }
        ship++;
      }
      write_fleet(fleetfd,&fleet);
      dispose_fleet(&fleet);
      close_fleet(fleetfd);
    } else {
      ui_error("Cannot open fleet.");
    }
  }
}

void depopulate_colony(void)
{
  if (g_Player.locus!=LOCUS_DOCKED) {
    ui_message("You need to be docked at a colony to depopulate it.");
  } else {
    ui_error("depopulate_colony() not yet written.");
  }
}

void show_fleet_population(const GString *fleetname)
{
  struct fleet_t fleet;
  struct pop_totals_t totals;
  if (load_fleet(fleetname->str, &fleet)) {
    extract_fleetpop_totals(&totals,&fleet);
    ui_show_pop_totals(&totals);
    dispose_pop_totals(&totals);
    dispose_fleet(&fleet);
  } else {
    ui_error("Error loading fleet.");
  }
}

void population_menu(void)
{
  int notdone=1;
  int c;
  /* Probably should get the totals once, and recalc only if forced. */
  while (notdone) {
    ui_message_ss("Population of fleet ",g_Player.fleet->str);
    show_fleet_population(g_Player.fleet);
    if (g_Player.locus==LOCUS_DOCKED) {
      ui_message_ss("Population of colony ",g_Player.dock->str);
      show_fleet_population(g_Player.dock);
    }
    show_population_menu();
    c=read_char("Command: ");
    switch (toupper(c)) {
    case 'D':
      depopulate_colony();
      break;
    case 'Q':
      notdone=0;
      break;
    case 'R':
      recruit_populace();
      break;
    case 'T':
      train_crew();
      break;
    }
  }
}

void charts_menu(void)
{
  int notdone=1;
  int key;
  int orientation=0;
  int emphasize=EMPH_MIN;
  struct coord_t c;
  string_to_coords(g_Player.coords,&c);
  while (notdone) {
    ui_show_chart(&c,orientation,emphasize);
    key=ui_charts_prompt();
    switch (toupper(key)) {
    case '\t':
      emphasize++;
      if (emphasize>EMPH_MAX) emphasize=EMPH_MIN;
      break;
    case '\r':
    case '\n':
      if (move_to_sector(&c)==0) {
        pause_prompt();
      } else {
        show_curr_location();
        pause_prompt();
      }
      break;
    case '?':
      ui_show_charts_help();
      break;
    case 'Q':
      return;
    case 'S':
      scan_from_sector();
      pause_prompt();
      break;
    case KEY_LEFT:
      if (c.x>0) { c.x--; }
      break;
    case KEY_RIGHT:
      if (c.x<g_Config_Horizon) { c.x++; }
      break;
    case KEY_UP:
      if (c.y>0) { c.y--; }
      break;
    case KEY_DOWN:
      if (c.y<g_Config_Horizon) { c.y++; }
      break;
    case KEY_NPAGE:
      if (c.y<g_Config_Horizon-15) { c.y+=15; }
      break;
    case KEY_PPAGE:
      if (c.y>15) { c.y-=15; }
      break;
    case '-':
      if (c.z>0) { c.z--; }
      break;
    case '+':
      if (c.z<g_Config_Horizon) { c.z++; }
      break;
    }
  }
}

void rename_body(void)
{
  int secfd;
  struct sector_t sect;
  GString *name;
  if ((secfd=open_sector(g_Player.coords->str))>=0) {
    read_sector(secfd,&sect);
    if ((g_Player.orbit<sect.numbodies) &&
      (sect.bodies[g_Player.orbit].name->len==0)) {
      if (g_Player.locus==LOCUS_LANDED) {
        close_sector(secfd);
        name=read_gstring("New name for body: ");
        if ((secfd=open_sector(g_Player.coords->str))>=0) {
          read_sector(secfd,&sect);
          if ((g_Player.orbit<sect.numbodies) &&
            (sect.bodies[g_Player.orbit].name->len==0)) {
            g_string_assign(sect.bodies[g_Player.orbit].name,
              name->str);
            write_sector(secfd,&sect);
          }
        } else {
          ui_error("Error reopening sector.");
        }
        g_string_free(name,TRUE);
      } else {
        ui_message("You must land on a body before you can name it.");
      }
    } else {
      if (g_Player.orbit>=sect.numbodies) {
        ui_message("That body has disappeared from the sector.");
      } else {
        ui_message_ss(sect.bodies[g_Player.orbit].name->str,
          " has already been named.");
      }
    }
    close_sector(secfd);
  } else {
    ui_error("Error opening sector");
  }
}

void read_fd_to_gstrlist(int fd, struct gstrlist_t *list)
{
  FILE *f=fdopen(fd,"r");
  char buf[1024];
  *list=null_gstrlist;
  if (flock(fd,LOCK_SH)==0) {
    if (f) {
      while (fgets(&buf[0],1000,f)) {
        gstrlist_add(list,&buf[0]);
        if ((list->strings[list->num-1]->len) &&
          (list->strings[list->num-1]->str[list->strings[list->num-1]->len-1]=='\n')) {
          g_string_truncate(list->strings[list->num-1],
            list->strings[list->num-1]->len-1);
        }
      }
      fclose(f);
    }
    flock(fd,LOCK_UN);
  }
}

void view_public_log()
{
  int fd;
  struct gstrlist_t list;
  GString *filename=g_string_new(g_Config.datapath->str);
  g_string_append(filename,MAKROKOSMOS_LOG_PUBLIC);
  fd=open(filename->str,O_RDONLY);
  if (fd>=0) {
    read_fd_to_gstrlist(fd,&list);
    close(fd);
  }
  add_log_to_chart(&list,&g_Player.logread);
  show_gstrlist("Log:",&list);
  dispose_gstrlist(&list);
  g_string_free(filename,TRUE);
}

void describe_planet(void)
{
  int	secfd;
  struct sector_t sect=null_sector;
  GString *origdesc;
  GString *newdesc;
  load_sector(g_Player.coords->str,&sect);
  if (g_Player.orbit>=sect.numbodies) {
    ui_message("The body you are on has vanished.");
    dispose_sector(&sect);
    return;
  }
  origdesc=g_string_new(sect.bodies[g_Player.orbit].desc->str);
  dispose_sector(&sect);
  newdesc=ui_edit_block("Editing body description.", origdesc,
    MAKROKOSMOS_MAX_DESCLEN);
  secfd=open_sector(g_Player.coords->str);
  if (secfd>=-1) {
    read_sector(secfd,&sect);
    if (gstreq(sect.bodies[g_Player.orbit].desc,origdesc)) {
      g_string_assign(sect.bodies[g_Player.orbit].desc,newdesc->str);
      write_sector(secfd,&sect);
    } else {
      ui_message_ss("Description has changed out from under us, to\n",
       sect.bodies[g_Player.orbit].desc->str);
    }
    close_sector(secfd);
  } else {
    ui_error("Error opening sector.");
  }
  g_string_free(newdesc,TRUE);
  g_string_free(origdesc,TRUE);
}

void describe_colony(void)
{
  int colfd;
  struct fleet_t col=null_fleet;
  GString *origdesc;
  GString *newdesc;
  if (load_fleet(g_Player.dock->str,&col)) {
    origdesc=g_string_new(col.desc->str);
    dispose_fleet(&col);
    newdesc=ui_edit_block("Editing colony description.", origdesc, 
      MAKROKOSMOS_MAX_DESCLEN);
    colfd=open_fleet(g_Player.dock->str, O_RDWR);
    if (colfd>=0) {
      read_fleet(colfd, &col, NULL);
      if (gstreq(col.desc, origdesc)) {
        g_string_assign(col.desc,newdesc->str);
        write_fleet(colfd, &col);
      } else {
        ui_message_ss("Description has changed out from under us, to\n",
          col.desc->str);
      }
      close_fleet(colfd);
    }
    g_string_free(origdesc,TRUE);
    g_string_free(newdesc,TRUE);
  }
}

void describe_locus(void)
{
  switch (g_Player.locus) {
  case LOCUS_LANDED:
    describe_planet();
    break;
  case LOCUS_DOCKED:
    describe_colony();
    break;
  default:
    ui_message("The current location is indescribable.");
  }
}

void handle_input(void)
{
  int targsat;
  int numsats;
  int *satellite_list;
  int c;
  struct sector_t sect;
  attrset(COLOR_PROMPT);
  printw("Command (%d turns) [? for help]: ",g_Player.turns);
  refresh();
  c=getch();
  attrset(COLOR_INPUT);
  printw("%c\n\n",c);
  switch (toupper(c)) {
  case ERR:
    ui_error("ERR returned on keypress.");
    break;
  case 0x04:
  case 0x05:
    describe_locus();
    break;
  case 0x06:
    split_fleet();
    break;
  case 0xC:
    clear();
    break;
  case '0':
  case 0xF: /* ^O */
    claim_ownership();
    break;
  case ':':
    send_chat();
    break;
  case '?':
    show_help();
    break;
  case 'A':
    attack_fleet();
    break;
  case 'B':
    switch (g_Player.locus) {
    case LOCUS_SECTOR:
    case LOCUS_ORBIT:
    case LOCUS_LANDED:
      board_fleet();
      break;
    case LOCUS_DOCKED:
      build_menu();
      break;
    }
    break;
  case 'C':
    switch (g_Player.locus) {
    case LOCUS_LANDED:
      colonize();
      break;
    default:
      charts_menu();
    }
    break;
  case 'D':
    dock_fleet();
    break;
  case 'E':
    equip_menu();
    break;
  case 'F':
    organize_fleets();
    break;
  case 'H':
    view_public_log();
    break;
  case 'L':
    switch (g_Player.locus) {
    case LOCUS_SECTOR:
      printw("Cannot land - not orbiting anything.\n");
      break;
    case LOCUS_ORBIT:
      printw("Landing...\n");
      move_ofleet_to_land(g_Player.coords->str,g_Player.fleet,g_Player.orbit);
      g_Player.locus=LOCUS_LANDED;
      break;
    case LOCUS_LANDED:
      printw("You congratulate yourself on your excellent piloting\nand wonder why you keep hearing giggling whenever your back is turned.\n");
      break;
    case LOCUS_DOCKED:
      printw("Your fleet lifts off from the colony and settles down again in a deserted area.\n");
      g_Player.locus=LOCUS_LANDED;
      break;
    case LOCUS_HYPERSPACE:
      leave_hyperspace();
      break;
    default:
      ui_error("Unknown locus.  Cannot land.");
    }
    refresh();
    break;
  case 'M':
    move_to_sector(NULL);
    break;
  case 'N':
    rename_body();
    break;
  case 'O':
    switch (g_Player.locus) {
    case LOCUS_SECTOR:
      load_sector(g_Player.coords->str,&sect);
      show_satellite_list(&sect,-1,&satellite_list,&numsats);
      if (g_Player.turns<=0) {
        ui_message("Insufficient turns to enter orbit.");
        break;
      }
      if (numsats>1) {
        targsat=read_int("Enter orbit around: ");
        if ((targsat>=1) && (targsat<numsats)) {
          if (satellite_list[targsat]>=0) {
            move_sfleet_to_orbit(g_Player.coords->str,
              g_Player.fleet,satellite_list[targsat]);
            g_Player.orbit=satellite_list[targsat];
            g_Player.locus=LOCUS_ORBIT;
            g_Player.turns--;
          } else {
            printw("Invalid input.\n");
            refresh();
          }
        } else {
          ui_message("Invalid input.\n");
        }
      } else {
        ui_message("Nothing orbitable in this sector.");
        refresh();
      }
      dispose_sector(&sect);
      break;
    case LOCUS_ORBIT:
      load_sector(g_Player.coords->str,&sect);
      show_satellite_list(&sect,g_Player.orbit,&satellite_list,&numsats);
      if (g_Player.turns<=0) {
        ui_message("Insufficient turns to change orbit.\n");
        break;
      }
      targsat=read_int("Enter orbit around: ");
      if ((targsat>=0) && (targsat<numsats)) {
        if (satellite_list[targsat]>=0) {
          move_ofleet_to_orbit(g_Player.coords->str,g_Player.orbit,
            g_Player.fleet,satellite_list[targsat]);
          g_Player.orbit=satellite_list[targsat];
          g_Player.turns--;
        } else if (satellite_list[targsat]==-1) {
          ui_message("Leaving orbit");
          move_ofleet_to_sector(g_Player.coords->str,g_Player.orbit,
            g_Player.fleet);
          g_Player.orbit=0;
          g_Player.locus=LOCUS_SECTOR;
          g_Player.turns--;
        } else {
          ui_error("Logic error in orbit change.");
        }
      }
      g_free(satellite_list);
      dispose_sector(&sect);
      break;
    case LOCUS_LANDED:
    case LOCUS_DOCKED:
      if (g_Player.turns>0) {
        move_lfleet_to_orbit(g_Player.coords->str,g_Player.fleet,g_Player.orbit);
        g_Player.locus=LOCUS_ORBIT;
        printw("Taking off.\n");
        g_Player.turns--;
      } else {
        ui_message("Insufficient turns to lift off.");
      }
      break;
    case LOCUS_HYPERSPACE:
      break;
    default:
      printw("Unknown locus %d.  Cannot orbit.\n",g_Player.locus);
    }
    break;
  case 'P': 
    population_menu();
    break;
  case 'Q':
    switch (read_char("Are you sure you want to quit? ")) {
      case 'Y':
      case 'y':
        shutdown_game();
    }
    printw("\n");
    break;
  case 'R':
    repair();
    break;
  case 'S':
    scan();
    break;
  case 'T':
    show_current_time();
    break;
  default:
    printw("Invalid command.\n");
  }
  refresh();
}

void main_loop(void)
{
  int notdone=1;
  while (notdone) {
    show_curr_location();
    check_chat();
    handle_input();
  }
}

int main(int argc, char **argv)
{
  // Init interface.
  initialize();
  main_loop();  
  return 255;
}
