#include <stdio.h>
#include <ncurses.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include "microcosm.h"
#include "util.h"
#include "sectors.h"
#include "populii.h"
#include "ui_all.h"

void add_popstr_to_item(struct item_t *i, char *popstr)
{
  i->numpopstrs++;
  i->popstrs=g_realloc(i->popstrs,sizeof(GString *)*i->numpopstrs);
  i->popstrs[i->numpopstrs-1]=g_string_new(popstr);
}

void populate_item(struct item_t *i, struct populii_t *p)
{
  int x=0;
  int y,z;
  GString *s;
  char scratch[128];
  if (p->total<=0) {
    return;
  }
  scratch[127]='\0';
  while (x<p->numpops) {
    s=g_string_new("");
    g_string_sprintf(s,"S%s:",p->pops[x].species.name->str);
    y=0;
    while (y<p->pops[x].species.sexes) {
      snprintf(&scratch[0],127,"%dX",y);
      g_string_append(s,&scratch[0]);
      z=0;
      while (z<=p->pops[x].species.senescence) {
        if (p->pops[x].num[y][z]) {
          snprintf(&scratch[0],127,"%d=%d;",z,p->pops[x].num[y][z]);
          g_string_append(s,&scratch[0]);
        }
        if (s->len>900) {
          add_popstr_to_item(i,s->str);
          g_string_free(s,TRUE);
          s=g_string_new("");
          g_string_sprintf(s,"S%s:%dX",p->pops[x].species.name->str,y);
        }
        z++;
      }
      y++;
    }
    if ((s->len) && (p->pops[x].total)) {
      add_popstr_to_item(i,s->str);
    }
    g_string_free(s,TRUE);
    x++;
  }
  i->modified=1;
  i->updated=p->lastupdated;
}

void dispose_species(struct species_t *s)
{
  g_string_free(s->name,TRUE);
  g_string_free(s->genders,TRUE);
  g_string_free(s->habitat,TRUE);
}

void dispose_pop(struct pop_t *p)
{
  int x=0;
  while (x>p->species.sexes) {
    g_free(p->num[x]);
    x++;
  }
  g_free(p->num);
  dispose_species(&p->species);
}

void dispose_populii(struct populii_t *p)
{
  int x=0;
  while (x<p->numpops) {
    dispose_pop(&p->pops[x]);
    x++;
  }
  g_free(p->pops);
  p->pops=NULL;
  p->total=0;
  p->numpops=0;
  p->lastupdated=0;
}

void dispose_item_popstrs(struct item_t *i)
{
  int x=0;
  while (x<i->numpopstrs) {
    g_string_free(i->popstrs[x++],TRUE);
  }
  g_free(i->popstrs);
  i->numpopstrs=0;
  i->popstrs=NULL;
}

void init_demographics(struct pop_t *p)
{
  int x=0;
  int y;
  p->num=g_malloc(sizeof(int *)*p->species.sexes);
  while (x<p->species.sexes) {
    p->num[x]=g_malloc(sizeof(int *)*(p->species.senescence+1));
    y=0;
    while (y<=p->species.senescence) {
      p->num[x][y]=0;
      y++;
    }
    x++;
  }
}


void load_species(char *str, struct species_t *s)
{
  FILE *f;
  GString *filename=g_string_new(g_Config.datapath->str);
  GString *args,*cmd;
  char	buf[1024];
  int	len;
  g_string_append(filename,MAKROKOSMOS_SPECIES_PATH);
  g_string_append(filename,str);
  g_string_assign(s->name,str);
  f=fopen(filename->str,"r");
  if (f) {
    while (fgets(&buf[0],1024,f)) {
      len=strlen(&buf[0]);
      if ((len) && (buf[len-1]=='\n')) {
        buf[len-1]='\0';
      }
      if (buf[0]!='#') {
        split(&buf[0],&cmd,&args);
        if (strcmp(cmd->str,"Genders")==0) {
          g_string_assign(s->genders,args->str);
        } else if (strcmp(cmd->str,"Habitat")==0) {
          g_string_assign(s->habitat,args->str);
        } else if (strcmp(cmd->str,"HP")==0) {
          s->hp=atoi(args->str);
        } else if (strcmp(cmd->str,"MaxPress")==0) {
          s->maxpress=atof(args->str);
        } else if (strcmp(cmd->str,"MaxTemp")==0) {
          s->maxtemp=atoi(args->str);
        } else if (strcmp(cmd->str,"Menopause")==0) {
          s->menopause=atoi(args->str);
        } else if (strcmp(cmd->str,"MinPress")==0) {
          s->minpress=atof(args->str);
        } else if (strcmp(cmd->str,"MinTemp")==0) {
          s->mintemp=atoi(args->str);
        } else if (strcmp(cmd->str,"Mortality")==0) {
          s->mortality=atof(args->str);
        } else if (strcmp(cmd->str,"Puberty")==0) {
          s->puberty=atoi(args->str);
        } else if (strcmp(cmd->str,"ReproduceNum")==0) {
          s->reproducenum=atoi(args->str);
        } else if (strcmp(cmd->str,"ReproduceOdds")==0) {
          s->reproduceodds=atof(args->str);
        } else if (strcmp(cmd->str,"Senescence")==0) {
          s->senescence=atoi(args->str);
        } else if (strcmp(cmd->str,"Sexes")==0) {
          s->sexes=atoi(args->str);
        } else {
          printw("Incomprehensible species line '%s'\n",&buf[0]);
          refresh();
        }
        g_string_free(cmd,TRUE);
        g_string_free(args,TRUE);
      }
    }
    fclose(f);
  } else {
    ui_error("Error opening species file.");
  }
}

int populacenum_c(struct populii_t *p, GString *species) {
  int x=0;
  while (x<p->numpops) {
    if (gstreq(species,p->pops[x].species.name)) {
      return x;
    }
    x++;
  }
  p->numpops++;
  p->pops=g_realloc(p->pops,sizeof(struct pop_t)*p->numpops);
  p->pops[p->numpops-1]=null_pop;
  load_species(species->str,&p->pops[p->numpops-1].species);
  init_demographics(&p->pops[p->numpops-1]);
  return p->numpops-1;
}

void extract_item_populii(struct item_t *i, struct populii_t *p)
{
  int x=0;
  int y;
  int a=0;
  int cycle;
  int sex;
  int species=-1;
  GString *speciesname;
  *p=null_populii;
  p->lastupdated=i->updated;
  if (i->popstrs) {
    while (x<i->numpopstrs) {
      y=0;
      while (y<=i->popstrs[x]->len) {
        switch (i->popstrs[x]->str[y]) {
        case '0' ... '9':
          a=(a*10)+(i->popstrs[x]->str[y]-'0');
          break;
        case ';':
          p->pops[species].num[sex][cycle]+=a;
          p->pops[species].total+=a;
          a=0;
          break;
        case '=':
          cycle=a;
          a=0;
          break;
        case 'S':
          a=y+1;
          while ((i->popstrs[x]->str[++y]!=':') && (y<i->popstrs[x]->len)) {
            /* Sanity check to ensure we don't iterate off the end. */
          }
          i->popstrs[x]->str[y]='\0';
          speciesname=g_string_new(&i->popstrs[x]->str[a]);
          species=populacenum_c(p,speciesname);
          g_string_free(speciesname,TRUE);
          i->popstrs[x]->str[y]=':';
          a=0;
          break;
        case 'X':
          if (a<p->pops[x].species.sexes) {
            sex=a;
            a=0;
          } else {
            printw("Invalid sex number %d (>%d) in population string.\n",
              a,p->pops[x].species.sexes);
            sex=0;
            a=0;
          }
          break;
        case '\0': /* Don't whine about nulls. */
          break;
        default:
          printw("Unknown character %c in population string.\n",
            i->popstrs[x]->str[y]);
	  /* Cheesy hack to keep negative numbers from screwing with
		population growth much. */
          cycle=p->pops[species].species.senescence;
          refresh();
        }
        y++;
      }
      x++;
    }
  }
  x=0;
  while (x<p->numpops) {
    p->total+=p->pops[x].total;
    x++;
  }
}


int fertilesets(struct pop_t *p)
{
  int sex=0;
  int x=0;
  int num[20];
  int repeats=1;
  int lowest;
  if (p->species.sexes>20) {
    ui_error(">20 sexes in your species.  fertilesets() returns 0.");
    return 0;
  }
  sex=0;
  while (sex<p->species.sexes) {
    int age=p->species.puberty;
    num[sex]=0;
    while (age<p->species.menopause) {
      num[sex]+=p->num[sex][age++];
    }
    sex++;
  }
  sex=0;
  while (x<p->species.reproducenum) {
    x++;
    sex++;
    if (sex>p->species.sexes) {
      sex=0;
      repeats++;
    }
  }
  x=0;
  while (x<sex) {
    num[x]/=repeats;
    x++;
  }
  while (x<p->species.sexes) {
    if (x<p->species.reproducenum) {
      num[x]/=(repeats-1);
    }
    x++;
  }
  sex=0;
  lowest=num[0];
  x=p->species.sexes < p->species.reproducenum ? p->species.sexes : p->species.reproducenum;
  while (sex<x) {
    lowest=lowest<num[sex] ? lowest : num[sex];
    sex++;
  }
  return lowest;
}

void fertile_dieback(struct pop_t *p, int *total)
{
  int x=0;
  int y;
  int new;
  while (x<p->species.sexes) {
    y=p->species.menopause;
    while (--y>=p->species.puberty) {
      if (p->num[x][y]) {
        if (p->num[x][y]<10) {
          if (probability(p->species.mortality)) {
            *total-=p->num[x][y];
            p->total-=p->num[x][y];
            p->num[x][y]=0;
          }
        } else {
          new=p->num[x][y]*(1-p->species.mortality);
          *total-=p->num[x][y]-new;
          p->total-=p->num[x][y]-new;
          p->num[x][y]=new;
        }
      }
    }
    x++;
  }
  if (*total<0) {
    *total=0;
  }
}

void elder_dieback(struct pop_t *p, int *total)
{
  int x=0;
  int y;
  while (x<p->species.sexes) {
    y=p->species.senescence;
    while (--y>=p->species.menopause) {
      if (p->num[x][y]) {
        if (probability(p->species.mortality)) {
          *total-=p->num[x][y];
          p->total-=p->num[x][y];
          p->num[x][y]=0;
        }
      }
    }
    x++;
  }
  if (*total<0) {
    *total=0;
  }
}

void senescent_dieback(struct pop_t *p, int *total)
{
  int x=0;
  while (x<p->species.sexes) {
    if (probability(p->species.mortality)) {
      *total-=p->num[x][p->species.senescence];
      p->total-=p->num[x][p->species.senescence];
      p->num[x][p->species.senescence]=0;
    }
    x++;
  }
  if (*total<0) {
    *total=0;
  }
}

void youth_dieback(struct pop_t *p, int *total)
{
  int x=0;
  int y;
  while (x<p->species.sexes) {
    y=p->species.puberty;
    while (--y>=0) {
      if (probability(p->species.mortality)) {
	/* Should probably add the protection code from fertile_dieback() */
        *total-=p->num[x][y];
        p->total-=p->num[x][y];
        p->num[x][y]=0;
      }
    }
    x++;
  }
}

void dieback(struct pop_t *p, int goalpop, int *total)
{
  while (p->total>goalpop) {
    senescent_dieback(p,total);
    if (*total<=goalpop) {
      return;
    }
    senescent_dieback(p,total);
    elder_dieback(p,total);
    if (*total<=goalpop) {
      return;
    }
    senescent_dieback(p,total);
    elder_dieback(p,total);
    fertile_dieback(p,total);
    if (*total<=goalpop) {
      return;
    }
    senescent_dieback(p,total);
    elder_dieback(p,total);
    fertile_dieback(p,total);
    youth_dieback(p,total);
    if (*total<=goalpop) {
      return;
    }
  }
}

void age_populace(struct pop_t *p, int *total)
{
  int sex=0;
  int age;
  while (sex<p->species.sexes) {
    age=p->species.senescence;
    p->total-=p->num[sex][age];
    *total-=p->num[sex][age];
    while (age>0) {
      p->num[sex][age]=p->num[sex][age-1];
      age--;
    }
    sex++;
  }
  sex=0;
  while (sex<p->species.sexes) {
    p->num[sex][0]=0;
    sex++;
  }
}

void age_populii(struct populii_t *p, int *total)
{
  int x=0;
  while (x<p->numpops) {
    age_populace(&p->pops[x],total);
    x++;
  }
}

void grow_populace(struct pop_t *p, int amount)
{
  int y=0;
  while (y<amount) {
    p->num[randval_i(p->species.sexes)][0]++;
    y++;
  }
}

void shanghai_crew(struct pop_t *p, struct ship_t *s, int goalpop, int *total)
{
  int x=0;
  int age=p->species.senescence;
  int sex=0;
  int diff;
  while (x<s->numitems) {
    if (s->items[x].class!=ICLASS_HABITAT) {
      while (s->items[x].crew<s->items[x].maxcrew) {
        if (p->num[sex][age]) {
          diff=s->items[x].maxcrew-s->items[x].crew;
          if (p->num[sex][age]>diff) {
            s->items[x].crew+=diff;
            p->num[sex][age]-=diff;
            *total-=diff;
          } else {
            s->items[x].crew+=p->num[sex][age];
            *total-=p->num[sex][age];
            p->num[sex][age]=0;
          }
          if (*total<=goalpop) {
            return;
          }
          if (s->items[x].crew>=s->items[x].maxcrew) {
            goto itemfull;
          }
        }
        sex++;
        if (sex>=p->species.sexes) {
          sex=0;
          age--;
          if (age<=p->species.menopause) {
            goto shanghai_breeders;
          }
        }
      }
    }
    itemfull:
    x++;
  }
  shanghai_breeders:
  age=p->species.menopause;
  sex=0;
  while (x<s->numitems) {
    if (s->items[x].class!=ICLASS_HABITAT) {
      while (s->items[x].crew<s->items[x].maxcrew) {
        if (p->num[sex][age]) {
          diff=lesser(*total-goalpop,s->items[x].maxcrew-s->items[x].crew);
          if (p->num[sex][age]>diff) {
            s->items[x].crew+=diff;
            p->num[sex][age]-=diff;
            *total-=diff;
          } else {
            s->items[x].crew+=p->num[sex][age];
            *total-=p->num[sex][age];
            p->num[sex][age]=0;
          }
          if (*total<=goalpop) {
            return;
          }
          if (s->items[x].crew>=s->items[x].maxcrew) {
            goto itemfull2;
          }
        }
        sex++;
        if (sex>=p->species.sexes) {
          sex=0;
          age--;
          if (age<=p->species.puberty) {
            return;
          }
        }
      }
    }
    itemfull2:
    x++;
  }
}

int total_populii(struct populii_t *p)
{
  int x=0;
  int total=0;
  while (x<p->numpops) {
    total+=p->pops[x].total;
    x++;
  }
  return total;
}

int resum_total_populii(struct populii_t *p)
{
  int pop=0;
  int age;
  int sex;
  while (pop<p->numpops) {
    sex=0;
    p->pops[pop].total=0;
    while (sex<p->pops[pop].species.sexes) {
      age=0;
      while (age<=p->pops[pop].species.senescence) {
        if (p->pops[pop].num[sex][age]<0) {
          p->pops[pop].num[sex][age]=0;
          ui_message("Negative number found in population.  Fixing.");
        }
        p->pops[pop].total+=p->pops[pop].num[sex][age];
        age++;
      }
      sex++;
    }
    pop++;
  }
  return total_populii(p);
}

void breed_to_time(struct populii_t *p, int maxpop, struct ship_t *s,
  int time)
{
  int total=0;
  int x=0;
  int y,z,totalgrowth=0,okgrowth;
  int sets;
  int overpopulated;
  int *growth=g_malloc(sizeof(int)*p->numpops);
  if (p->numpops<=0) {
    return;
  }
  total=total_populii(p);
  if (total<0) {
    total=resum_total_populii(p);
  }
//  if (p->lastupdated+(MAKROKOSMOS_MAX_AGING/8)<time) {
//    ui_message("Calculating population growth.  This may take a while.");
//    if (p->lastupdated+MAKROKOSMOS_MAX_AGING<time) {
//      ui_message("Population has been abandoned for a long while...");
//      p->lastupdated=time-MAKROKOSMOS_MAX_AGING;
//    }
//  }
  while (p->lastupdated<time) {
    /* Aging */
    age_populii(p,&total);
    /* Now determine max possible growth */
    x=0;
    totalgrowth=0;
    while (x<p->numpops) {
      sets=fertilesets(&p->pops[x]);
      if ((int)(sets*p->pops[x].species.reproduceodds)) {
        growth[x]=sets;
        totalgrowth+=sets;
      } else {
        growth[x]=0;
        y=0;
        while (y++<sets) {
          if (probability(p->pops[x].species.reproduceodds)) {
            growth[x]++;
          }
        }
        totalgrowth+=growth[x];
      }
      x++;
    }
    /* How much of that is allowed? */
    if (totalgrowth+total>maxpop) {
      if (total>maxpop) {
        x=0;
        while (x<p->numpops) {
          shanghai_crew(&p->pops[x],s,maxpop-growth[x],&total);
          dieback(&p->pops[x],greater(maxpop-growth[x],maxpop/10),&total);
          x++;
        }
        if (total<0) {
          total=resum_total_populii(p);
        }
      }
      okgrowth=(maxpop-total)/p->numpops;
      overpopulated=1;
    } else {
      overpopulated=0;
      okgrowth=maxpop-total;
    }
    x=0;
    while (x<p->numpops) {
      z=lesser(growth[x],okgrowth);
      grow_populace(&p->pops[x],z);
      p->pops[x].total+=z;
      growth[x]-=z;
      total+=z;
      x++;
    }
    /* Now handle die-offs for overpopulation */
    if (overpopulated) {
      x=0;
      while (x<p->numpops) {
        shanghai_crew(&p->pops[x],s,maxpop-growth[x],&total);
        dieback(&p->pops[x],greater(maxpop-growth[x],maxpop/10),&total);
        x++;
      }
    }
    p->lastupdated++;
  }
  g_free(growth);
}

void species_assign(struct species_t *dest, struct species_t *src)
{
  dispose_species(dest);
  *dest=*src;
  dest->name=g_string_new(src->name->str);
  dest->genders=g_string_new(src->genders->str);
  dest->habitat=g_string_new(src->habitat->str);
}

void add_populace(struct populii_t *p, struct species_t *s)
{
  p->numpops++;
  p->pops=g_realloc(p->pops,p->numpops*sizeof(struct pop_t));
  p->pops[p->numpops-1]=null_pop;
  species_assign(&p->pops[p->numpops-1].species,s);
  init_demographics(&p->pops[p->numpops-1]);
}

int extract_seed_population_default(struct pop_t *src, struct pop_t *seed)
{
  int numextracted=0;
  int sex=0;
  int age;
  int *seedpop=g_malloc(sizeof(int)*src->species.reproducenum);
  while (sex<src->species.reproducenum) {
    age=src->species.puberty;
    while (age<src->species.menopause) {
      if (src->num[sex][age]) {
        seedpop[sex]=age;
        numextracted++;
/*        printw("Debug: %d-cycle-old member of sex %d in seed.\n", 
          age,sex); */
        refresh();
        goto gotone;
      }
      age++;
    }
    age=src->species.puberty;
    while (--age>=0) {
      if (src->num[sex][age]) {
        numextracted++;
        seedpop[sex]=age;
/*        printw("Debug: %d-cycle-old member of sex %d in seed.\n",
          age,sex); */
        refresh();
        goto gotone;
      }
    }
    gotone:
    sex++;
  }
  if (numextracted==src->species.reproducenum) {
    sex=0;
    while (sex<src->species.reproducenum) {
      seed->num[sex][seedpop[sex]]++;
      src->num[sex][seedpop[sex]]--;
      src->total--;
      seed->total++;
      sex++;
    }
    g_free(seedpop);
    return TRUE;
  }
  g_free(seedpop);
  return FALSE;
}

int absorb_pop_maxage(struct pop_t *dest, struct pop_t *src, int maxage)
{
  int age;
  int sex=0;
  if (src->species.sexes!=dest->species.sexes) {
    ui_error("Attempted to absorb a population with a different number of sexes.");
    return 0;
  }
  while (sex<dest->species.sexes) {
    age=0;
    while (age<maxage) {
      int amt;
      if ((amt=src->num[sex][age])) {
        dest->num[sex][age]+=amt;
        dest->total+=amt;
        src->total-=amt;
        src->num[sex][age]=0;
      }
      age++;
    }
    sex++;
  }
  return 1;
}

int extract_seed_population_clusterfuck(struct pop_t *src, struct pop_t *seed)
{
  int sex=0;
  int num=0;
  int age;
  int *needed=g_malloc(sizeof(int)*src->species.sexes);
  while (sex<src->species.sexes) {
    needed[sex]=0;
    sex++;
  }
  sex=0;
  while (num<src->species.reproducenum) {
    needed[sex]++;
    num++;
    sex++;
    if (sex>=src->species.sexes) {
      sex=0;
    }
  }
  num=0;
  sex=0;
  while (sex<src->species.sexes) {
    age=src->species.puberty;
    while (age<src->species.menopause) {
      if (src->num[sex][age]) {
        int amt = needed[sex]<src->num[sex][age] ?
          needed[sex] : src->num[sex][age];
        seed->num[sex][age]+=amt;
        src->num[sex][age]-=amt;
/*        src->total-=amt;*/
        seed->total+=amt;
        needed[sex]-=amt;
        if (needed[sex]<=0) {
          age=src->species.menopause;
        }
      }
      age++;
    }
    sex++;
  }
  sex=0;
  while (sex<src->species.sexes) {
    if (needed[sex]>0) {
      age=0;
      while (age<src->species.puberty) {
        if (src->num[sex][age]) {
          int amt = needed[sex]<src->num[sex][age] ?
            needed[sex] : src->num[sex][age];
          seed->num[sex][age] += amt;
          seed->total+=amt;
          src->total-=amt;
          src->num[sex][age] -= amt;
          needed[sex]-=amt;
          if (needed[sex]<=0) {
            age=src->species.puberty;
          }
        }
        age++;
      }
      if (needed[sex]>=0) {
        if (seed->total>seed->species.reproducenum) {
          ui_error("WARNING: Seed exceeds required size.");
        }
        absorb_pop_maxage(src,seed,src->species.menopause);
        g_free(needed);
        return FALSE;
      }
    }
    sex++;
  }
  src->total-=seed->total;
  return TRUE;
}

int extract_seed_population(struct pop_t *src, struct pop_t *seed)
{
  if (src->species.sexes>=src->species.reproducenum) {
    return extract_seed_population_default(src,seed);
  }
  return extract_seed_population_clusterfuck(src,seed);
}

void extract_half_populii(struct populii_t *src, struct populii_t *dest,
  struct gstrlist_t *specie)
{
  int species=0;
  int flop=0;
  int age;
  int sex;
  int destpop;
  int amt=0;
  *dest=null_populii;
  while (species<src->numpops) {
    if (gstrlistindex(specie,src->pops[species].species.name)>=0) {
      destpop=populacenum_c(dest,src->pops[species].species.name);
      sex=0;
      while (sex<src->pops[species].species.sexes) {
        age=0;
        while (age<=src->pops[species].species.senescence) {
          if (src->pops[species].num[sex][age]) {
            amt=src->pops[species].num[sex][age];
            if (amt&1) {
              amt += flop;
              flop ^= 2;
            }
            amt>>=1;
            dest->pops[destpop].num[sex][age]+=amt;
            src->pops[species].num[sex][age]-=amt;
            dest->pops[destpop].total+=amt;
            dest->total+=amt;
          }
          age++;
        }
        sex++;
      }
      src->pops[species].total-=dest->pops[destpop].total;
    }
    species++;
  }
  src->total-=dest->total;
}

void del_populace(struct populii_t *p, GString *species)
{
  int x=0;
  while (x<p->numpops) {
    if (gstreq(p->pops[x].species.name,species)) {
      p->numpops--;
      p->pops[x]=p->pops[p->numpops];
      p->pops=g_realloc(p->pops,p->numpops*sizeof(struct pop_t));
    }
    x++;
  }
}

void extract_seed_populii(struct populii_t *src, struct populii_t *seed,
	struct gstrlist_t *specie)
{
  int x=0;
  int species;
  *seed=null_populii;
  while (x<src->numpops) {
    if (specie && (gstrlistindex(specie,src->pops[x].species.name)>=0)) {
      species=populacenum_c(seed,src->pops[x].species.name);
      if (extract_seed_population(&src->pops[x],&seed->pops[species])) {
        src->total-=seed->pops[species].total;
        seed->total+=seed->pops[species].total;
      } else {
        del_populace(seed,src->pops[x].species.name);
      }
    }
    x++;
  }
  seed->lastupdated=src->lastupdated;
}

void absorb_pop_limited(struct populii_t *dest, struct pop_t *src,
  int max)
{
  int age,sex;
  int spec=populacenum_c(dest,src->species.name);
  int curr=dest->total;
  if (dest->total+src->total<=max) {
    age=0;
    sex=0;
    while (sex<src->species.sexes) {
      age=0;
      while (age<src->species.senescence) {
        dest->pops[spec].num[sex][age]+=src->num[sex][age];
        src->num[sex][age]=0;
        age++;
      }
      sex++;
    }
    dest->total+=src->total;
    dest->pops[spec].total+=src->total;
    src->total=0;
  } else {
    age=0;
    src->total-=max-dest->total;
    while (age<src->species.senescence) {
      sex=0;
      while (sex<src->species.sexes) {
        if (curr+src->num[sex][age]>max) {
          dest->pops[spec].num[sex][age]+=max-curr;
          src->num[sex][age]-=max-curr;
          curr+=max-curr;
        } else {
          dest->pops[spec].num[sex][age]+=src->num[sex][age];
          curr+=src->num[sex][age];
          src->num[sex][age]=0;
        }
        sex++;
      }
      age++;
    }
    dest->pops[spec].total+=max-dest->total;
    dest->total=max;
  }
}

void absorb_item_populii_specie(struct populii_t *dest, struct item_t *src,
  const struct gstrlist_t *specie, int max)
{
  struct populii_t	itempop;
  int			x=0;
  extract_item_populii(src,&itempop);
  while (x<itempop.numpops) {
    if (gstrlistindex(specie,itempop.pops[x].species.name)>=0) {
      absorb_pop_limited(dest,&itempop.pops[x],max);
    }
    x++;
  }
  dispose_item_popstrs(src);
  populate_item(src,&itempop);
  dispose_populii(&itempop);
}

void absorb_populii(struct populii_t *dest, struct populii_t *src)
{
  int x=0;
  int species,y,z;
  while (x<src->numpops) {
    species=populacenum_c(dest,src->pops[x].species.name);
    y=0;
    while (y<dest->pops[species].species.sexes) {
      z=0;
      while (z<=dest->pops[species].species.senescence) {
        dest->pops[species].num[y][z]+=src->pops[x].num[y][z];
        src->pops[x].num[y][z]=0;
        z++;
      }
      y++;
    }
    dest->pops[species].total+=src->pops[x].total;
    x++;
  }
  dest->total+=src->total;
}

void absorb_populii_limited(struct populii_t *dest, struct populii_t *src,
  int max)
{
  if (dest->total+src->total<=max) {
    absorb_populii(dest,src);
    dispose_populii(src);
  } else {
    int x=0;
    int d;
    while (x<src->numpops) {
      d=src->pops[x].total;
      absorb_pop_limited(dest,&src->pops[x],max);
      src->total-=(d-src->pops[x].total);
      x++;
    }
  }
}

int breedable_default(struct pop_t *p)
{
  int sex=0;
  int age;
  int notfound;
  while (sex<p->species.reproducenum) {
    notfound=1;
    age=0;
    while (age<p->species.menopause) {
      if (p->num[sex][age]) {
        notfound=0;
        age=p->species.menopause;
      }
      age++;
    }
    if (notfound) {
      return 0;
    }
    sex++;
  }
  return 1;
}

int breedable_clusterfuck(struct pop_t *p)
{
  int needed[20];
  int x=0;
  int sex=0;
  if (p->species.sexes>=20) {
    ui_error("Invalid sex count >20.");
    return 0;
  }
  while (x<p->species.sexes) {
    needed[x++]=0;
  }
  x=0;
  while (x<p->species.reproducenum) {
    needed[sex]++;
    x++;
    sex++;
    if (sex>=p->species.sexes) {
      sex=0;
    }
  }
  sex=0;
  while (sex<p->species.sexes) {
    int age=0;
    while (age<p->species.menopause) {
      if (p->num[sex][age]) {
        int amt = needed[sex]<=p->num[sex][age] ?
          needed[sex] : p->num[sex][age];
        needed[sex]-=amt;
        if (needed[sex]<=0) {
          age=p->species.menopause;
        }
      }
      age++;
    }
    if (needed[sex]>0) {
      return 0;
    }
    sex++;
  }
  return 1;
}

int breedable(struct pop_t *p)
{
  if (p->species.reproducenum<=p->species.sexes) {
    return breedable_default(p);
  }
  return breedable_clusterfuck(p);
}

int populacenum(struct populii_t *p, GString *s)
{
  int x=0;
  while (x<p->numpops) {
    if (gstreq(s,p->pops[x].species.name)) {
      return x;
    }
    x++;
  }
  return -1;
}

void move_subpop_to_item(struct fleet_t *f, struct populii_t *pops,
  struct pop_t *p,
  int *okgrowth, int *sex, int *age,
  int destpop, int ship, int item)
{
  while (*okgrowth>0) {
    if (p->num[*sex][*age]) {
      pops->pops[destpop].num[*sex][*age]++;
      pops->pops[destpop].total++;
      pops->total++;
      p->num[*sex][*age]--;
      p->total--;
      --*okgrowth;
      if (p->total<=0) {
        *okgrowth=0;
      }
    }
    ++*sex;
    if (*sex>=p->species.sexes) {
      *sex=0;
      ++*age;
      if (*age>p->species.senescence) {
        *age=0;
        if (p->total<=0) {
          *okgrowth=0; /* Break out of infinite loop. */
        }
      }
    }
  }
  dispose_item_popstrs(&f->ships[ship].items[item]);
  populate_item(&f->ships[ship].items[item],pops);
  f->ships[ship].items[item].updated=MAKROKOSMOS_NOW;
  if (f->ships[ship].items[item].modified==0) {
    f->ships[ship].items[item].modified=1;
  }
  if (f->ships[ship].modified==0) {
    f->ships[ship].modified=1;
  }
}

void move_population_to_fleet(struct fleet_t *f, struct pop_t *p)
{
  int			ship=0;
  int			item;
  int			age=0;
  int			sex=0;
  int			okgrowth;
  int			destpop;
  int			x;
  struct itemlist_t	list=null_itemlist;
  struct populii_t pops;
  while (ship<f->numships) {
    item=0;
    while (item<f->ships[ship].numitems) {
      if (f->ships[ship].items[item].class==ICLASS_HABITAT) {
        itemlist_push(&list,ship,item);
        if (f->ships[ship].items[item].numpopstrs) {
          extract_item_populii(&f->ships[ship].items[item],&pops);
          okgrowth=f->ships[ship].items[item].maxcrew-pops.total;
          if ((destpop=populacenum(&pops,p->species.name))>=0) {
            move_subpop_to_item(f,&pops,p,&okgrowth,&sex,&age,
              destpop,ship,item);
          }
          dispose_populii(&pops);
        }
      }
      if (p->total<=0) {
        ship=f->numships;
        item=f->ships[ship].numitems;
      }
      item++;
    }
    ship++;
  }
  if (p->total>=0) {
    x=0;
    while ((x<list.listlen) && (p->total>0)) {
      extract_item_populii(
        &f->ships[list.list[x].ship].items[list.list[x].item],&pops);
      okgrowth=
        f->ships[list.list[x].ship].items[list.list[x].item].maxcrew-pops.total;
      destpop=populacenum_c(&pops,p->species.name);
      move_subpop_to_item(f,&pops,p,&okgrowth,&sex,&age,destpop,
        list.list[x].ship,list.list[x].item);
      dispose_populii(&pops);
      x++;
    }
  }
  dispose_itemlist(&list);
}


int shanghai_itempop_senescent(struct item_t *i, int totalamt)
  /* Returns the number of crewmembers shanghaied. */
{
  struct populii_t populii;
  int species=0;
  int shanghaied=0;
  int amt=totalamt;
  int batch;
  int age,sex;
  extract_item_populii(i,&populii);
  while (species<populii.numpops) {
    age=populii.pops[species].species.senescence;
    while (age>populii.pops[species].species.menopause) {
      sex=0;
      while (sex<populii.pops[species].species.sexes) {
        batch=lesser(amt,populii.pops[species].num[sex][age]);
        amt-=batch;
        shanghaied+=batch;
        populii.pops[species].num[sex][age]-=batch;
        if (amt<=0) {
          goto done;
        }
        sex++;
      }
      age--;
    }
    species++;
  }
  done:
  if (shanghaied) {
    populate_item(i,&populii);
  }
  dispose_populii(&populii);
  return shanghaied;
}

int shanghai_itempop_maximal(struct item_t *i, int totalamt)
{
  struct populii_t pop,seed;
  int shanghaied=0;
  int amt=totalamt;
  int batch,sex,age,species=0;
  extract_item_populii(i,&pop);
  extract_seed_populii(&pop,&seed,NULL);
  while (species<pop.numpops) {
    age=pop.pops[species].species.senescence;
    while (age>=0) {
      sex=0;
      while (sex<pop.pops[species].species.sexes) {
        if (pop.pops[species].num[sex][age]) {
          batch=lesser(amt,pop.pops[species].num[sex][age]);
          pop.pops[species].num[sex][age]-=batch;
          //  IIRC, I don't need to update the totals here.
          amt-=batch;
          shanghaied+=batch;
          if (amt<=0) goto done;
        }
        sex++;
      }
      age--;
    }
    species++;
  }
  done:
  if (shanghaied) {
    absorb_populii(&seed,&pop);
    populate_item(i,&seed);
  }
  dispose_populii(&pop);
  dispose_populii(&seed);
  return shanghaied;
}

int habitability(char *zone, struct species_t *species)
{
  int		retval=2;
  int		x=0;
  int		chempos=0;
  double	envamt;
  int		habamt;
  envamt=extract_amt(zone,"p+");
  if (envamt>species->maxpress) {
    retval=2-((envamt-species->maxpress)/(species->maxpress-species->minpress));
  }
  envamt=extract_amt(zone,"p-");
  if (envamt<species->minpress) {
    retval=lesser(retval,2-((species->minpress-envamt)/(species->maxpress-species->minpress)));
  }
  envamt=extract_amt(zone,"t+");
  if (envamt>species->maxtemp) {
    retval=lesser(retval,2-((envamt-species->maxtemp)/(species->maxtemp-species->mintemp)));
  }
  envamt=extract_amt(zone,"t-");
  if (envamt<species->mintemp) {
    retval=lesser(retval,2-((species->mintemp-envamt)/(species->maxtemp-species->mintemp)));
  }
  while (x<species->habitat->len) {
    switch (species->habitat->str[x]) {
    case '>':
      species->habitat->str[x]='\0';
      habamt=atoi(&species->habitat->str[x+1]);
      envamt=extract_amt(zone,&species->habitat->str[chempos]);
      if (envamt<habamt) {
        if (envamt==0) {
          envamt=1; /* I don't like this. */
        }
        retval=lesser(retval,2-(habamt/envamt));
      }
      species->habitat->str[x]='>';
      break;
    case '<':
      species->habitat->str[x]='\0';
      habamt=atoi(&species->habitat->str[x+1]);
      envamt=extract_amt(zone,&species->habitat->str[chempos]);
      if (envamt>habamt) {
        retval=lesser(retval,2-(envamt/habamt));
      }
      species->habitat->str[x]='<';
      break;
    case ',':
      chempos=x+1;
      break;
    default:
    }
    x++;
  }
  /* Return values:
   *    Habitability level 2: Conditions entirely within habitable range.
   *    Habitability Level 1: Conditions partially habitable.  Conditions do
   *		not exceed habitable-range by more than habitable-range.
   *    Habitability level 0:
   *            0a: Conditions partly habitable, and maximum not off by more than range*2
   *            0b: Conditions not off by more than range.
   *    -1:     Off by range*2.  No overlap.
   *    Additional distortions: For each [range's] worth of uninhabitability, subtract 1.
   *
   */
  return retval;
}

int speciesindex(struct populii_t *p, GString *g)
{
	/* Returns -1 if species not found; otherwise, returns the index
	   of the species in the populii array. */
  int x=0;
  while (x<p->numpops) {
    if (gstreq(p->pops[x].species.name,g)) {
      return x;
    }
    x++;
  }
  return -1;
}

int habitability_to_populii(char *zone, struct populii_t *p,
  struct gstrlist_t *specie)
{
  int   x=0;
  int	retval=2;
  int   lasthabitability=2;
  int	s;
  while (x<specie->num) {
    s=speciesindex(p,specie->strings[x]);
    if (s>=0) {
      lasthabitability=habitability(zone,&p->pops[s].species);
      retval=lesser(lasthabitability,retval);
    }
    x++;
  }
  return retval;
}

GString *genderstring(GString *genders, int num)
{
  int x=0;
  int y=0;
  GString *s;
  while ((x<genders->len) && (y<num)) {
    if (genders->str[x]==',') y++;
    x++;
  }
  s=g_string_new(&genders->str[x]);
  x=0;
  while ((x<s->len) && (s->str[x]!=',')) {
    x++;
  }
  if ((x>=0) && (s->str[x]==',')) {
    g_string_truncate(s,x);
  }
  return s;
}

void create_newplayer_population(struct populii_t *p, struct species_t *s)
{
  int x=0;
  int pop=0;
  int add=MAKROKOSMOS_NEWPLAYER_POP/s->sexes;
  *p=null_populii;
  pop=populacenum_c(p,s->name);
  species_assign(&p->pops[pop].species,s);
  while (x<s->sexes) {
    p->pops[pop].num[x][s->puberty]=add;
    x++;
  }
  p->pops[pop].total=add*s->sexes;
  p->total=p->pops[pop].total;
}

void move_populii_to_fleet(struct fleet_t *f, struct populii_t *p)
{
  int x=0;
  int oldtotal;
  while (x<p->numpops) {
    oldtotal=p->pops[x].total;
    move_population_to_fleet(f,&p->pops[x]);
    if (p->pops[x].total<oldtotal) {
      p->total-=(oldtotal-p->pops[x].total);
    }
    if (p->pops[x].total) {
      return;
    }
    x++;
  }
}

void trim_populii(struct populii_t *p, struct populii_t *overflow,
  int num)
{
  int age=p->pops[0].species.senescence;
  int sex=p->pops[0].species.sexes-1;
  int pop=0;
  int destpop;
  int move;
  destpop=populacenum_c(overflow,p->pops[pop].species.name);
  while (p->total>num) {
    if (p->pops[pop].num[sex][age]) {
      move=greater(p->pops[pop].num[sex][age]-1,1);
      if (move>p->total-num) {
        move=p->total-num;
      }
      overflow->pops[destpop].num[sex][age]+=move;
      overflow->pops[destpop].total+=move;
      overflow->total+=move;
      p->pops[pop].num[sex][age]-=move;
      p->pops[pop].total-=move;
      p->total-=move;
    }
    sex--;
    if (sex<0) {
      sex=p->pops[pop].species.sexes-1;
      age--;
      if (age<0) {
        pop++;
        if (pop>=p->numpops) {
          pop=0;
        }
        age=p->pops[pop].species.senescence;
        sex=p->pops[pop].species.sexes-1;
        destpop=populacenum_c(overflow,p->pops[pop].species.name);
      }
    }
  }
}

void extract_species_params(struct species_t *s, struct species_params_t *p)
{
  int basetemp=s->mintemp+((s->maxtemp-s->mintemp)/2);
  float basepress=s->minpress+((s->maxpress-s->minpress)/2);
  p->points=-1;
  p->temptolerance=(basetemp-s->mintemp)/cbrt(basetemp);
  p->pressuretolerance=(basepress-s->minpress)/cbrt(basepress);
}

struct listpane_data_t *habitatpane_data(GString *habitat)
{
  int x=0;
  struct listpane_data_t *data=g_malloc(sizeof(struct listpane_data_t));
  struct habitatreq_item *hreqs=g_malloc(0);
  int chempos=0;
  data->size=sizeof(struct habitatreq_item);
  data->num=0;
  while (x<habitat->len) {
    switch (habitat->str[x]) {
	case '>':
	case '<':
      hreqs=g_realloc(hreqs,sizeof(struct habitatreq_item)*(data->num+1));
      hreqs[data->num].cond=habitat->str[x];
      habitat->str[x]='\0';
      hreqs[data->num].chem=g_string_new(&habitat->str[chempos]);
      habitat->str[x]=hreqs[data->num].cond;
      hreqs[data->num].amt=atoi(&habitat->str[x+1]);
      data->num++;
      break;
    case ',':
      chempos=x+1;
    }
    x++;
  }
  data->data=(char *)hreqs;
  return data;
}

GString *habitatpane_gstring(struct listpane_data_t *data)
{
  GString *ret=g_string_new("");
  char buf[40];
  int x=0;
  struct habitatreq_item *hreqs=(struct habitatreq_item *)data->data;
  if (data->size != sizeof(struct habitatreq_item)) {
    ui_error("habitatpane_gstring(): converting a missized record.");
    return ret;
  }
  while (x<data->num) {
    g_string_append(ret,hreqs[x].chem->str);
    g_string_append_c(ret,hreqs[x].cond);
    snprintf(&buf[0],39,"%d,",hreqs[x].amt);
    g_string_append(ret,&buf[0]);
    x++;
  }
  return ret;
}

void extract_popline_totals(struct pop_totals_t *totals, GString *line)
{
  int x=0;
  int y;
  int species=-1;
  int a;
  while (x<line->len) {
    switch (line->str[x]) {
    case 'S':
      y=x;
      while ((y<line->len) && (line->str[++y]!=':')) {
      }
      if (y<line->len) {
        line->str[y]='\0';
        a=totals->list.num;
        species=gstrlistindex_str_c(&totals->list,&line->str[x+1]);
        if (totals->list.num != a) {
          totals->total=g_realloc(totals->total,totals->list.num*sizeof(int));
          totals->total[species]=0;
        }
        line->str[y]=':';
        x=y;
        a=0;
      }
      break;
    case '0'...'9':
      a*=10;
      a+=line->str[x]-'0';
      break;
    case ';':
      if (species>=0) {
        totals->total[species]+=a;
      }
      /* No break here, as we want the accumulator reset. */
    default:
      a=0;
    }
    x++;
  }
}

void extract_itempop_totals(struct pop_totals_t *totals, struct item_t *item)
{
  int x=0;
  while (x<item->numpopstrs) {
    extract_popline_totals(totals, item->popstrs[x]);
    x++;
  }
}

void extract_fleetpop_totals(struct pop_totals_t *totals, struct fleet_t *fleet)
{
  int ship=0;
  *totals=null_pop_totals; 
  while (ship<fleet->numships) {
    int item=0;
    while (item<fleet->ships[ship].numitems) {
      switch (fleet->ships[ship].items[item].class) {
      case ICLASS_HABITAT:
        extract_itempop_totals(totals, &fleet->ships[ship].items[item]);
      }
      item++;
    }
    ship++;
  }
}

void dispose_pop_totals(struct pop_totals_t *totals)
{
  dispose_gstrlist(&totals->list);
  g_free(totals->total);
}
