/*
 * Curses UI routines
 * show_*, read_*, ui_*
 */

//#define __USE_ISOC99

#include <math.h>
#include <ncurses.h>
#include <glib.h>
#include <errno.h>
#include <string.h>
#include "microcosm.h"
#include "util.h"
#include "vectors.h"
#include "populii.h"
#include "factions.h"
#include "ui_all.h"
#include "ui_curses.h"
#include "sectors.h"
#include "charts.h"
#include "fleets.h"

int COLOR_CHART_BLANK=COLOR_PAIR(10)|A_BOLD;

void ui_configure(struct player_t *player)
{
  int c=0;
  while (1) {
    text("Select a colour for unexplored sectors\n");
    printw("(If you only see one option, it means your terminal cannot display\n");
    printw("the others.)\n");
    init_pair(COLOR_TEMP1_NUM, COLOR_BLACK, COLOR_BLACK);
    attrset(COLOR_PAIR(COLOR_TEMP1_NUM)|A_BOLD);
    printw("0: Use bright black/dark grey.\n");
    init_pair(COLOR_TEMP2_NUM, COLOR_YELLOW, COLOR_BLACK);
    attrset(COLOR_PAIR(COLOR_TEMP2_NUM));
    printw("1: Use brown.\n");
    refresh();
    switch (c=read_char("Unexplored Colour: ")) {
    case '0':
      player->theme.chart_blank=0x666666;
      player->theme.custom=1;
      return;
    case '1':
      player->theme.chart_blank=0x808000;
/*      player->theme.custom=1; */
      return;
    default:
    }
  } 
}

void pause_prompt(void)
{
  attrset(COLOR_PROMPT);
  printw("[Press any key]");
  refresh();
  getch();
  printw("\n");
}

void show_habitat(GString *h)
{
  printw("%s\n",h->str);
}

int ui_species_edit_prompt(int points)
{
  int c;
  attrset(COLOR_PROMPT);
  printw("Action [");
  attrset(COLOR_INPUT);
  printw("%d",points);
  attrset(COLOR_PROMPT);
  printw(" bioengineering points remaining] [");
  hotkey('?');
  attrset(COLOR_PROMPT);
  printw(": Help]: ");
  attrset(COLOR_INPUT);
  c=getch();
  printw("%c\n",c);
  return c;
}

int ui_charts_prompt(void)
{
  attrset(COLOR_PROMPT);
  printw("Charts command [");
  hotkey('?');
  return read_char(": Help]: ");
}

void show_file_paged(const char *file)
{
  FILE *f;
  int x,height;
  char s[1024];
  GString *path=g_string_new(g_Config.datapath->str);
  g_string_append(path,MAKROKOSMOS_HELP_PATH);
  g_string_append(path,file);
  getmaxyx(stdscr,height,x);
  f=fopen(path->str,"r");
  attrset(COLOR_TEXT);
  if (f) {
    x=1;
    while (fgets(&s[0],1023,f)) {
      s[1023]='\0';
      printw("%s",&s[0]);
      x++;
      if (x>=height-2) {
        pause_prompt();
        attrset(COLOR_TEXT);
        x=1;
      }
    }
    fclose(f);
    printw("\n");
  } else {
    ui_config_error_ss("Cannot open text file ",file,errno);
  }
  pause_prompt();
}

void show_species_editor_help(void)
{
  show_file_paged("species.txt");
}

void show_species_info(struct species_t *s, struct species_params_t *p)
{
  int		x;
  GString	*str;
  clear();
  attrset(COLOR_TEXT);
  printw("Species configuration menu:\n\n");
  attrset(COLOR_HOTKEY);
  printw("N");
  attrset(COLOR_TEXT);
  printw(": Species name: %s\n",s->name->str);
  attrset(COLOR_HOTKEY);
  printw("H");
  attrset(COLOR_TEXT);
  printw("+ ");
  attrset(COLOR_HOTKEY);
  printw("h");
  attrset(COLOR_TEXT);
  printw("- : Average HP: %d\n",s->hp);
  attrset(COLOR_HOTKEY);
  printw("S");
  attrset(COLOR_TEXT);
  printw(": Sexes: %d (",s->sexes);
  if (s->sexes) {
    x=0;
    while (x<s->sexes) {
      attrset(COLOR_HOTKEY);
      printw("%d",x+1);
      attrset(COLOR_TEXT);
      str=genderstring(s->genders,x);
      printw(": %s, ",str->str);
      g_string_free(str,TRUE);
      x++;
    }
  } else {
    printw("not defined");
  }
  printw(")\n");
  attrset(COLOR_HOTKEY);
  printw("#");
  attrset(COLOR_TEXT);
  printw(": Number to reproduce: %d\nAge at: ",s->reproducenum);
  hotkey('F');
  attrset(COLOR_TEXT);
  printw(": Start of fertility: %d, ",s->puberty);
  hotkey('M');
  attrset(COLOR_TEXT);
  printw(": Menopause: %d, ",s->menopause);
  hotkey('E');
  attrset(COLOR_TEXT);
  printw(": Senescence: %d\n",s->senescence);
  hotkey('R');
  attrset(COLOR_TEXT);
  printw(": Mortality: %f    ",s->mortality);
  hotkey('O');
  attrset(COLOR_TEXT);
  printw(": Odds of reproduction: %f\nTemperature: ",s->reproduceodds);
  hotkey('T');
  attrset(COLOR_TEXT);
  printw(": Base temp: %d    ",s->maxtemp-((s->maxtemp-s->mintemp)>>1));
  hotkey('t');
  attrset(COLOR_TEXT);
  printw(": Tolerance: %d    (Range: %d-%d)\nPressure: ",
    p->temptolerance, s->mintemp, s->maxtemp);
  hotkey('P');
  attrset(COLOR_TEXT);
  printw(": Base pressure: %f\n",p->basepressure);
//    (float)s->maxpress-((s->maxpress-s->minpress)/2));
  hotkey('p');
  attrset(COLOR_TEXT);
  printw(": variation tolerance: %d (Range: %f-%f)\n",
    p->pressuretolerance,
    (float)s->minpress, (float)s->maxpress);
  hotkey('A');
  text(": Atmospheric requirements: ");
  show_habitat(s->habitat);
  hotkey('Q');
  text(": Quit and save.\n");
  refresh();
}

int curses_color(int color)
{
  int red=(color & 0xff0000)>=0x800000;
  int green=(color & 0xff00)>=0x008000;
  int blue=(color & 0x000ff)>=0x000080;
  if (red) {
    if (green) {
      if (blue) {
        return COLOR_WHITE;
      } else {
        return COLOR_YELLOW;
      }
    } else {
      if (blue) {
        return COLOR_MAGENTA;
      } else {
        return COLOR_RED;
      }
    }
  } else {
    if (green) {
      if (blue) {
        return COLOR_CYAN;
      } else {
        return COLOR_GREEN;
      }
    } else {
      if (blue) {
        return COLOR_BLUE;
      }
    }
  }
  return COLOR_BLACK;
}

int curses_attribs(int color)
{
  if (((color & 0xff0000) >= 0x990000) || 
    ((color & 0x00ff00) >= 0x009900) || ((color & 0x0000ff)>=0x000099)) {
    return A_BOLD;
  } else if (color == 0x666666) {
    return A_BOLD;
  }
  return 0;
}

void ui_init(void)
{
  initscr(); cbreak(); noecho(); scrollok(stdscr,TRUE);
  start_color();
  init_pair(COLOR_TEXT_NUM,COLOR_GREEN,COLOR_BLACK);
  init_pair(COLOR_HOTKEY_NUM,COLOR_CYAN,COLOR_BLACK);
  init_pair(COLOR_PROMPT_NUM,COLOR_MAGENTA,COLOR_BLACK);
  init_pair(COLOR_INPUT_NUM,COLOR_CYAN,COLOR_BLACK);
  init_pair(COLOR_ERROR_NUM,COLOR_RED,COLOR_BLACK);
  init_pair(6,COLOR_WHITE,COLOR_BLACK);
  init_pair(7,COLOR_CYAN,COLOR_BLACK);
  init_pair(8,COLOR_BLUE,COLOR_BLACK);
  init_pair(9,COLOR_GREEN,COLOR_BLACK);
  init_pair(10,curses_color(g_Player.theme.chart_blank),COLOR_BLACK);
  COLOR_CHART_BLANK=curses_attribs(g_Player.theme.chart_blank)|COLOR_PAIR(10);
  keypad(stdscr,TRUE);
}

void ui_shutdown(void)
{
  endwin();
}

void show_fleet_menu_help(void)
{
  attrset(COLOR_TEXT);
  printw("Fleet menu help:\n");
  attrset(COLOR_HOTKEY);
  addch('?');
  attrset(COLOR_TEXT);
  printw(": Help\n");
  attrset(COLOR_HOTKEY);
  addch('B');
  attrset(COLOR_TEXT);
  printw(": Board\n");
  attrset(COLOR_HOTKEY);
  addch('O');
  attrset(COLOR_TEXT);
  printw(": Organize fleets\n");
  attrset(COLOR_HOTKEY);
  addch('R');
  attrset(COLOR_TEXT);
  printw(": Repair (unimplimented)\n");
  attrset(COLOR_HOTKEY);
  addch('S');
  attrset(COLOR_TEXT);
  printw(": Split fleet\n");
  attrset(COLOR_HOTKEY);
  addch('Q');
  attrset(COLOR_TEXT);
  printw(": Quit\n\n");
  refresh();
}

void draw_ship_bar_target(WINDOW *win, struct ship_t *s, int line)
{
  wmove(win,line,0);
  if (s) {
    if (s->immobile) {
      wprintw(win,"I*%s* ",s->base->str);
    } else {
      wprintw(win," *%s* ",s->base->str);
    }
  } else {
    wprintw(win," ** ");
  }
  wrefresh(win);
}

void draw_ship_bar_unselected(WINDOW *win,struct ship_t *s, int line)
{
  wmove(win,line,0);
  wattrset(win,COLOR_LIGHTBAR_UNSELECTED);
  if (s) {
    if (s->immobile) {
      waddch(win,'I');
    } else {
      waddch(win,' ');
    }
    wprintw(win," %s  ",s->base->str);
  } else {
    wprintw(win,"    ");
  }
  wrefresh(win);
}

void draw_item_bar(struct fleetpane_t *p,int pos)
{
  wmove(p->win,pos-p->spage,0);
  wattrset(p->win,COLOR_LIGHTBAR_UNSELECTED);
  if (p->fleet->ships[p->pos].items[pos].equipped) {
    waddch(p->win,'E');
  } else {
    waddch(p->win,' ');
  }
  if (p->fleet->ships[p->pos].items[pos].immobile) {
    wprintw(p->win,"I %s  ",p->fleet->ships[p->pos].items[pos].base->str);
  } else {
    wprintw(p->win,"  %s  ",p->fleet->ships[p->pos].items[pos].base->str);
  }
}

void draw_item_bar_unselected(struct fleetpane_t *p)
{
  draw_item_bar(p,p->spos);
}

void draw_pane_bar_selected(struct fleetpane_t *p)
{
  switch (p->state) {
  case PANE_FLEET:
    draw_ship_bar_selected(p);
    break;
  case PANE_SHIP:
    draw_item_bar_selected(p);
    break;
  case PANE_ITEM:
    /* At the item level, this is not yet relevant. */
    break;
  default:
    ui_error("Unknown pane state.");
  }
}

void draw_item_bar_selected(struct fleetpane_t *pane)
{
  wmove(pane->win,pane->spos-pane->spage,0);
  wattron(pane->win,A_STANDOUT);
  if (pane->fleet->ships[pane->pos].items[pane->spos].equipped) {
    waddch(pane->win,'E');
  } else {
    waddch(pane->win,' ');
  }
  if (pane->fleet->ships[pane->pos].items[pane->spos].immobile) {
    wprintw(pane->win,"I[%s] ",
      pane->fleet->ships[pane->pos].items[pane->spos].base->str);
  } else {
    wprintw(pane->win," [%s] ",
      pane->fleet->ships[pane->pos].items[pane->spos].base->str);
  }
  wrefresh(pane->win);
  wattroff(pane->win,A_STANDOUT);
}

void draw_pane_bar_unselected(struct fleetpane_t *p)
{
  switch (p->state) {
  case PANE_FLEET:
    draw_ship_bar_unselected(p->win,&p->fleet->ships[p->pos],p->pos-p->page);
    break;
  case PANE_SHIP:
    draw_item_bar_unselected(p);
    break;
  case PANE_ITEM:
    /* At the item level, this is not yet relevant. */
    break;
  default:
    ui_error("Unknown pane state.");
  }
}

void draw_ship_bar_selected(struct fleetpane_t *pane)
{
  wmove(pane->win,pane->pos-pane->page,0);
  wattron(pane->win,A_STANDOUT);
  if (pane->pos<pane->fleet->numships) {
    if (pane->fleet->ships[pane->pos].immobile) {
      waddch(pane->win,'I');
    } else {
      waddch(pane->win,' ');
    }
    wprintw(pane->win,"[%s] ",pane->fleet->ships[pane->pos].base->str);
  } else {
    wprintw(pane->win," [] ");
  }
  wrefresh(pane->win);
  wattroff(pane->win,A_STANDOUT);
}

void draw_item_pane(struct fleetpane_t *pane)
{
  struct item_t *i=&pane->fleet->ships[pane->pos].items[pane->spos];
  wmove(pane->win,0,0);
  wprintw(pane->win, "Class: %s\n",item_class_string(i->class));
  wprintw(pane->win,"HP: %d/%d\n",i->hp,i->maxhp);
  wprintw(pane->win,"Surface area: %d\n",i->surface);
  wprintw(pane->win,"Volume: %d\n",i->volume);
  if (i->cargo->len) {
    wprintw(pane->win,"Cargo: %s\nCargo Qty: %d/%d\n",i->cargo->str,
      i->cargoqty,i->cargomax);
  }
  wrefresh(pane->win);
}

void draw_fleet_pane(struct fleetpane_t *pane)
{
  int x;
  wclear(pane->win);
  switch (pane->state) {
  case PANE_FLEET:
    x=pane->page;
    while (x<pane->page+pane->height) {
      if (x<pane->fleet->numships) {
        draw_ship_bar_unselected(pane->win,&pane->fleet->ships[x],x-pane->page);
      } else {
        x=pane->page+pane->height;
      }
      x++;
    }
    break;
  case PANE_SHIP:
    x=pane->spage;
    while (x<pane->spage+pane->height) {
      if (x<pane->fleet->ships[pane->pos].numitems) {
        draw_item_bar(pane,x);
      } else {
        x=pane->spage+pane->height;
      }
      x++;
    }
    break;
  case PANE_ITEM:
    draw_item_pane(pane);
    break;
  default:
    ui_error("Invalid pane state.");
    return;
  }
  draw_pane_bar_selected(pane);
//  wrefresh(pane->win);
}

void show_fleet_menu(const char *currfleet)
{
  attrset(COLOR_TEXT);
  printw("Fleet menu [in %s]\n",currfleet);
  attrset(COLOR_PROMPT);
  printw("Command (? for help): ");
  attrset(COLOR_INPUT);
  refresh();
}

int read_int(const char *prompt)
{
  int ret=0;
  int done=0;
  int chars=0;
  int key;
  attrset(COLOR_PROMPT);
  printw("%s", prompt);
  attrset(COLOR_INPUT);
  while (done==0) {
    key=getch();
    switch (key) {
    case '0' ... '9':
      ret=(ret*10)+(key-'0');
      printw("%d",key-'0');
      refresh();
      chars++;
      break;
    case 8:
    case 127:
    case KEY_BACKSPACE:
      if (chars) {
        chars--;
        ret=(ret/10);
        printw("\010 \010");
        refresh();
      }
      break;
    case 10:
    case 13:
    case KEY_ENTER:
      done=1;
      break;
    default:
      beep();
    }
  }
  printw("\n");
  refresh();
  if (chars) {
    return ret;
  } else {
    return -1;
  }
}

int read_char(const char *prompt)
{
  int c;
  attrset(COLOR_PROMPT);
  printw("%s",prompt);
  c=getch();
  if ((c>=' ') && (c<='~')) {
    attrset(COLOR_INPUT);
    printw("%c\n",c);
  }
  return c;
}

void show_help_landed(void)
{
  text("Commands available while landed:\n");
  hotkey('A');
  text(":  Attack               ");
  hotkey('O');
  text(":     Orbit\n");
  hotkey('B');
  text(":  Board fleet          ");
  hotkey('N');
  text(":     Name body\n");
  hotkey('C');
  text(":  Colonize             ");
  attrset(COLOR_HOTKEY);
  printw("^O");
  text(", ");
  hotkey('0');
  text(": Claim ownership\n");
  hotkey('D');
  text(":  Dock                 ");
  hotkey('R');
  text(":     Repair\n");
  attrset(COLOR_HOTKEY);
  printw("^D");
  text(": Describe body        ");
  hotkey('S');
  text(":     Scan [Unimplemented]\n");
  hotkey('E');
  text(":  Equip                ");
  hotkey('T');
  text(":     Time\n");
  hotkey('F');
  text(":  Fleet organization   ");
  hotkey(':');
  text(":     Hyperwave Commlink\n");
  attrset(COLOR_HOTKEY);
  printw("^F");
  text(": Split fleet          ");
  hotkey('?');
  text(":     Help\n");
  hotkey('H');
  text(":  History              ");
  hotkey('Q');
  text(":     Quit / quit to main\n");
}

void ui_error(const char *message)
{
  attrset(COLOR_ERROR);
  printw("Error: %s\nPlease inform deekoo@tentacle.net; it is probably a bug. \n",message);
  refresh();
}

void ui_error_sd(const char *s, int d)
{
  attrset(COLOR_ERROR);
  printw("Error: %s%d\nPlease inform deekoo@tentacle.net; it is probably a bug. \n",s,d);
  refresh();
}

void ui_error_sss(const char *s1, const char *s2, const char *s3)
{
  attrset(COLOR_ERROR);
  printw("Error: %s%s%s\nPlease inform deekoo@tentacle.net; it is probably a bug. \n",s1,s2,s3);
  refresh();
}

void ui_config_error_ss(const char *s1, const char *s2, int err)
{
  attrset(COLOR_ERROR);
  printw("Error: %s%s\nError code: %d (%s)\nPlease inform the game's admin.\n",
    s1,s2,err,g_strerror(err));
  refresh();
}

void ui_config_error_sscsc(const char *s1, const char *s2,
  char c1, const char *s3, char c2, int err)
{
  attrset(COLOR_ERROR);
  printw("Error: %s%s%c%s%c\nError code: %d (%s)\nPlease inform the game's admin.\n",
    s1,s2,c1,s3,c2,err,g_strerror(err));
  refresh();
}


void ui_config_error_sss(const char *s1, const char *s2, 
  const char *s3, int err)
{
  attrset(COLOR_ERROR);
  printw("Error: %s%s%s\n",s1,s2,s3);
  if (err) {
    printw("System error code: %d (%s)\n",err,g_strerror(err));
  } else {
    printw("No system error code available.\n");
  }
  printw("Please inform the game's admin.\n");
  refresh();
}

void ui_config_error(const char *message, int err)
{
  attrset(COLOR_ERROR);
  printw("%s\n",message);
  if (err!=-1) {
    printw("Error code %d (%s)\n",err,g_strerror(err));
  }
  printw("Please inform the game's admin.\n");
  refresh();
}

void show_build_help(void)
{
  hotkey('A');
  text(": Atmosphere miner\n");
  hotkey('S');
  text(": Ship\n");
  hotkey('Q');
  text(": Quit to main\n");
  hotkey('Y');
  text(": Shipyard\n");
}

void show_help_docked(void)
{
  text("Commands available while docked:\n");
  hotkey('?');
  text(":  Help                       ");
  hotkey('L');
  text(": Land outside colony\n");
  hotkey('B');
  text(":  Build                      ");
  hotkey('O');
  text(": Orbit\n");
  hotkey('D');
  text(":  Dock                       ");
  hotkey('P');
  text(": Population menu\n");
  hotkey('^');
  addch('D');
  text(": Describe colony.\n");
  hotkey('E');
  text(":  Equip                      ");
  hotkey('R');
  text(": Repair\n");
  hotkey('F');
  text(":  Fleet organization         ");
  hotkey('T');
  text(": Time\n");
  hotkey('Q');
  text(":  Quit to main               ");
  hotkey(':');
  text(": Hyperwave Commlink\n");
}

void show_location_docked(struct sector_t *sect)
{
  struct fleet_t col=null_fleet;
  printw("Docked at colony %s on planet %s in sector %s\n",
    g_Player.dock->str,sect->bodies[g_Player.orbit].name->str,
    g_Player.coords->str);
  load_fleet(g_Player.dock->str, &col);
  if (col.desc) {
    printw("%s\n", col.desc->str);
  } else {
    if (sect->bodies[g_Player.orbit].desc->len) {
      printw("%s\n", sect->bodies[g_Player.orbit].desc->str);
    }
  }
}

void show_single_orbiter(int num, struct body_t *b)
{
  printw("%d: %s", num, body_name(b));
  if (b->owner->len) {
    printw("   (Claimed by %s)\n", b->owner->str);
  } else {
    printw("\n");
  }
}

float read_float(char *prompt)
{
  int	notdone=1;
  int	retint=0;
  int	retdec=0;
  int	decpos=1;
  int	decimal=0;
  int	c;
  attrset(COLOR_PROMPT);
  printw("%s",prompt);
  attrset(COLOR_INPUT);
  refresh();
  while (notdone) {
    c=getch();
    switch (c) {
    case 0x08:
    case 0x7F:
    case KEY_BACKSPACE:
      if (decimal) {
        if (retdec) {
          retdec/=10;
          decpos/=10;
        } else {
          decimal=0;
        }
        printw("\010 \010");
      } else {
        if (retint) {
          retint/=10;
          printw("\010 \010");
        }
      }
      break;
    case 0xA:
    case 0xD:
      notdone=0;
      break;
    case '0'...'9': 
      if (decimal) {
        retdec*=10;
        retdec+=c-'0';
        decpos*=10;
      } else {
        retint*=10;
        retint+=c-'0';
      }
      addch(c);
      break;
    case '.':
      if (decimal==0) {
        decimal=1;
        addch('.');
        break;
      }
    default:
      beep();
    }
  }
  printw("\n");
  return retint+((float)retdec/decpos);
}

void show_equip_help(void)
{
  text("Equip menu:\n");
  hotkey('?');
  text(": Help\n")
  hotkey('E');
  text(": Equip everything\n");
  hotkey('Q');
  text(": Quit to main\n");
  hotkey('T');
  text(": Toggle item (incomplete)\n");
  refresh();
}

void show_help_orbit()
{
  text("Orbit help:\n");
  hotkey('?');
  text(":  Help                ");
  hotkey('L');
  text(": Land\n");
  hotkey('A');
  text(":  Attack              ");
  hotkey('O');
  text(": Orbit\n");
  hotkey('B');
  text(":  Board               ");
  hotkey('R');
  text(": Repair\n");
  hotkey('E');
  text(":  Equip               ");
  hotkey('T');
  text(": Time\n");
  hotkey('F');
  text(":  Fleet organization  ");
  hotkey(':');
  text(": Hyperwave Commlink\n");
  attrset(COLOR_HOTKEY);
  printw("^F");
  text(": Split fleet         ");
  hotkey('Q');
  text(": Quit\n");
}

void dispose_fleetpane(struct fleetpane_t *pane)
{
  delwin(pane->win);
}

void fleetpane_down(struct fleetpane_t *pane)
{
  switch (pane->state) {
  case PANE_FLEET:
    if (pane->pos<pane->fleet->numships-1) {
      draw_ship_bar_unselected(pane->win,&pane->fleet->ships[pane->pos],
        pane->pos);
      pane->pos++;
      if (pane->pos>=pane->page+pane->height) {
        pane->page+=pane->height;
        draw_fleet_pane(pane);
      }
      draw_ship_bar_selected(pane);
    }
    break;
  case PANE_SHIP:
    if (pane->spos+1<pane->fleet->ships[pane->pos].numitems) {
      draw_item_bar_unselected(pane);
      pane->spos++;
      if (pane->spos>=pane->spage+pane->height) {
        pane->spage+=pane->height;
        draw_fleet_pane(pane);
      }
      draw_item_bar_selected(pane);
    }
    break;
  }
}

void draw_equip_footer(struct fleetpane_t *pane)
{
  move(pane->height+1,0);
  switch (pane->state) {
  case PANE_FLEET:
    attrset(COLOR_TEXT);
    addch('[');
    attrset(COLOR_HOTKEY);
    printw("Enter");
    text("]: view ship.  [");
    attrset(COLOR_HOTKEY);
    printw("Up");
    text("], [");
    attrset(COLOR_HOTKEY);
    printw("Down");
    text("]: Move lightbar.  ");
    hotkey('Q');
    text(": Quit\n");
    clrtoeol();
    break;
  case PANE_SHIP:
    attrset(COLOR_TEXT);
    addch('[');
    attrset(COLOR_HOTKEY);
    printw("Enter");
    text("]: view item.  [");
    attrset(COLOR_HOTKEY);
    printw("Up");
    text("], [");
    attrset(COLOR_HOTKEY);
    printw("Down");
    text("]: Move lightbar.  ");
    hotkey('Q');
    text(": Quit\n[");
    attrset(COLOR_HOTKEY);
    printw("Left");
    text("]: Back  [");
    attrset(COLOR_HOTKEY);
    printw("Space");
    text("]: Equip item.\n");
    break;
  case PANE_ITEM:
    attrset(COLOR_TEXT);
    addch('[');
    attrset(COLOR_HOTKEY);
    printw("Left");
    text("]: Return to parent ship.\n");
    clrtoeol();
    break;
  default:
    attrset(COLOR_ERROR);
    printw("Error: Unknown pane state %d.\n",pane->state);
  }
  refresh();
}

void draw_bar_selected(struct fleetpane_t *pane)
{
  switch (pane->state) {
  case PANE_FLEET:
    draw_ship_bar_selected(pane);
    break;
  case PANE_SHIP:
    draw_item_bar_selected(pane);
  }
}

void show_tech_difference(struct tech_t *real, struct tech_t *req)
{
  text("Required: ");
  if (req->bio>real->bio) {
    printw("Biochemical: %d needed, %d available.\n",
      req->bio, real->bio);
  }
  if (req->emf>real->emf) {
    printw("Electromagnetic: %d needed, %d available.\n",
      req->emf, real->emf);
  }
  if (req->grav>real->grav) {
    printw("Gravitotemporal: %d needed, %d available.\n",
      req->grav, real->grav);
  }
  if (req->logic>real->logic) {
    printw("Logical: %d needed, %d available.\n",
      req->logic, real->logic);
  }
  if (req->mech>real->mech) {
    printw("Mechanical: %d needed, %d available.\n",
      req->mech, real->mech);
  }
  if (req->psi>real->psi) {
    printw("Psionomagical: %d needed, %d available.\n",
      req->psi, real->psi);
  }
  if (req->quantum>real->quantum) {
    printw("Quantum: %d needed, %d available.\n",
      req->quantum, real->quantum);
  }
}

void fleetpane_up(struct fleetpane_t *pane)
{
  switch (pane->state) {
  case PANE_FLEET:
    if (pane->pos>0) {
      if (pane->pos<pane->fleet->numships) {
        draw_ship_bar_unselected(pane->win,&pane->fleet->ships[pane->pos],
          pane->pos);
      } else {
        wmove(pane->win,0,pane->pos);
        clrtoeol();
      }
      pane->pos--;
      if (pane->pos<pane->page) {
        pane->page-=pane->height;
        draw_fleet_pane(pane);
      }
      draw_ship_bar_selected(pane);
    }
    break;
  case PANE_SHIP:
    if (pane->spos>0) {
      draw_item_bar_unselected(pane);
      pane->spos--;
      if (pane->spos<pane->spage) {
        pane->spage-=pane->height;
        draw_fleet_pane(pane);
      }
      draw_item_bar_selected(pane);
    }
  }
}

void show_help_sector(void)
{
  text("Sector help:\n");
  hotkey('A');
  text(":  Attack                 ");
  hotkey('O');
  text(":     Orbit\n");
  hotkey('B');
  text(":  Board fleet            ");
  attrset(COLOR_HOTKEY);
  printw("^O");
  text(", ");
  hotkey('0');
  text(": Claim ownership\n");
  hotkey('C');
  text(":  Charts                 ");
  hotkey('P');
  text(": Population menu\n");
  hotkey('E');
  text(":  Equip                  ");
  hotkey('R');
  text(":     Repair\n");
  hotkey('F');
  text(":  Fleet organization     ");
  hotkey('T');
  text(":     Time\n");
  attrset(COLOR_HOTKEY);
  printw("^F");
  text(": Split fleet            ");
  hotkey(':');
  text(":     Hyperwave Commlink\n");
  hotkey('H');
  text(":  History                ");
  hotkey('?');
  text(":     Help\n");
  hotkey('M');
  text(":  Move                   ");
  hotkey('Q');
  text(":     Quit\n");
}

GString *read_gstring_low(int mask)
{
  int c;
  int notdone=1;
  GString *g=g_string_new("");
  attrset(COLOR_INPUT);
  refresh();
  while (notdone) {
    c=getch();
    switch (c) {
    case KEY_BACKSPACE:
      if (g->len>0) {
        if (!mask) { printw("\010 \010"); }
        g_string_truncate(g,g->len-1);
      }
      break;
    case '\t':
      beep();
      break;
    case 0x0A:
    case 0x0D:
      notdone=0;
      printw("\n");
      break;
    default:
      if (c>=' ') {
        g_string_append(g," ");
        g->str[g->len-1]=c;
        if (!mask) {
          printw("%c",c);
        }
      } else {
        beep();
      }
    }
  }
  return g;
}

GString *read_gstring(const char *prompt)
{
  attrset(COLOR_PROMPT);
  printw("%s",prompt);
  return read_gstring_low(FALSE);
}

GString *read_accesscode(const char *prompt)
{
  attrset(COLOR_PROMPT);
  printw("%s",prompt);
  return read_gstring_low(TRUE);
}

GString *read_gstring_scs(const char *s1, char c, const char *s2)
{
  attrset(COLOR_PROMPT);
  printw("%s%c%s",s1,c,s2);
  return read_gstring_low(FALSE);
}

void show_species_info_box(struct species_t *s)
{
  struct species_params_t params;
  extract_species_params(s,&params);
  show_species_info(s,&params);
  attrset(COLOR_PROMPT);
  printw("[ Press any key ]");
  attrset(COLOR_INPUT);
  getch();
  printw("\n");
}

void ui_init_scan_result(struct coord_t *c, int range)
{
  attrset(COLOR_TEXT);
  printw("Scanning from %d/%d/%d (range %d):\n",c->x,c->y,c->z,range);
  refresh();
}

void ui_add_scan_result_bodies(int bodies, struct coord_t *c)
{
  printw("%d bodies detected in sector %d/%d/%d\n",bodies,c->x,c->y,c->z);
  refresh();
}

void ui_add_scan_result_system(struct sector_t *s, struct coord_t *c)
{
  int b=0;
  int stars=0;
  while (b<s->numbodies) {
    if (s->bodies[b].mass>1.6E29) stars++;
    b++;
  }
  if (stars) {
    printw("%d large bodies, probably stars, detected in sector %d/%d/%d\n",
      stars,c->x,c->y,c->z);
    refresh();
  }
}

void ui_add_scan_result_fleets(struct sector_t *s, struct coord_t *c)
{
  struct gstrlist_t list=null_gstrlist;
  int body=0,x=0;
  while (x<s->numfleets) {
    factionlist_add_fleet(&list,s->fleets[x]);
    x++;
  }
  while (body<s->numbodies) {
    factionlist_add_orbital_fleets(&list,&s->bodies[body]);
    factionlist_add_landed_fleets(&list,&s->bodies[body]);
    body++;
  }
  if (list.num) {
    printw("Fleets detected in sector %d/%d/%d: ",c->x,c->y,c->z);
    x=0;
    while (x<list.num) {
      if (x>0) {
        printw(", %s",list.strings[x]->str);
      } else {
        printw("%s",list.strings[x]->str);
      }
      x++;
    }
    dispose_gstrlist(&list);
    printw("\n");
    refresh();
  }
}

void ui_message(const char *message)
{
  attrset(COLOR_TEXT);
  printw("%s\n",message);
  refresh();
}

void ui_debug(const char *message)
{
  attrset(COLOR_TEXT);
  printw("Debug: %s\n",message);
  refresh();
}

void ui_message_sds(const char *message1, int d, const char *message2)
{
  printw("%s%d%s\n",message1,d,message2);
  refresh();
}

void ui_message_sd(const char *s, int d)
{
  printw("%s%d\n",s,d);
  refresh();
}

void ui_message_sss(const char *s1, const char *s2, const char *s3)
{
  attrset(COLOR_TEXT);
  printw("%s%s%s\n",s1,s2,s3);
  refresh();
}

void ui_message_sdsds(const char *s1, int d1, const char *s2, int d2,
  const char *s3)
{
  printw("%s%d%s%d%s\n",s1,d1,s2,d2,s3);
  refresh();
}

void ui_message_ss(const char *s1, const char *s2)
{
  attrset(COLOR_TEXT);
  printw("%s%s\n",s1,s2);
  refresh();
}

void ui_alertbox(char *message)
{
  ui_message(message);
  pause_prompt();
}

void set_chartitem_colour(char chartitem)
{
  if (chartitem & CHART_VISITED) {
    attrset(COLOR_CHART_VISITED);
  } else if (chartitem & CHART_FLEETSCANNED) {
    if ((chartitem & CHART_FULLSCANNED) ==CHART_FULLSCANNED) {
      attrset(COLOR_CHART_FULLSCANNED);
    } else {
      attrset(COLOR_CHART_FLEETSCANNED);
    }
  } else if (chartitem & CHART_BODYSCANNED) {
    attrset(COLOR_CHART_BODYSCANNED);
  } else {
    attrset(COLOR_CHART_BLANK);
  }
}

void ui_show_charts_help(void)
{
  clear();
  text("The following commands are supported in chart mode:\n\n");
  attrset(COLOR_HOTKEY);
  printw("Arrow keys");
  text(": Scroll chart in X and Y axes.\n\n");
  hotkey('+');
  text(" and ");
  hotkey('-');
  text(": Scroll chart on Z axis.\n\n");
  attrset(COLOR_HOTKEY);
  printw("Enter");
  text(": Move your ship to the currently-centered position on the chart.\n");
  printw(" (Note that you'll only be able to move your ship if you're not orbiting\n");
  printw(" or landed.)\n\n");
  attrset(COLOR_HOTKEY);
  printw("Tab");
  text(": Change emphasis mode.  At this time, the chart has two emphasis\n");
  printw("modes: Default and unexplored.  Default tries to show you as much\n");
  printw("information about areas you've explored (planets, fleets, how\n");
  printw("they've been scanned) as possible, while unexplored tries to show\n");
  printw("where the nearest unexplored areas are.\n\n");
  hotkey('Q');
  text(": Quit charts.  Leaves you wherever you last moved your ship to.\n\n");
  printw("Legend:\n");
  set_chartitem_colour(CHART_BLANK);
  addch('?');
  text(": Uncharted sector.\n");
  text("f: Fleet                   ");
  attrset(COLOR_CHART_FLEETSCANNED);
  printw(".");
  text(": Scanned for fleets, none found\n");
  text("o: Body(ies)               ");
  set_chartitem_colour(CHART_BODYSCANNED);
  addch('.');
  text(": Scanned for bodies, none found.\n");
  printw(".: Scanned sector          ");
  set_chartitem_colour(CHART_FULLSCANNED);  
  addch('.');
  text(": Scanned for bodies and fleets.\n+: Symbols following this are above your current Z-axis position.\n-: Symbols following are below your current Z-axis position.\n\n");
  
  pause_prompt();
}

void draw_chart_cyl(char *cyl, int planesize, int zzer, int zrange,
  int emphasize)
{
  int chars;
  int x;
  char chartitem=0;
  char cyllow=0;
  char cylhigh=0;
  x=(0-zrange)-zzer;
  if (zzer) {
    printw("");
  }
  switch (emphasize) {
  case EMPH_DEFAULT:
    while (x<zzer) {
      if (cyl[planesize*x]& (CHART_FLEETS|CHART_BODIES)) {
        cyllow |= cyl[planesize*x];
      }
      x++;
    }
    chartitem=cyl[0];
    x++;
    while (x<=zrange-zzer) {
      if (cyl[planesize*x]& (CHART_FLEETS|CHART_BODIES)) {
        cylhigh |= cyl[planesize*x];
      }
      x++;
    }
    break;
  case EMPH_UNSCANNED:
    while (x<zzer) {
      if ((cyl[planesize*x] & CHART_FULLSCANNED)==0) cyllow=0-x;
      x++;
    }
    chartitem=((cyl[0]&CHART_FULLSCANNED)==0);
    x++;
    while (x<zzer+zrange) {
      if ((cyl[planesize*x] & CHART_FULLSCANNED)==0) {
        cylhigh=x;
        x=zzer+zrange;
      }
      x++;
    }
    if (cyllow || cylhigh || chartitem) {
      attrset(COLOR_CHART_BLANK);
      if (cyllow) {
        if (cylhigh) {
          if (chartitem) {
            printw("?+-? ");
          } else {
            if (cylhigh==cyllow) {
              printw("+-%d? ",cylhigh);
            } else {
              printw("+-?  ");
            }
          }
        } else {
          if (chartitem) {
            printw("?-%d? ",cyllow);
          } else {
            printw("-%d?  ",cyllow);
          }
        }
      } else {
        if (cylhigh) {
          if (chartitem) {
            printw("?+%d? ",cylhigh);
          } else {
            printw("+%d?  ",cylhigh);
          }
        } else /* cyl only */ {
          printw("?    ");
        }
      }
      return;
    } else {
      chartitem=cyl[0];
    }
    break;
  default:
    ui_error("draw_chart_cyl has been fed an invalid chart feature to emphasize.");
    return;
  }
  set_chartitem_colour(chartitem);
  chars=5;
  if (chartitem==CHART_BLANK) {
    addch('?');
    chars--;
  } else {
    if (chartitem & CHART_COLONIES) {
      addch('c');
      chars--;
    } else if (chartitem & CHART_BODIES) {
      addch('o');
      chars--;
    }
    if (chartitem & CHART_FLEETS) {
      addch('f');
      chars--;
    }
    if (chartitem & CHART_MYFORCES) {
      addch('@');
      chars--;
    }
    if (chartitem & CHART_DETAILS) {
      addch('#');
      chars--;
    }
    if (chars==5) {
      addch('.');
      chars--;
    }
  }
  if (cyllow & CHART_BODIES) {
    if (cylhigh & CHART_BODIES) {
      set_chartitem_colour(cyllow & cylhigh);
      if (chars>=3) {
        printw("+-o");
        chars-=3;
      } else if (chars>=2) {
        printw("+-");
        chars-=2;
      }
    } else {
      if (chars) {
        set_chartitem_colour(cyllow);
        if (chars>=2) {
          printw("-o");
          chars-=2;
        } else {
          addch('-');
          chars--;
        }
      }
    }
  } else {
    if (cylhigh & CHART_BODIES) {
      set_chartitem_colour(cylhigh);
      if (chars>=2) {
        printw("+o");
        chars-=2;
      } else if (chars) {
        addch('+');
        chars--;
      }
    }
  }
  while (chars>0) {
    addch(' ');
    chars--;
  }
}

void show_chart_keys(int currz, int curremph)
{
  attrset(COLOR_TEXT);
  addch('[');
  attrset(COLOR_HOTKEY);
  printw("Enter");
  text("] Move    [");
  attrset(COLOR_HOTKEY);
  printw("Left");
  attrset(COLOR_TEXT);
  addch('/');
  attrset(COLOR_HOTKEY);
  printw("Right");
  text("]: move X   [");
  attrset(COLOR_HOTKEY);
  printw("Up");
  attrset(COLOR_TEXT);
  addch('/');
  attrset(COLOR_HOTKEY);
  printw("Down");
  text("]: move Y  [");
  hotkey('+');
  attrset(COLOR_TEXT);
  addch('/');
  hotkey('-');
  text("]: Move Z (");
  attrset(COLOR_INPUT);
  printw("%d",currz);
  text(")\n[");
  attrset(COLOR_HOTKEY);
  printw("Tab");
  text("]: Change emphasis (");
  attrset(COLOR_INPUT);
  switch(curremph) {
  case EMPH_DEFAULT:
    printw("Default");
    break;
  case EMPH_UNSCANNED:
    printw("Unscanned");
    break;
  default:
    attrset(COLOR_ERROR);
    printw("UNKNOWN");
  }
  text(")    [");
  hotkey('Q');
  text("]: Quit charts.\n");
}

void ui_show_chart(struct coord_t *c, int orientation, int emphasize)
{
  int maxy,maxx,miny=0,minx=0,horx=g_Config_Horizon,hory=g_Config_Horizon;
  int x;
  int buflen;
  int zrange=5;
  int planesize;
  char *slice;
  struct coord_t d,e;
  getmaxyx(stdscr,maxy,maxx);
  maxy-=5;
  maxx/=5;
  maxx-=2;
  if (maxx & 1) {
   maxx--;
  }
  d.y=c->y-(maxy/2);
  d.x=c->x-(maxx/2);
  d.z=c->z-zrange;
  if (d.y<0) miny=0-d.y;
  if (d.x<0) minx=0-d.x;
  e=(struct coord_t){x: c->x+(maxx/2)-1, y: c->y+(maxy/2), z: c->z+zrange};
  if (e.x>g_Config_Horizon) horx-=d.x;
  if (e.y>g_Config_Horizon) hory-=d.y;
  planesize=((e.x-d.x)+1)*((e.y-d.y)+1);
  buflen=planesize*((e.z-d.z)+1);
  slice=g_malloc(buflen);
  load_chart_slice(&d,&e,slice,buflen);
  clear();
  move(0,0);
  printw("Y \\ X");
  x=c->x-(maxx/2);
  attrset(COLOR_TEXT);
  while (x<c->x) {
    printw("%-5d",x);
    x++;
  }
  attrset(COLOR_PROMPT);
  printw("v%-4d",x++);
  attrset(COLOR_TEXT);
  while (x<c->x+(maxx/2)) {
    printw("%-5d",x);
    x++;
  }
  printw("Z=%d",c->z);
  move(1,75);
  printw("%d",c->z);
  move(1,0);
  d.y=0;
  while (d.y<maxy) {
    if (d.y!=maxy/2) {
      attrset(COLOR_TEXT);
      printw("%-5d",d.y+c->y-(maxy/2));
    } else {
      attrset(COLOR_PROMPT);
      printw("%-4d>",d.y+c->y-(maxy/2));
    }
    d.x=0;
    while (d.x<maxx) {
      if ((d.x<minx) || (d.y<miny) || (d.x>horx) || (d.y>hory)) {
        attrset(COLOR_CHART_BLANK);
        printw("     ");
      } else {
        draw_chart_cyl(&slice[d.x+(d.y*maxx)+((zrange)*planesize)],
          planesize,0,zrange,emphasize);
      }
      d.x++;
    }
    printw("\n");
    d.y++;
  }
  g_free(slice);
  show_chart_keys(c->z,emphasize);
  refresh();
}

void show_zone_composition(GString *z)
{
  printw("Composition: %s\n",z->str);
  refresh();
}

void show_body_scan(struct body_t *b)
{
  int zone=0;
  GString *s=body_mass_str(b->mass);
  attrset(COLOR_TEXT);
  printw("Body: %s Mass: %s Zones: %d\n",body_name(b),s->str,b->numzones);
  printw("Orbit: %.0f\n",b->orbradius);
  while (zone<b->numzones) {
    show_zone_composition(b->zones[zone]);
    zone++;
  }
  g_string_free(s,TRUE);
}

void show_location_orbit(struct body_t *b, char *sector)
{
  int x=0;
  printw("Orbiting %s in sector %s\n",body_name(b),sector);
  if (b->numofleets) {
    printw("Orbiting fleets:\n");
    while (x<b->numofleets) {
      printw("%s\n",b->ofleets[x]->str);
      x++;
    }
  }
}

void show_location_landed(struct body_t *b)
{
  int x=0;
  printw("Landed on %s in sector %s\n",b->name->str[0] ? b->name->str :
    "[Unnamed Body]" ,g_Player.coords->str);
  if (b->owner->len) {
    printw("Claimed by %s\n",b->owner->str);
  }
  printw("Class: %s",body_class_string(b->class));
  if (b->orbradius) {
    printw("    Orbit: %.0f",b->orbradius);
  }
  if (b->radius) {
    printw("    Radius: %d",b->radius);
  }
  printw("\n");
  if (b->desc->len) {
    printw("%s\n",b->desc->str);
  }
  if (b->numlfleets) {
    printw("Landed fleets:\n");
    while (x<b->numlfleets) {
      printw("%s\n",b->lfleets[x]->str);
      x++;
    }
  }
  if (b->numcolonies) {
    printw("Colonies:\n");
    x=0;
    while (x<b->numcolonies) {
      printw("%s\n",b->colonies[x]->str);
      x++;
    }
  }
}

void show_gstrlist(char *header, struct gstrlist_t *list)
{
  int maxx,maxy;
  int line=0,x=0;
  getmaxyx(stdscr,maxy,maxx);
  attrset(COLOR_TEXT);
  ui_message(header);
  while (x<list->num) {
    printw("%s\n",list->strings[x]->str);
    x++;
    line++;
    if (line>maxy-3) {
      pause_prompt();
      attrset(COLOR_TEXT);
      line=0;
    }
  }
}

void ui_pane_warning(struct fleetpane_t *pane, const char *s)
{
  wattrset(pane->win,COLOR_ERROR);
  wprintw(pane->win,"\n%s",s);
  wclrtoeol(pane->win);
  wprintw(pane->win,"\n");
  wattrset(pane->win,COLOR_PROMPT);
  wprintw(pane->win,"[Press any key]");
  wclrtoeol(pane->win);
  wrefresh(pane->win);
  getch();
  draw_fleet_pane(pane);
}

void draw_species_bar_checkbox(struct specie_pane_t *pane)
{
  move(pane->starty+(pane->pos-pane->page),3);
  if (pane->selected[pane->pos]) {
    addch('X');
  } else {
    addch(' ');
  }
  refresh();
}

void draw_species_bar_selected(struct specie_pane_t *pane)
{
  move(pane->starty+(pane->pos-pane->page),2);
  addch('<');
  move(pane->starty+(pane->pos-pane->page),4);
  addch('>');
  refresh();
}

void draw_species_bar_unselected(struct specie_pane_t *pane)
{
  move(pane->starty+(pane->pos-pane->page),2);
  addch(' ');
  move(pane->starty+(pane->pos-pane->page),4);
  addch(' ');
}

void draw_specie_pane(struct specie_pane_t *pane)
{
  int x,y;
  getmaxyx(stdscr,y,x);
  if (!pane->drawn) {
    if (y>pane->specie->numpops) {
      if (scrl(pane->specie->numpops)!=ERR) {
        getyx(stdscr,y,x);
        pane->starty=y-pane->specie->numpops;
        pane->pagesize=pane->specie->numpops+1;
      } else {
        ui_error("Error scrolling screen.");
      }
      pane->pagesize=y+1;
    } else {
      pane->starty=0;
      pane->pagesize=y-1;
    }
    pane->drawn=1;
  }
  move(pane->starty,0);
  x=0;
  while (x<lesser(pane->specie->numpops-pane->page,y)) {
    printw(" [ ");
    if (pane->selected[x+pane->page]) {
      addch('X');
    } else {
      addch(' ');
    }
    printw(" ] %-3d %s\n",
      habitability(pane->zone->str,&pane->specie->pops[x+pane->page].species),
      pane->specie->pops[x+pane->page].species.name->str);
    x++;
  }
  printw(" [ * ] Done ");
  draw_species_bar_selected(pane);
  refresh();
}

void hide_specie_pane(struct specie_pane_t *pane)
{
  int x,y;
  getmaxyx(stdscr,y,x);
  move(pane->starty+pane->pagesize,x-1);
  move(pane->starty+pane->pagesize+1,x-1);
  scrl(1);
  printw("\n\n",pane->pagesize);
  refresh();
}

void show_satellite_list(struct sector_t *sect, int body, int **list, int *num)
{
  int x=0;
  int lines=0;
  *list=g_malloc(sizeof(int));
  *num=1;
  if (body>=0) {
    if (sect->bodies[body].primary>-1) {
      *list[0]=sect->bodies[body].primary;
      printw("0: %s\n",body_name(&sect->bodies[*list[0]]));
    } else {
      printw("0: Leave orbit\n");
      *list[0]=-1;
    }
  }
  while (x<sect->numbodies) {
    if (sect->bodies[x].primary==body) {
      if (lines>20) {
        printw("[Mo[i]re]\n");
        lines=0;
        refresh();
        getch();
      }
      printw("%d: %s\n",*num,body_name(&sect->bodies[x]));
      lines++;
      ++*num;
      *list=g_realloc(*list,*num*sizeof(int));
      (*list)[*num-1]=x;
    }
    x++;
  }
  refresh();
}

void gstring_to_list(GString *orig, struct gstrlist_t *list, int maxlines)
{
  int	notdone=1;
  char *line=orig->str;
  int	x=0;
  while (notdone) {
    if (orig->str[x]=='\n') {
      orig->str[x]='\0';
      gstrlist_add(list,line);
      orig->str[x]='\n';
      line=&orig->str[x+1];
    }
    if (list->num>=maxlines) {
      notdone=0;
    } else {
      x++;
      if (x>orig->len) {
        gstrlist_add(list,line);
        notdone=0;
      }
    }
  }
  if (list->num==0) {
    gstrlist_add(list,"");
  }
}

struct edit_window {
  int               lin;
  int               col;
  struct gstrlist_t block;
  int               statline;
  int               maxlin;
  int               maxcol;
  const char        *header;
  int               size;
};

void draw_edit_window(struct edit_window *e)
{
  int lin=0;
  clear();
  attrset(COLOR_INPUT);
  move(1,0);
  while (lin<e->block.num) {
    printw("%s\n",e->block.strings[lin]->str);
    lin++;
  }
  move(0,0);
  attrset(COLOR_TEXT);
  printw("%s",e->header);
  if (e->maxlin>e->size+2) {
    move(e->size+1,0);
    addch('[');
    attrset(COLOR_HOTKEY);
    printw("F9");
    text("] Quit without saving.            [");
    attrset(COLOR_HOTKEY);
    printw("F10");
    text("]: Quit and save.");
  }
  move(e->lin+1,e->col);
  attrset(COLOR_INPUT);
}

int trim_lines(struct gstrlist_t *list, int cols)
{
  int ret=0;
  int x=0;
  while (x<list->num) {
    if (list->strings[x]->len>cols) {
      ret++;
      g_string_truncate(list->strings[x],cols);
    }
    x++;
  }
  return ret;
}

void init_editor(struct edit_window *e,
  const char *header, GString *orig, int size)
{
  e->lin=0;
  e->col=0;
  e->header=header;
  e->block=null_gstrlist;
  gstring_to_list(orig,&e->block,size);
  e->statline=24;
  getmaxyx(stdscr,e->maxlin,e->maxcol);
  if (e->maxcol>78) {
    e->maxcol=78;
  }
  e->size=size;
  if (size+2<=e->maxlin) {
    e->statline=size+2;
  }
}

void edit_down(struct edit_window *e)
{
  if (e->lin<e->size-1) {
    e->lin++;
    if (e->block.num<=e->lin) {
      gstrlist_add(&e->block,"");
    }
    if (e->block.strings[e->lin]->len<e->col) {
      e->col=e->block.strings[e->lin]->len;
    }
    move(e->lin+1,e->col);
  }
}

void edit_up(struct edit_window *e)
{
  if (e->lin>0) {
    e->lin--;
    if (e->block.strings[e->lin]->len<e->col) {
      e->col=e->block.strings[e->lin]->len;
    }
    move(e->lin+1,e->col);
  }
}

void edit_backspace(struct edit_window *e)
{
  GString *old;
  if (e->col>0) {
    if (e->col>e->block.strings[e->lin]->len) {
      e->col--;
      return;
    }
    old=e->block.strings[e->lin];
    old->str[e->col-1]='\0';
    e->block.strings[e->lin]=g_string_new(old->str);
    g_string_append(e->block.strings[e->lin],&old->str[e->col]);
    e->col--;
    move(e->lin+1,e->col);
    printw("%s",&e->block.strings[e->lin]->str[e->col]);
    addch(' ');
    move(e->lin+1,e->col);
    g_string_free(old,TRUE);
  }
}

void edit_del(struct edit_window *e)
{
  GString *old;
  if (e->col>0) {
    if (e->col>=e->block.strings[e->lin]->len) {
      return;
    }
    old=e->block.strings[e->lin];
    old->str[e->col]='\0';
    e->block.strings[e->lin]=g_string_new(old->str);
    if (e->col<old->len-1) {
      g_string_append(e->block.strings[e->lin],&old->str[e->col+1]);
    }
    g_string_free(old,TRUE);
  } else {
    if (e->block.strings[e->lin]->len>0) {
      old=e->block.strings[e->lin];
      e->block.strings[e->lin] = g_string_new(&old->str[1]);
      g_string_free(old,TRUE);
    } else {
      return;
    }
  }
  move(e->lin+1,e->col);
  printw("%s",&e->block.strings[e->lin]->str[e->col]);
  addch(' ');
  move(e->lin+1,e->col);
}

GString *ui_edit_block(const char *header, GString *orig, int maxlines)
{
  struct edit_window e;
  int	notdone=1;
  int   key;
  char  shortstr[2]={' ','\0'};
  int	x;
  GString *ret;
  clear();
  init_editor(&e,header,orig,maxlines);
  x=trim_lines(&e.block,78);
  draw_edit_window(&e);
  if (x) {
    move(e.statline,1);
    attrset(COLOR_ERROR);
    if (x==1) {
      printw("1 outsized line truncated to 78 characters.");
    } else {
      printw("%d outsized lines truncated to 78 characters.",x);
    }
    attrset(COLOR_INPUT);
    move(e.lin+1,e.col);
    refresh();
  }
  move(1,0);
  refresh();
  notdone=1;
  while (notdone) {
    key=getch();
    switch (key) {
    case 0xA:
    case 0xD:
      edit_down(&e);
      e.col=0;
      move(e.lin+1,e.col);
      break;
    case 0xC:
      draw_edit_window(&e);
      break;
    case 0x8:
    case KEY_BACKSPACE:
      edit_backspace(&e);
      break;
    case KEY_DC:
      edit_del(&e);
      break;
    case KEY_DOWN:
      edit_down(&e);
      break;
    case KEY_END:
      e.col=e.block.strings[e.lin]->len;
      move(e.lin+1,e.col);
      break;
    case KEY_F(9):
      dispose_gstrlist(&e.block);
      return g_string_new(orig->str);
      break;
    case KEY_F(10):
      notdone=0;
      break;
    case KEY_HOME:
      e.col=0;
      move(e.lin+1,0);
      break;
    case KEY_LEFT:
      if (e.col>0) {
        e.col--;
      }
      move(e.lin+1,e.col);
      break;
    case KEY_RIGHT:
      if (e.col<e.maxcol) {
        e.col++;
      }
      move(e.lin+1,e.col);
      break;
    case KEY_UP:
      edit_up(&e);
      break;
    default:
      if (key>=' ' && key<='~') {
        if (e.block.strings[e.lin]->len>=78) {
          beep();
          move(e.statline,1);
          attrset(COLOR_TEXT);
          clrtoeol();
          printw("Max line length is 78 characters.");
          attrset(COLOR_INPUT);
          refresh();
          move(e.lin+1,e.col);
        } else {
          if (e.col>=e.block.strings[e.lin]->len) {
            shortstr[0]=key;
            g_string_append(e.block.strings[e.lin],&shortstr[0]);
            addch(key);
            refresh();
          } else {
            shortstr[0]=key;
            if (e.col>0) {
              g_string_insert(e.block.strings[e.lin],e.col,&shortstr[0]);
            } else {
              g_string_prepend(e.block.strings[e.lin],&shortstr[0]);
            }
            printw("%s",&e.block.strings[e.lin]->str[e.col]);
            move(e.lin+1,e.col);
            refresh();
          }
          e.col++;
          move(e.lin+1,e.col);
        }
      } else {
        move(e.statline,1);
        attrset(COLOR_TEXT);
        clrtoeol();
        printw("Odd key 0x%X pressed.   ",key);
        attrset(COLOR_INPUT);
        move(e.lin+1,e.col);
        refresh();
      }
    }
  }
  ret=g_string_new(e.block.strings[0]->str);
  x=1;
  while (x<e.block.num) {
    g_string_append(ret,"\n");
    g_string_append(ret,e.block.strings[x]->str);
    x++;
  }
  dispose_gstrlist(&e.block);
  return ret;
}

void ui_show_chatline(const char *line)
{
  char *colon=strstr(line,": ");
  GString *s;
  if (colon) {
    s=g_string_new(line);
    g_string_truncate(s,colon-line);
    attrset(COLOR_HOTKEY);
    printw("%s",s->str);
    g_string_free(s,TRUE);
    text(": ");
    attrset(COLOR_INPUT);
    printw("%s\n",&line[(colon-line)+2]);
  } else {
    attrset(COLOR_ERROR);
    printw("Invalid chat line '%s'\n",line);
  }
}

struct ui_progbar_t ui_progbar_init(const char *label, int min, int max)
{
  struct ui_progbar_t bar;
  int x,y,maxx;
  bar.text=g_string_new(label);
  bar.min=min;
  bar.max=max;
  bar.curr=min;
  bar.lastupdated=min;
  getmaxyx(stdscr,y,maxx);
  bar.cellsize=(max-min)/(maxx-3);
  bar.currcell=1;
  getyx(stdscr,y,x);
  bar.screenline=y;
  move(y,maxx-12);
  attrset(COLOR_LIGHTBAR_UNSELECTED);
  printw("%-10d",max);
  attrset(COLOR_TEXT);
  addch('}');
  move(y,0);
  addch('{');
  attrset(COLOR_LIGHTBAR_UNSELECTED);
  printw("%d %s",min,label);
  refresh();
  return bar;
}

void ui_progbar_advance(struct ui_progbar_t *bar, int to)
{
  bar->curr=to;
  while (to-bar->lastupdated>bar->cellsize) {
    to-=bar->cellsize;
    move(bar->screenline,bar->currcell);
    attrset(COLOR_LIGHTBAR_UNSELECTED|A_STANDOUT);
    addch(A_CHARTEXT & inch());
    bar->lastupdated+=bar->cellsize;
    bar->currcell++;
  }
  refresh();
}

void ui_progbar_destroy(struct ui_progbar_t *bar)
{
  int y,x;
  g_string_free(bar->text,TRUE);
  getyx(stdscr,y,x);
  if (y==bar->screenline) {
    move(y,0);
    clrtoeol();
    refresh();
  }
}

int screen_coord(float y, int maxy)
{
  if (y>=1) {
    if (y>maxy) {
      return maxy;
    }
    return lrint(y);
  }
  if (y<=-1) {
    return maxy+lrint(y);
  }
  if (y==0) {
    return 0;
  }
  y*=maxy;
  if (y<0) {
    return maxy+lrint(y);
  }
  return y;
}

void ui_pane_create(struct pane_t *pane, float ymin, float xmin, float ymax, float xmax)
{
  int	symax,sxmax;
  getmaxyx(stdscr,symax,sxmax);
  pane->ymin=screen_coord(ymin,symax);
  pane->ymax=screen_coord(ymax,symax);
  pane->xmin=screen_coord(xmin,sxmax);
  pane->xmax=screen_coord(xmax,sxmax);
  pane->win=subwin(stdscr,
    pane->ymax-pane->ymin,pane->xmax-pane->xmin,
	pane->ymin,pane->xmin);
}

struct listpane_t *ui_listpane_create(struct listpane_data_t *data,
  float ymin, float xmin, float ymax, float xmax,
  void pane_draw_item(struct listpane_t *pane, int num, int style))
{
  /* xyminax can be: 0 - position 0.  >=1: Integer position.
     <=-1: That many cells back from the end of the screen.
     Positive fractions: That much of the screen's width, from 0.
     Negative fractions: That much of the screen's width, from max.
     */
  struct listpane_t *pane=g_malloc(sizeof(struct listpane_t));
  ui_pane_create(&pane->pane,ymin,xmin,ymax,xmax);
  pane->data = *data;
  pane->drawcode = pane_draw_item;
  pane->pagesize = pane->pane.ymax-pane->pane.ymin;
  pane->page = 0;
  pane->lin = 0;
  return pane;
}

void dispose_listpane(struct listpane_t *pane)
{
  delwin(pane->pane.win);
  g_free(pane);
}

void draw_habitatpane_line(struct listpane_t *pane, int num, int style)
{
  struct habitatreq_item *h=(struct habitatreq_item *)&pane->data.data[num*pane->data.size];
  if (pane->data.size != sizeof(struct habitatreq_item)) {
    ui_error("draw_habitatpane_line on mismatched habitat.");
    return;
  }
  if (num>pane->page+pane->pagesize || num <pane->page ) {
    ui_error("Attempt to draw habitatpane line out of bounds.");
    return;
  }
  if (num>=pane->data.num) {
    return;
  }
  wmove(pane->pane.win, num-pane->page, pane->pane.xmin);
  switch (style) {
  case STYLE_UNSELECTED:
    wattrset((pane->pane.win), COLOR_LIGHTBAR_UNSELECTED);
    wprintw(pane->pane.win,"%s %s %.2g%%", h->chem->str, (h->cond=='<' ? "less than" : 
      h->cond=='>' ? "more than" : "wibbly-errata"), h->amt/100.0);
    break;
  case STYLE_SELECTED:
    wattrset(pane->pane.win, COLOR_LIGHTBAR_UNSELECTED|A_STANDOUT);
    wprintw(pane->pane.win,"%s %s %.2g%%", h->chem->str, (h->cond=='<' ? "less than" :
      h->cond=='>' ? "more than" : "wibbly-errata"), h->amt/100.0);
    break;
  default:
    ui_error("Incomprehensible style number");
  }
  /* Does not refresh */
}

void draw_listpane(struct listpane_t *pane)
{
  int x=pane->page;
  wclear(pane->pane.win);
  while (x<pane->page+pane->pagesize) {
    pane->drawcode(pane,x,STYLE_UNSELECTED);
    x++;
  }
  pane->drawcode(pane,pane->lin,STYLE_SELECTED);
  wrefresh(pane->pane.win);
}

int ui_habitat_edit_prompt(int points)
{
  int c;
  int x,y;
  getmaxyx(stdscr,y,x);
  move (y-1,0);
  attrset(COLOR_PROMPT);
  printw("Editing habitat [");
  attrset(COLOR_INPUT);
  printw("%d",points);
  attrset(COLOR_PROMPT);
  printw(" bioengineering points] [");
  hotkey('?');
  attrset(COLOR_PROMPT);
  printw(": Help]: ");
  attrset(COLOR_INPUT);
  c=getch();
  if (c>=' ' && c<=0x7E) {
    addch(c);
  }
  return c;
}

void listpane_down(struct listpane_t *pane)
{
  if (pane->lin<pane->data.num-1) {
    pane->drawcode(pane,pane->lin,STYLE_UNSELECTED);
    pane->lin++;
    if (pane->lin>=pane->page+pane->pagesize) {
      pane->page+=pane->pagesize;
      draw_listpane(pane);
    } else {
      pane->drawcode(pane,pane->lin,STYLE_SELECTED);
      wrefresh(pane->pane.win);
    }
  }
}

void listpane_pagedown(struct listpane_t *pane)
{
  if (pane->page+pane->pagesize<pane->data.num) {
    pane->page+=pane->pagesize;
    pane->lin+=pane->pagesize;
    if (pane->lin>=pane->data.num) {
      pane->lin=pane->data.num-1;
    }
    draw_listpane(pane);
  }
}

void listpane_up(struct listpane_t *pane)
{
  if (pane->lin>0) {
    pane->drawcode(pane,pane->lin,STYLE_UNSELECTED);
    pane->lin--;
    if (pane->lin>pane->page) {
      pane->page-=pane->pagesize;
      draw_listpane(pane);
    } else {
      pane->drawcode(pane,pane->lin,STYLE_SELECTED);
      wrefresh(pane->pane.win);
    }
  }
}

void listpane_pageup(struct listpane_t *pane)
{
  if (pane->lin>=pane->pagesize) {
    pane->lin-=pane->pagesize;
    pane->page-=pane->pagesize;
    draw_listpane(pane);
  }
}

void ui_show_pop_totals(struct pop_totals_t *populii)
{
  int x=0;
  if (populii->total==0) {
    text("No population.\n");
    return;
  }
  while (x<populii->list.num) {
    printw("%10d %s\n",populii->total[x],populii->list.strings[x]->str);
    x++;
  }
}

void show_population_menu(void)
{
  if (g_Player.locus==LOCUS_DOCKED) {
    hotkey('D');
    text(": Depopulate colony\n");
    hotkey('R');
    text(": Recruit colonists\n");
  }
  hotkey('T');
  text(": Train crew\n");
  hotkey('Q');
  text(": Quit to main\n");
}
