#include <stdlib.h>
#include <stdio.h>
#include "microcosm.h"
#include "util.h"
#include "ui_all.h"

#define cgi_url "/technocracy/makrokosmos/makrokosmos.cgi"

int ui_input_read=0;
int ui_input_present=0;
struct gstrlist_t cgi_variable_names=null_gstrlist;
struct gstrlist_t cgi_variable_data=null_gstrlist;
GString *cgi_message_string=NULL;

void ui_init()
{
  cgi_message_string=g_string_new("");
}

void cgi_push_message(const char *s)
{
  g_string_append(cgi_message_string,s);
}

GString *cgi_variable()
{
  return g_string_new("ARGH!");
}

void read_cgi_input()
{
  char buf[1025];
  printf("Content-type: text/plain\n\n");
  while (fgets(&buf[0],1024,stdin)) {
    printf("Debug: %s\n",&buf[0]);
    buf[1024]='\0';
    if (feof(stdin)) {
      return;
    }
  }
}

int input_present()
{
  if (ui_input_read==0) {
    read_cgi_input();
    ui_input_read++;
    ui_input_present=1;
  }
  return ui_input_present;
}

void cgi_print_header()
{
  printf("Content-type: text/plain\015\012\015\012");
  printf("<HTML><HEAD><TITLE>Makrokosmos/WWW</TITLE></HEAD>\n");
  printf("<BODY bgcolor=\"#000000\" text=\"#00ffcc\" link=\"#ffffff\"\n");
  printf("alink=\"#cc00ff\" vlink=\"#999999\">\n");
}

void cgi_print_footer()
{
  printf("</BODY></HTML>\n\n");
}

void cgi_print_form(const char *prompt, int size)
{
  cgi_print_header();
  printf("%s",cgi_message_string->str);
  printf("<FORM ACTION=\"%s\" METHOD=\"POST\"><P>%s</P>\n",cgi_url,prompt);
  printf("<INPUT TYPE=TEXT NAME=\"%s\" size=%d>\n",prompt,size);
  printf("<INPUT TYPE=SUBMIT VALUE=\"Submit\"></FORM>\n");
  cgi_print_footer();
  exit(0);
}

int read_int(const char *prompt)
{
  GString *cgivar;
  int c;
  if (input_present()) {
    cgivar=cgi_variable(prompt);
    c=atoi(cgivar->str);
    return c;
  }
  cgi_print_form("prompt",7);
}

GString *read_gstring(const char *prompt)
{
  GString *cgivar;
  if (input_present()) {
    cgivar=cgi_variable(prompt);
    if (cgivar) {
      return cgivar;
    }
  }
  cgi_print_form("prompt",72);
}

int read_char(const char *prompt)
{
  GString *cgivar;
  if (input_present()) {
    cgivar=cgi_variable(prompt);
    if (cgivar->len) {
      return (int)cgivar->str[0];
    }
  }
  cgi_print_form(prompt,1);
}

void ui_error(const char *s)
{
  GString *str=g_string_new("");
  g_string_sprintf(str,"<P><B>ERROR: %s</B></P>\n",s);
  cgi_push_message(str->str);
  g_string_free(str,TRUE);
}

void ui_error_sss(const char *s1, const char *s2, const char *s3)
{
  GString *s=g_string_new("");
  g_string_sprintf(s,"<P><B>ERROR: %s%s%s</B></P>\n",s1,s2,s3);
  cgi_push_message(s->str);
  g_string_free(s,TRUE);
}

void ui_error_sd(const char *s, int d)
{
  GString *str=g_string_new("");
  g_string_sprintf(str,"<P><B>ERROR: %s%d</B></P>\n",s,d);
  cgi_push_message(str->str);
  g_string_free(str,TRUE);
}

void ui_config_error(const char *message, int err)
{
  GString *s=g_string_new("");
  g_string_sprintf(s,"<P><B>ERROR: %s (code %d)</B></P>\n",message,err);
  cgi_push_message(s->str);
  g_string_free(s,TRUE);
}

void ui_message(const char *s)
{
  GString *str=g_string_new("");
  g_string_sprintf(str,"<P>%s</P>\n",s);
  cgi_push_message(str->str);
  g_string_free(str,TRUE);
}

void ui_message_ss(const char *s1, const char *s2)
{
  GString *s=g_string_new("");
  g_string_sprintf(s,"<P>%s%s</P>\n",s1,s2);
  cgi_push_message(s->str);
  g_string_free(s,TRUE);
}

void ui_message_sds(const char *s1, int d, const char *s2)
{
  GString *s=g_string_new("");
  g_string_sprintf(s,"<P>%s%d%s</P>\n",s1,d,s2);
  cgi_push_message(s->str);
  g_string_free(s,TRUE);
}

void pause_prompt()
{
  /* Does nothing in CGI. */
}

void ui_show_charts_help(void)
{
  ui_error("Chart help for http not yet written.");
}

void ui_show_chart(void)
{
  ui_error("ui_show_chart for http not yet written.");
}

void draw_pane_bar_selected(void)
{
  ui_error("draw_pane_bar_selected for http not yet written.");
}

void draw_pane_bar_unselected(void)
{
  ui_error("draw_pane_bar_unselected for http not yet written.");
}

void draw_ship_bar_target(void)
{
  ui_error("draw_ship_bar_target for http not yet written.");
}

void draw_fleet_pane(void)
{
  ui_error("draw_fleet_pane for http not yet written.");
}

void pane_up(void)
{
  ui_error("pane_up for http not yet written.");
}

void pane_down(void)
{
  ui_error("pane_down for http not yet written.");
}

void ui_pane_warning(const char *s)
{
  ui_error(s);
}

void show_tech_difference(struct tech_t *real, struct tech_t *req)
{
  ui_error("show_tech_difference not yet written for http.");
}

void show_gstrlist (char *header, struct gstrlist_t *list)
{
  GString *s=g_string_new("");
  int x=0;
  ui_message(header);
  while (x<list->num) {
    g_string_sprintf(s,"%s",list->strings[x]->str);
    x++;
  }
  g_string_free(s,TRUE);
}

void show_zone_composition(GString *z)
{
  ui_message_ss("Zone composition: ",z->str);
}

void ui_add_scan_result_fleets(void)
{
  ui_error("ui_add_scan_result_fleets() not yet written."); //
}
void ui_add_scan_result_bodies(void)
{
  ui_error("ui_add_scan_result_bodies() not yet written."); //
}

void draw_equip_footer(void)
{
  ui_error("draw_equip_footer for http irrelevant.");
}

void show_body_scan(struct body_t *b)
{
  GString *s=g_string_new("");
  GString *mass=body_mass_str(b->mass);
  g_string_sprintf(s,"Body: %s   Mass: %s  Zones: %d",
    body_name(b), mass->str, b->numzones);
  ui_message(s->str);
  g_string_free(s,TRUE);
  g_string_free(mass,TRUE);
}
