#include <sys/file.h>
#include <fcntl.h>
#include <time.h>
#include <stdio.h>
#include <glib.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include "microcosm.h"
#include "util.h"
#include "ui_all.h"

int gstreq(const GString *s1, const GString *s2)
{
  if (s1==s2) {
     return TRUE;
  }
  if (s1->len==s2->len) {
    return (strcmp(s1->str,s2->str)==0);
  }
  return FALSE;
}

void g_string_assign_interpolated(GString *s, char *orig)
{
  char *newline;
  g_string_assign(s,"");
  while ((newline=strstr(orig,"\\n"))) {
    newline[0]='\0';
    g_string_append(s,orig);
    newline[0]='\\';
    g_string_append(s,"\n");
    orig=newline+2;
  }
  g_string_append(s,orig);
}

int gstrlistindex_c(struct gstrlist_t *l, GString *s)
{
  int x=0;
  while (x<l->num) {
    if (gstreq(l->strings[x],s)) {
      return x;
    }
    x++;
  }
  l->num++;
  l->strings=g_realloc(l->strings,l->num*sizeof(GString *));
  l->strings[l->num-1]=g_string_new(s->str);
  return l->num-1;
}

int gstrlistindex(const struct gstrlist_t *l, const GString *s)
{
  int x=0;
  while (x<l->num) {
    if (gstreq(l->strings[x],s)) {
      return x;
    }
    x++;
  }
  return -1;
}

int gstrlistindex_str_c(struct gstrlist_t *l, const char *s)
{
  int len=strlen(s);
  int x=0;
  while (x<l->num) {
    if (l->strings[x]->len==len) {
      if (strcmp(l->strings[x]->str,s)==0) {
        return x;
      }
    }
    x++;
  }
  l->num++;
  l->strings=g_realloc(l->strings,l->num*sizeof(GString *));
  l->strings[l->num-1]=g_string_new(s);
  return l->num-1;
}

void dispose_gstrlist(struct gstrlist_t *l)
{
  if (l->strings) {
    g_free(l->strings);
  }
  *l=null_gstrlist;
}

double lesser_d(double x, double y)
{
  if (x<y) {
    return x;
  }
  return y;
}

void log_to_file(char *file, char *s1, char *s2)
{
  GString *logname=g_string_new(g_Config.datapath->str);
  FILE *f;
  time_t now;
  int fd;
  struct tm *now_tm;
  g_string_append(logname,file);
  fd=open(logname->str,O_CREAT|O_RDWR, g_Config.datamode);
  if (fd>=0) {
    if (flock(fd,LOCK_EX)==0) {
      f=fdopen(fd,"a");
      if (f) {
        now=time(NULL);
        now_tm=gmtime(&now);
        fprintf(f,"%.24s: %s=%s\n",asctime(gmtime(&now)),s1,s2);
        fclose(f);
      } else {
        ui_error("Error fopen()ing logfile.");
      }
      flock(fd,LOCK_UN);
    } else {
      ui_error("Error locking logfile.");
    }
    close(fd);
  } else {
    ui_error("Error fdopen()ing logfile.");
  }
  g_string_free(logname,TRUE);
}

void log_public(char *s1, char *s2)
{
  log_to_file(MAKROKOSMOS_LOG_PUBLIC,s1,s2);
}

void log_private(char *s1, char *s2)
{
  log_to_file(MAKROKOSMOS_LOG_PRIVATE,s1,s2);
}

void gstrlist_add(struct gstrlist_t *l, char *s)
{
  l->num++;
  l->strings=g_realloc(l->strings,l->num*sizeof(GString *));
  l->strings[l->num-1]=g_string_new(s);
}

void split(char *c, GString **cmd, GString **args)
{
  char *spacepos=strstr(c,"=");
  if (spacepos) {
    spacepos[0]='\0';
    *cmd=g_string_new(c);
    *args=g_string_new(spacepos+1);
    spacepos[0]='=';
  } else {
    *cmd=g_string_new(c);
    *args=g_string_new("");
  }
}

int randval_i(int max)
{
  return ((double)random()/RAND_MAX)*max;
}

double randval_d(double max)
{
  return (double)((double)random()/RAND_MAX)*max;
}

void write_string_interpolated(FILE *f, char *s)
{
  char *newline;
  while ((newline=strstr(s,"\n"))) {
    newline[0]='\0';
    fprintf(f,"%s\\n",s);
    newline[0]='\n';
    s=newline+1;
  }
  fprintf(f,"%s",s);
}
